/*******************************************************************************
* FILE NAME: ibasstrm.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*     IBaseStream                                                              *                                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Taligent.    *
*   (C) Copyright Taligent, Inc. 1996 - All rights reserved.                   *
*                                                                              *
*******************************************************************************/

#include <ibasstrm.hpp>
#include <icontext.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBASSTRM_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| IBaseStream::IBaseStream                                                     |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IBaseStream::IBaseStream()
{
    fContext = 0;
    fDefaultContext = new IContext;
}

/*------------------------------------------------------------------------------
| IBaseStream::~IBaseStream                                                    |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IBaseStream::~IBaseStream()
{
   delete fDefaultContext;
}



/*------------------------------------------------------------------------------
| IBaseStream::setContext                                                      |
|                                                                              |
| Store the passed context.                                                    |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::setContext(IContext* c)
{
    fContext = c;
    return *this;
}

/*------------------------------------------------------------------------------
| IBaseStream::context                                                         |
|                                                                              |
| Return the current context.                                                  |
------------------------------------------------------------------------------*/
IContext*
IBaseStream::context() const
{
    return fContext;
}

/*------------------------------------------------------------------------------
| IBaseStream::defaultContext                                                  |
|                                                                              |
| Return the default context.                                                  |
------------------------------------------------------------------------------*/
IContext*
IBaseStream::defaultContext() const
{
    return ((IContext*) fDefaultContext);
}


/*------------------------------------------------------------------------------
| IBaseStream::seekRelative                                                    |
|                                                                              |
| Seek to an offset from the current position in the stream.                   |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::seekRelative(IBaseStream::PositionDelta offset)
{
    seek((IBaseStream::Position) (offset + position()));
    return *this;
}

/*------------------------------------------------------------------------------
| IBaseStream::setLogicalEndOfStream                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::setLogicalEndOfStream (IBaseStream::Position endOfStream)
{
    setPhysicalEndOfStream(endOfStream);
    return *this;

}

/*------------------------------------------------------------------------------
| IBaseStream::logicalEndOfStream                                              |
|                                                                              |
| Return the position of the logical end of the stream.                        |
------------------------------------------------------------------------------*/
IBaseStream::Position
IBaseStream::logicalEndOfStream () const
{
    return physicalEndOfStream();
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read bytes of void* data.                                |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(void* what, size_t byteCount)
{
    handleReadBufferEmpty(what, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write bytes of void* data.                               |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const void* what,  size_t byteCount)
{
    handleWriteBufferFull(what, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read null terminated char* data.                         |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(char* s)
{
    unsigned long i = 0;
    do {
        read((void*)&s[i], IBaseStream::kCharBytes);
    } while (s[i++] != '\0');
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read a signed character.                                 |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(signed char& c)
{
   handleReadBufferEmpty((void*) &c, IBaseStream::kSignedCharBytes);
   return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an unsigned character.                              |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(unsigned char& c)
{
   handleReadBufferEmpty((void*) &c, IBaseStream::kUnsignedCharBytes);
   return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read a short.                                            |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(short& c)
{
   handleReadBufferEmpty((void*) &c, IBaseStream::kShortBytes);
   return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an unsigned short.                                  |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(unsigned short& c)
{
    handleReadBufferEmpty((void*) &c, IBaseStream::kUnsignedShortBytes);
    return *this;
}



/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read a long.                                             |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(long& c)
{
    handleReadBufferEmpty((void*) &c, IBaseStream::kLongBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an unsigned long.                                   |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(unsigned long& c)
{
    handleReadBufferEmpty((void*) &c, IBaseStream::kUnsignedLongBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read a float.                                            |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(float& c)
{
    handleReadBufferEmpty((void*) &c, IBaseStream::kFloatBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read a double.                                           |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(double& c)
{
    handleReadBufferEmpty((void*) &c, IBaseStream::kDoubleBytes);
    return *this;
}



/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write a null terminated character string.                |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const char* c)
{
    size_t byteCount = strlen(c) + 1;
    handleWriteBufferFull((void*)c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write a signed character.                                |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(signed char c)
{
    handleWriteBufferFull((const void*) &c, IBaseStream::kSignedCharBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an unsigned character.                             |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(unsigned char c)
{
    handleWriteBufferFull((const void*) &c, IBaseStream::kUnsignedCharBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write a short.                                           |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(short c)
{
    handleWriteBufferFull((const void*) &c, IBaseStream::kShortBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an unsigned short.                                 |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(unsigned short c)
{
    handleWriteBufferFull((const void*) &c, IBaseStream::kUnsignedShortBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write a long.                                            |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(long c)
{
   handleWriteBufferFull((const void*) &c, IBaseStream::kLongBytes);
   return *this;
}



/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an unsigned long.                                  |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(unsigned long c)
{
    handleWriteBufferFull((const void*) &c, IBaseStream::kUnsignedLongBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write a float.                                           |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(float c)
{
    handleWriteBufferFull((const void*) &c, IBaseStream::kFloatBytes);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write a double.                                          |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(double c)
{
    handleWriteBufferFull((const void*) &c, IBaseStream::kDoubleBytes);
    return *this;
}



/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of signed characters.                      |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(signed char c[], unsigned long count)
//  Note:   This is optimized, knowing that IBaseStream::kSignedCharBytes == 1.
{
    handleReadBufferEmpty((void*) c, (size_t)(IBaseStream::kSignedCharBytes*count));
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of unsigned characters.                    |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(unsigned char c[], unsigned long count)
{
    handleReadBufferEmpty((void*) c, (size_t)(IBaseStream::kUnsignedCharBytes*count));
    return *this;
}

/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of shorts.                                 |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(short c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kShortBytes);
    handleReadBufferEmpty((void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of unsigned shorts.                        |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(unsigned short c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kUnsignedShortBytes);
    handleReadBufferEmpty((void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of longs.                                  |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(long c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kLongBytes);
    handleReadBufferEmpty((void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of unsigned longs.                         |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(unsigned long c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kUnsignedLongBytes);
    handleReadBufferEmpty((void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of floats.                                 |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(float c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kFloatBytes);
    handleReadBufferEmpty((void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::read                                                            |
|                                                                              |
| Stream primitive to read an array of doubles.                                |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::read(double c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kDoubleBytes);
    handleReadBufferEmpty((void*) c, byteCount);
    return *this;
}



/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of signed characters.                     |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const signed char c[], unsigned long count)
//  Note:   This is optimized, knowing that IBaseStream::kSignedCharBytes == 1.
{
    handleWriteBufferFull((void*) c, (size_t)(IBaseStream::kSignedCharBytes*count));
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of unsigned characters.                   |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const unsigned char c[], unsigned long count)
{
    handleWriteBufferFull((void*) c, (size_t)(IBaseStream::kUnsignedCharBytes*count));
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of shorts.                                |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const short c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kShortBytes);
    handleWriteBufferFull((const void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of unsigned shorts.                       |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const unsigned short c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kUnsignedShortBytes);
    handleWriteBufferFull((const void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of longs.                                 |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const long c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kLongBytes);
    handleWriteBufferFull((const void*) c, byteCount);
    return *this;
}


/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of unsigned longs.                        |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const unsigned long c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kUnsignedLongBytes);
    handleWriteBufferFull((const void*) c, byteCount);
    return *this;
}



/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of floats.                                |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const float c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kFloatBytes);
    handleWriteBufferFull((const void*) c, byteCount);
    return *this;
}



/*------------------------------------------------------------------------------
| IBaseStream::write                                                           |
|                                                                              |
| Stream primitive to write an array of doubles.                               |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::write(const double c[], unsigned long count)
{
    size_t byteCount = (size_t)(count * IBaseStream::kDoubleBytes);
    handleWriteBufferFull((const void*) c, byteCount);
    return *this;
}





/*------------------------------------------------------------------------------
| IBaseStream::flush                                                           |
|                                                                              |
| Tell the stream that output buffer is to be flushed.                         |
------------------------------------------------------------------------------*/
IBaseStream&
IBaseStream::flush()
{
    // this function intentionally left blank
    return *this;
}

/*------------------------------------------------------------------------------
| IBaseStream::flush                                                           |
|                                                                              |
| Return true if the stream can be written to.                                 |
------------------------------------------------------------------------------*/
IBase::Boolean IBaseStream::isWriteable() const
{
    return (true);
}



/*------------------------------------------------------------------------------
| IBaseStream::writeVersion                                                    |
|                                                                              |
| Write the version of this stream implementation.                             |
------------------------------------------------------------------------------*/
void
writeVersion( IBaseStream& toWhere, const IBaseStream::VersionNumber version )
{
    version >>= toWhere;
}

/*------------------------------------------------------------------------------
| IBaseStream::readVersion                                                     |
|                                                                              |
| Write the version of the stream.                                             |
------------------------------------------------------------------------------*/
IBaseStream::VersionNumber
readVersion( IBaseStream& fromWhere )
{
    IBaseStream::VersionNumber retval = 0;
    retval <<= fromWhere;
    return retval;
}

/*------------------------------------------------------------------------------
| IBaseStream::operator<<=                                                     |
|                                                                              |
| Stream into an IString.                                                      |
------------------------------------------------------------------------------*/
IBaseStream& operator<<=(IString& c, IBaseStream& s)
{
    unsigned long len;
    s.read(len);
    char * buf = new char [len];
    s.read((void*)buf, (size_t)len);
    //-----------------------------------------------------
    // DEFECT 7526 : must copy FULL buffer, WITH null chars
    //-----------------------------------------------------
    c = IString(buf, len);    //c = buf;
    //-----------------------------------------------------
    delete [] buf;
    return s;
}

/*------------------------------------------------------------------------------
| IBaseStream::operator>>=                                                     |
|                                                                              |
| Write an IString to a stream.                                                |
------------------------------------------------------------------------------*/
IBaseStream& operator>>=(const IString& c, IBaseStream& s)
{
    unsigned long len = c.size() + 1;
    s.write(len);
    s.write((const char *)c, (size_t)len);
    return s;
}

