/*******************************************************************************
* FILE NAME: istring9.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the advanced "<type1>2<type2>"    *
*   IString conversion functions.                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <istring.hpp>

#include <ibuffer.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ISTRING9_CPP_
  #include <ipagetun.h>
#endif

#include "inlsutil.hpp"

/*------------------------------------------------------------------------------
| IString::c2b                                                                 |
|                                                                              |
| If the receiver is null, then the result is a null string. Otherwise, a char |
| array 8 times as long as the receiver is allocated.  This string is filled   |
| with ones and zeros, 8 from each byte of the receiver string.                |
------------------------------------------------------------------------------*/
IString &IString :: c2b ( )
  {
  if ( this->length() != 0 )
    {
    IString     old( this->buffer() );
    const char *pSource = old.data();
    unsigned    n       = old.length();

    this->initBuffer( 0, IBuffer::checkMultiplication( n, 8 ), 0, 0, 0, 0, '0');

    char *pDest = this->data();

    while ( n-- )
      {
      char c = *pSource++;
      unsigned int mask = 1 << 8;
      while ( mask >>= 1 )
         *pDest++ |= ( ( c & mask ) != 0 ); // Mask bit on -> '1'.
      }
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::c2d                                                                 |
|                                                                              |
| If the receiver is null, then it becomes '0'. Otherwise, we convert the      |
| receiver to digits by:                                                       |
|                                                                              |
|   - Starting with the first 4 bytes to an unsigned  long and then            |
|     to a IString using the constructor that accepts an unsigned long         |
|   - "Multiplying" by 256 and adding in the next character for each           |
|      of the rest of the characters in the receiver.                          |
------------------------------------------------------------------------------*/
IString & IString :: c2d ( )
{
#ifdef IC_NLS
  if (IString::isInternationalized()) {
    if ( this->length() != 0 ) {
      IString old( this->buffer() );
      unsigned const char *pSource = (unsigned char*) old.data();

      // Set init to value of first n bytes (n <= 4 ):
      unsigned int  n    = ( old.length() < 4 ) ? old.length() : 4;
      unsigned long init = *pSource++;
      while ( --n )
        init = init * 256 + *pSource++;

      // Initialize receiver with IString equivalent of this value:
      this->initBuffer( init );

      // Now handle additional bytes:
      if ( old.length() > 4 ) {
        n = old.length() - 4;
        // Pad with leading zeros to streamline math:
        unsigned leadzeros = IBuffer::checkMultiplication( n, 3 );
        IString old1( this->buffer() );

        this->initBuffer( 0, leadzeros, old1.data(), old1.length(),
                          0, 0, '0');
        while ( n-- )
          this->decimalMath( *pSource++ );
        // Strip leading zeros:
        if ( ( n = indexThatIsNot( this->data(), this->length(), '0' ) ) > 1 ) {
          IString old2( this->buffer() );

          this->initBuffer( old2.data() + n - 1, old2.length() - n + 1 );
        }
      }
    }
    return *this;

  } else {
#endif
    if ( this->length() != 0 )
      {
      IString old( this->buffer() );
      unsigned const char *pSource = (unsigned char*) old.data();

      // Set init to value of first n bytes (n <= 4 ):
      unsigned int  n    = ( old.length() < 4 ) ? old.length() : 4;
      unsigned long init = *pSource++;
      while ( --n )
        init = init * 256 + *pSource++;

      // Initialize receiver with IString equivalent of this value:
      this->initBuffer( init );

      // Now handle additional bytes:
      if ( old.length() > 4 )
        {
        n = old.length() - 4;
        // Pad with leading zeros to streamline math:
        this->rightJustify( IBuffer::checkAddition
                              ( this->length(),
                                IBuffer::checkMultiplication( n, 3 ) ), '0' );
        while ( n-- )
          this->decimalMath( *pSource++ );
        // Strip leading zeros:
        if ( ( n = this->indexOfAnyBut( "0" ) ) > 1 )
          this->remove( 1, n - 1 );
        }
      }
    return *this;
#ifdef IC_NLS
  }
#endif
}

/*------------------------------------------------------------------------------
| IString::c2x                                                                 |
|                                                                              |
| If the receiver is null, return a null string.  Otherwise, allocate a char   |
| array twice as big as the receiver's.  Then, we fill the array with the      |
| ascii representation of the nibbles of each byte of the receiver string.     |
------------------------------------------------------------------------------*/
static const char hexDigits[] = "0123456789ABCDEF";

IString & IString :: c2x ( )
  {
  if ( this->length() != 0 )
    {
    IString old( this->buffer() );
    const unsigned char *pSource = (unsigned char *) old.data();
    unsigned             n       = old.length();

    this->initBuffer( 0, n, 0, n );

    char* pDest = this->data();

    while ( n-- )
      {
      unsigned char c = *pSource++;
      *pDest++ = hexDigits[ c / 16 ];
      *pDest++ = hexDigits[ c % 16 ];
      }
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::b2c                                                                 |
|                                                                              |
| If the receiver is null, do nothing.  If the receiver is not comprised of    |
| binary digits, reset the receiver to null.                                   |
|                                                                              |
| Otherwise, we process bytes of the receiver, 8 at a time, adding into a      |
| character the corresponding bits.  The resulting character is placed into    |
| the new buffer.                                                              |
------------------------------------------------------------------------------*/
IString & IString :: b2c ( )
  {
  if ( this->length() != 0 && this->isBinaryDigits() )
    {
    IString old( this->buffer() );
    const char  *pSource  = old.data();
    unsigned int newLen   = ( old.length() + 7 ) / 8,
                 bitCount = ( old.length() - 1 ) % 8 + 1;

    this->initBuffer( 0, newLen );

    unsigned char *pDest = (unsigned char *) this->data();

    while ( newLen-- )
      {
      unsigned char c = 0;
      while ( bitCount-- )
        {
        c <<= 1; // Multiply previous bits by 2.
        c += ( *pSource++ == '1' ); // Add: '1'->1,'0'->0
        }
      *pDest++ = c;
      bitCount = 8;
      }
    }
  else
    *this = null;
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::b2d                                                                 |
|                                                                              |
| Convert the receiver to "character" and then to decimal.                     |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::b2x                                                                 |
|                                                                              |
| Convert the receiver to "character" and then to hexidecimal.                 |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString :: d2c                                                               |
|                                                                              |
| If the receiver is not decimal digits, reset to null.                        |
|                                                                              |
| Otherwise, we convert to binary by:                                          |
|                                                                              |
|    - Converting up front as many digits as can fit in a long.                |
|    - Converting additional digits by multiplying by 10 and adding            |
|      in additional digits.                                                   |
------------------------------------------------------------------------------*/
IString &IString :: d2c ( )
{
#ifdef IC_NLS
  if (IString::isInternationalized()) {
    if ( isAllDigits( this->data(), this->length() ) ) {
      if ( this->length() != 0 ) {
        // Set init to value of first n-1 bytes (n == maxLong size):
        unsigned int  n    = ( lengthOf( maxLong ) <= this->length() ) ?
                               (lengthOf( maxLong )-1)
                               :
                               this->length();
        IString tmpStr( this->data(), n );
        unsigned long init = tmpStr.asInt();
        IString old( this->buffer() );

        // Initialize receiver with appropriate bytes from init:
        n = sizeof init;
        this->initBuffer( 0, n );
        unsigned char *p = *this;
        while ( n-- ) {
          p[ n ] = (unsigned char) ( init % 256 );
          init /= 256;
        }

        // Now handle additional bytes:
        if ( old.length() > lengthOf( maxLong ) - 1 ) {
          unsigned const char *pSource = (unsigned char*) old.data() +
                                                   lengthOf( maxLong ) - 1;
          n = old.length() - lengthOf( maxLong ) + 1;
          // Pad with leading zeros to streamline math:
          IString old1( this->buffer() );

          this->initBuffer( 0, n/3, old1.data(), old1.length() );
          while ( n-- )
            this->binaryMath( *pSource++ - '0' );
        }

        // Strip leading zeros:
        unsigned firstNonZero = indexThatIsNot( this->data(), this->length(),
                                                '\0' );
        if (firstNonZero == 0) { // All zeros
          *this = IString('\0');
        } else if (firstNonZero > 1) { // At least one leading zeros
          IString old2( this->buffer() );

          this->initBuffer( old2.data() + firstNonZero - 1,
                            old2.length() - firstNonZero + 1 );
        }
      }
    } else {
      *this = IString::null;
    }
    return *this;

  } else {
#endif
    if ( this->isDigits() )
      {
      if ( this->length() != 0 )
        {
        // Set init to value of first n-1 bytes (n == maxLong size):
        unsigned int  n    = ( lengthOf( maxLong ) <= this->length() ) ?
                               (lengthOf( maxLong )-1)
                               :
                               this->length();
        unsigned long init = this->subString( 1, n ).asInt();
        IString old( this->buffer() );

        // Initialize receiver with appropriate bytes from init:
        n = sizeof init;
        this->initBuffer( 0, n );
        unsigned char *p = *this;
        while ( n-- )
          {
          p[ n ] = (unsigned char) ( init % 256 );
          init /= 256;
          }

        // Now handle additional bytes:
        if ( old.length() > lengthOf( maxLong ) - 1 )
          {
          unsigned const char *pSource = (unsigned char*) old.data() +
                                                    lengthOf( maxLong ) - 1;
          n = old.length() - lengthOf( maxLong ) + 1;
          // Pad with leading zeros to streamline math:
          this->rightJustify( this->length() + n/3, 0 );
          while ( n-- )
            this->binaryMath( *pSource++ - '0' );
          }

        // Strip leading zeros:
        this->stripLeading( '\0' );
        // If null, reset to "0":
        if ( this->length() == 0 )
          *this = IString('\0');
        }
      }
    else
      *this = IString::null;
    return *this;
#ifdef IC_NLS
  }
#endif
}

/*------------------------------------------------------------------------------
| IString::d2b                                                                 |
|                                                                              |
| Convert the receiver to "character" and then to binary.                      |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::d2x                                                                 |
|                                                                              |
| Convert the receiver to "character" and then to hex.                         |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| x2c                                                                          |
|                                                                              |
| Function to transform hex digit to number between 0 and 15:                  |
------------------------------------------------------------------------------*/
static inline unsigned char x2c( unsigned char x )
  {
#if 0
  // This validation is weak -- can create problem on EBCDIC code pages
  return ( x > '9' )
         ? ( x > 'F' )
           ? x - 'a' + 10
           : x - 'A' + 10
         : x - '0';
#else
  return ( '0' <= x && x <= '9' )
         ? x - '0'
         : ( 'a' <= x && x <= 'f' )
           ? x - 'a' + 10
           : x - 'A' + 10;
#endif
  }

/*------------------------------------------------------------------------------
| IString::x2c                                                                 |
|                                                                              |
| If the receiver is null, return a null string.  If the receiver is not       |
| comprised of hex digits, return a null string.                               |
|                                                                              |
| Otherwise, allocate a new buffer one half as big.  Process the source        |
| string two bytes (nibbles) at a time, generating one character per 2         |
| hex digits.                                                                  |
------------------------------------------------------------------------------*/
IString & IString :: x2c ( )
{
  if ( this->length() != 0 ) {
    Boolean allHex;
#ifdef IC_NLS
    if (IString::isInternationalized()) {
      allHex = isAllHexDigits( this->data(), this->length() );
    } else
#endif
      allHex = this->isHexDigits();

    if ( allHex )
      {
      IString old( this->buffer() );
      const unsigned char  *pSource = (unsigned char *) old.data();
      unsigned int          newLen  = ( old.length() + 1 ) / 2;

      // Get first digit, if length is odd, presume leading '0':
      unsigned char c = ( old.length() % 2 ) ? '0' : *pSource++;

      this->initBuffer( 0, newLen );

      unsigned char *pDest = (unsigned char *) this->data();

      while ( newLen-- )
        {
        // Next character is 16 times first digit plus second digit:
        *pDest++ = ::x2c( c ) * 16 + ::x2c( *pSource++ );
        // Get first digit of next character:
        c = *pSource++;
        }
      }
    else
      *this = IString::null;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IString::x2b                                                                 |
|                                                                              |
| Convert the receiver to "character" and then to binary digits.               |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::x2d                                                                 |
|                                                                              |
| Convert the receiver to "character" and then to decimal.                     |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::decimalMath                                                         |
|                                                                              |
| We process the digits of the receiver one at a time from right to left.  We  |
| calculate the result of multiplying this digit by 256 and adding in the      |
| input "carry in" digit (which must be in the range [0, 255].                 |
|                                                                              |
| The result is then divided by 10.   The remainder is used as the result      |
| digit and the quotient is propogated as the "carry" into the calculation of  |
| the next digit to the left.                                                  |
|                                                                              |
| If, after all digits are processed, the carry is non-zero, then we convert   |
| this to decimal digits and prepend it to what was calculated so far.         |
|                                                                              |
| The receiver is prefixed with sufficient leading zeros to ensure there will  |
| be sufficient room for these carry digits.                                   |
------------------------------------------------------------------------------*/
void IString :: decimalMath ( unsigned char newDigit )
  {
  IString old( this->buffer() );
  const unsigned char *pSource = (unsigned char *) old.data() +
                                                   old.length() - 1;

  // Add enough leading zeros to handle carry out of high order digit:

  unsigned int n;
  // First non-0 digit.
#ifdef IC_NLS
  if (IString::isInternationalized()) {
    n = indexThatIsNot( this->data(), this->length(), '0' );
  } else
#endif
    n = this->indexOfAnyBut( "0" );

  if ( n == 0 )
    n = old.length(); // All digits are zeros.
  else
    n--;            // n-1 are zeros.
  // Never remove leading zeros.
  if ( n > 3 )
    n = 3;
  this->initBuffer( 0, old.length(),
                    0, 3 - n,
                    0, 0,
                    '0' );                   // At least 3 leading 0s.
  unsigned char *pDest = (unsigned char *) this->data() + this->length() - 1;

  n = old.length();
  unsigned int carry = newDigit;
  while ( n-- )
    {
    unsigned int digit = *(pSource--) - (unsigned char) '0';
    digit = 256u * digit + carry;
    carry = digit / 10u;
    *pDest-- = '0' + (unsigned char) ( digit % 10u );
    }

  while ( carry )
    {
    *pDest-- = '0' + (unsigned char) ( carry % 10u );
    carry = carry / 10u;
    }
  }

/*------------------------------------------------------------------------------
| IString::binaryMath                                                          |
|                                                                              |
| We process the bytes of the receiver one at a time from right to left.  We   |
| calculate the result of multiplying each byte by 10 and adding in the input  |
| "carry in" digit (which must be in the range [0, 9]).                        |
|                                                                              |
| The result is then divided by 256.   The remainder is used as the result     |
| byte and the quotient is propogated as the "carry" into the calculation of   |
| the next byte to the left.                                                   |
|                                                                              |
| If, after all bytes are processed, the carry is non-zero, then this value is |
| used to set the leading byte.  The receiver is prefixed with a zero byte to  |
| ensure there is room for this.                                               |
------------------------------------------------------------------------------*/
void IString :: binaryMath ( unsigned char newDigit )
  {
  IString old( this->buffer() );
  const unsigned char *pSource = (unsigned char *) old.data() +
                                                   old.length() - 1;

  // Add enough leading zeros to handle carry out of high order byte:

  unsigned int n;
  // First non-0 byte.
#ifdef IC_NLS
  if (IString::isInternationalized()) {
    n = indexThatIsNot( this->data(), this->length(), '\0' );
  } else
#endif
    n = this->indexOfAnyBut( "\0" );

  n = ( n == 1 );  // Add a zero only if first digit is non-zero.
  this->initBuffer( 0, old.length(), 0, n );
  unsigned char *pDest = (unsigned char *) this->data() + this->length() - 1;

  n = old.length();
  unsigned int carry = newDigit;
  while ( n-- )
    {
    unsigned int digit = *pSource--;
    digit = 10u * digit + carry;
    carry = digit / 256u;
    *pDest-- = (unsigned char) ( digit % 256u );
    }

  if ( carry )
    *pDest = carry;
  }
