/*******************************************************************************
* FILE NAME: icmpfram.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IComponentFrameWindow                                                    *
*     IAswFrameWindow                                                          *
*     IAswFrameSpace                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "icmpfram.hpp"

// ICLUI Headers
#include <imenubar.hpp>
#include <itbarfrm.hpp>

// Document Framework Headers
#include <iguibndl.hpp>
#include <istatnry.hpp>
#include <iecmpimp.hpp>
#include <iembmod.hpp>


/*******************************************************************************
*                                                                              *
*   Class IComponentFrameWindow                                                *
*                                                                              *
*   Main frame window for the component                                        *
*                                                                              *
*******************************************************************************/

// -----------------------------------------------------------------------------
// Constuctor
// -----------------------------------------------------------------------------
IComponentFrameWindow::IComponentFrameWindow() :
    IFrameWindow( (const char*)IBase::messageText( IC_CDF_UNTITLED_NAME ) ),
    fpAswFrameWindow( NULL ),
    fpMenuBase( NULL ),
    fpFrameSpace( NULL ),
    fpActivePresSite( NULL ),
    fpMenuBar( NULL ),
    fpToolBar( NULL ),
    fpAccel( NULL ),
    fDummy( 0, this, this ),
    fbDeactivateFlag( false ),
    fbDelayedAdjustment( false )
{
}

// -----------------------------------------------------------------------------
// setup
// -----------------------------------------------------------------------------
// if a pointer to the Aswan control is supplied then this is a container, and
// we should instantiate and setup an Aswan FrameWindow object
CAswResult
IComponentFrameWindow::setup( CAswControl* pAswControl )
{
    CAswResult ar( S_OK, ASD_OK );

    // if this is a container (aswan control was supplied as a parameter)
    // then construct an Aswan frame window object
    if( pAswControl )
    {
        fpAswFrameWindow = new IAswFrameWindow( *this );
        ASSERT( fpAswFrameWindow != NULL );

        fpMenuBase = new CAswContMenu( pAswControl );
        ar = fpMenuBase->Setup();
        IASWASSERTSTATE( ar );

        fpFrameSpace = new IAswFrameSpace( *this );
        ar = fpFrameSpace->Setup();
        IASWASSERTSTATE( ar );

        // ASWAN requires the frame window to have the WS_CLIPCHILDREN style
        HWND hWnd = (HWND)(void*)handle();
        DWORD dwStyle = ::GetWindowLong( hWnd, GWL_STYLE );
        ::SetWindowLong( hWnd, GWL_STYLE, dwStyle | WS_CLIPCHILDREN );

        ar = fpAswFrameWindow->Setup( hWnd, NULL, FALSE, fpMenuBase, fpFrameSpace, NULL );
        IASWASSERTSTATE( ar );

    }

    return ar;
}


// -----------------------------------------------------------------------------
// Destructor
// -----------------------------------------------------------------------------
IComponentFrameWindow::~IComponentFrameWindow()
{

    // delete menu bar
    delete fpMenuBar;

    // delete tool bar
    delete fpToolBar;

    // delete accelerator
    delete fpAccel;

    // shutdown and delete frame space
    if( fpFrameSpace )
    {
        fpFrameSpace->Shutdown( TRUE );
        delete fpFrameSpace;
    }

    // shutdown and delete menu base
    if( fpMenuBase )
    {
        fpMenuBase->Shutdown( TRUE );
        delete fpMenuBase;
    }

    // shutdown and delete aswan frame window
    if( fpAswFrameWindow )
    {
        fpAswFrameWindow->SetSpaceHandler( NULL ); // aswan bug
        fpAswFrameWindow->Shutdown( TRUE );
        delete fpAswFrameWindow;
    }

}


// -----------------------------------------------------------------------------
// addDefaultMenuBar
// -----------------------------------------------------------------------------
void
IComponentFrameWindow::addDefaultMenuBar( Boolean bForceStandalone )
{
    // create the menu bar

    if( IComponentStationery::stationery().isEmbedded() && !bForceStandalone )
    {
        // -------------------------------
        // We are embedded
        // -------------------------------
        fpMenuBar = new IMenuBar( kEmbeddedMenu, this );

        // delete the old accelerator before creating a new one
        delete fpAccel;
        fpAccel = new IAccelerator( kEmbeddedMenu, this );
    }

    else
    {
        // -------------------------------
        // Non embedding menu
        // -------------------------------
        fpMenuBar = new IMenuBar( kStandaloneMenu, this );

        // delete the old accelerator before creating a new one
        delete fpAccel;
        fpAccel = new IAccelerator( kStandaloneMenu, this );

    }

    assert( fpMenuBar != NULL );
    try {
        fpMenuBar->disableItem( kEditUndo );
    }
    catch( IException& )
    {
    }

    assert( fpAccel != NULL );

    // plug the menu options into the Aswan container menu if we are a container
    if( IComponentStationery::stationery().isContainer() )
    {
        // setup the three groups
        // TODO: we are only implementing the "FILE" option - must implement CONTAINER and WINDOW
        fSharedMenu[0].nPosition = 0;
        fSharedMenu[0].nGroup = ASW_MENUGROUP_FILE;
        fpMenuBase->SetMenuInfo( (HMENU)(void*)fpMenuBar->menuHandle(), 1, fSharedMenu );
        fpMenuBase->SetRegularMenu( (HMENU)(void*)fpMenuBar->menuHandle() );

    } // we are a container

//    fpMenuBar->checkItem( kViewToolbar );

    // add dummy frame extension for in-place active object's toolbars
    addExtension( &fDummy,
                  IFrameWindow::aboveClient,
                  0,
                  IFrameWindow::none );

    // hide the dummy window or it will obscure the toolbars
    fDummy.hide();

    // add default tool bar
    fpToolBar = new IComponentToolBar( *this );
//    showToolBar( true );

}

// -----------------------------------------------------------------------------
// changeNegotiatedSpace
// -----------------------------------------------------------------------------
void
IComponentFrameWindow::changedNegotiatedSpace()
{
    // if we're a container, tell in-place active servers about window resize
    
    CAswResult ar(S_FALSE, ASD_OK);
    if ( fpFrameSpace )
    {
        ar = fpFrameSpace->ChangedNegotiatedSpace();
    }

    if ( ar.Scode() == S_FALSE )
        showToolBar( fpMenuBar->isItemChecked( kViewToolbar ) );

}

// -----------------------------------------------------------------------------
// delayAdjustment
// -----------------------------------------------------------------------------
void
IComponentFrameWindow::delayAdjustment()
{
    fbDelayedAdjustment = true;
}

// -----------------------------------------------------------------------------


// adjustToolBar 
// -----------------------------------------------------------------------------
void
IComponentFrameWindow::adjustToolBar()
{
    if ( fbDeactivateFlag )
    {
        // put the toolbar back if the user wants to see it
        showToolBar( fpMenuBar->isItemChecked( kViewToolbar ) );

        fbDeactivateFlag = false;
    }
    
    fbDelayedAdjustment = false;
    
}

// -----------------------------------------------------------------------------
// showToolBar 
// -----------------------------------------------------------------------------
void
IComponentFrameWindow::showToolBar( Boolean bShow )
{
    if ( bShow )
    {
        setExtensionSize( &fDummy, (unsigned long)fpToolBar->size().height() );
        fpToolBar->floatingFrame()->moveSizeTo( fDummy.rect() );
        fpToolBar->floatingFrame()->show();

        fpMenuBar->checkItem( kViewToolbar, true );
    }
    else
    {
        setExtensionSize( &fDummy, 0 );
        fpToolBar->floatingFrame()->hide();

        fpMenuBar->checkItem( kViewToolbar, false );
    }
        
}


/*******************************************************************************
*                                                                              *
*   Class IComponentToolBar                                                    *
*                                                                              *
*******************************************************************************/

IComponentToolBar::IComponentToolBar( IFrameWindow& frame ) :
    IToolBar( kToolBar, &frame, IToolBar::floating, false,
              IToolBar::defaultStyle() | IToolBar::noDragDrop ),
    fButtonCut( kEditCut, this, this ),
    fButtonCopy( kEditCopy, this, this ),
    fButtonPaste( kEditPaste, this, this ),
    fButtonAbout( kHelpAbout, this, this ),
    fpButtonNew( NULL ),
    fpButtonOpen( NULL ),
    fpButtonSave( NULL )
{
//    floatingFrame()->hide();
//    floatingFrame()->setParent( &frame );

    // get rid of the caption, border, etc
    HWND hWnd = (HWND)(void*)floatingFrame()->handle();
    long lStyle = ::GetWindowLong( hWnd, GWL_STYLE );
    ::SetWindowLong( hWnd, GWL_STYLE, lStyle & ~WS_CAPTION );

    setBackgroundColor( IColor::paleGray );

    // add the buttons
    if ( ! IComponentStationery::stationery().isEmbedded() )
    {
        fpButtonNew = new IToolBarButton( kFileNew, this, this );
        fpButtonOpen = new IToolBarButton( kFileOpen, this, this );
        fpButtonSave = new IToolBarButton( kFileSave, this, this );

        // new group
        addAsLast( fpButtonNew, true );
        addAsLast( fpButtonOpen );
        addAsLast( fpButtonSave );
    }

    // new group
    addAsLast( &fButtonCut, true );
    addAsLast( &fButtonCopy );
    addAsLast( &fButtonPaste );

    // new group
    addAsLast( &fButtonAbout, true );

}

IComponentToolBar::~IComponentToolBar()
{
    delete fpButtonNew;
    delete fpButtonOpen;
    delete fpButtonSave;
}

IToolBarFrameWindow*
IComponentToolBar::createFloatingFrame ( IFrameWindow* owner )
{
    // call the base class to create the floating frame
    IToolBarFrameWindow* floatingFrame = IToolBar::createFloatingFrame( owner );
    floatingFrame->setParent( owner );

    return floatingFrame;

}

/*******************************************************************************
*                                                                              *
*   Class IAswFrameWindow                                                      *
*                                                                              *
*   needed to override some methods                                            *
*                                                                              *
*******************************************************************************/

IAswFrameWindow::IAswFrameWindow( IComponentFrameWindow& frame ) :
    CAswFrameWindow( IComponentStationery::stationery() ),
    fpComponentFrameWindow( &frame )
{
}


// -----------------------------------------------------------------------------
// OnInPlaceActivatePassThru
// -----------------------------------------------------------------------------
HRESULT
IAswFrameWindow::OnInPlaceActivatePassThru( PASWINPLACEPRESSITE pIPPresSite,
                                            PASWREFHOLDER       pRefHoldOleInPlaceObj )
{
    fpComponentFrameWindow->fpActivePresSite = pIPPresSite;

    // turn on flag in the embedded object's presentation site that
    // indicates that the object is in place active. This will prevent the
    // grab handles from displaying
    ((IAswPresSite*)pIPPresSite)->setIsInPlace( true );

    // we must call the base class to perform important initialization
    return CAswFrameWindow::OnInPlaceActivatePassThru( pIPPresSite, pRefHoldOleInPlaceObj );
}


// -----------------------------------------------------------------------------
// OnInPlaceDeactivatePassThru
// -----------------------------------------------------------------------------
void
IAswFrameWindow::OnInPlaceDeactivatePassThru( PASWINPLACEPRESSITE pIPPresSite )
{
    fpComponentFrameWindow->fpActivePresSite = NULL;

    // tell the embedded object's pres site that the object is no longer in place active
    ((IAswPresSite*)pIPPresSite)->setIsInPlace( false );

    // call the base class to clean up
    CAswFrameWindow::OnInPlaceDeactivatePassThru( pIPPresSite );

    fpComponentFrameWindow->fbDeactivateFlag = true;

    // remove server tools unless we were told to delay
    if ( ! fpComponentFrameWindow->fbDelayedAdjustment )
        fpComponentFrameWindow->adjustToolBar();

}


/*******************************************************************************
*                                                                              *
*   Class IAswFrameSpace                                                       *
*                                                                              *
*   handles frame space negotiation for toolbars                               *
*                                                                              *
*******************************************************************************/

// -----------------------------------------------------------------------------
// Constuctor
// -----------------------------------------------------------------------------
IAswFrameSpace::IAswFrameSpace( IComponentFrameWindow& frame ) :
    CAswSpaceBase( IComponentStationery::stationery() ),
    fpComponentFrameWindow( &frame )
{
}

// -----------------------------------------------------------------------------
// Destructor
// -----------------------------------------------------------------------------
IAswFrameSpace::~IAswFrameSpace()
{
    Shutdown( TRUE );
}

// -----------------------------------------------------------------------------
// GetBorder
// -----------------------------------------------------------------------------
// Returns a RECT structure in which the object can put toolbars while
// active InPlace.
CAswResult
IAswFrameSpace::GetBorder( LPRECT prcBorder )
{
    // the base class implementation returns the client window
    CAswResult ar = CAswSpaceBase::GetBorder( prcBorder );

//    if( fpComponentFrameWindow->toolBar().location() == IToolBar::aboveClient )
//        prcBorder->bottom += fpComponentFrameWindow->toolBar().size().height();
//    else
    prcBorder->bottom += fpComponentFrameWindow->dummy().size().height();

    return ar;

}

// -----------------------------------------------------------------------------
// RequestBorderSpace
// -----------------------------------------------------------------------------
// Checks if tools can be installed around the  window frame while the
// object is active InPlace.
CAswResult
IAswFrameSpace::RequestBorderSpace( LPCBORDERWIDTHS lpBorderwidths )
{
    return CAswSpaceBase::RequestBorderSpace( lpBorderwidths );
}

// -----------------------------------------------------------------------------
// SetBorderSpace
// -----------------------------------------------------------------------------
// Allocates space for the border requested in the call to the
// RequestBorderSpace method.
CAswResult
IAswFrameSpace::SetBorderSpace( LPCBORDERWIDTHS lpBorderwidths )
{

    // call the base class to see if it's OK
    CAswResult ar = CAswSpaceBase::SetBorderSpace( lpBorderwidths );
    if( ASWSUCCEEDED( ar ) && lpBorderwidths )
    {
        // turn off our toolbar

        // is the toolbar there?
        if( fpComponentFrameWindow->menuBar().isItemChecked( kViewToolbar ) )
        {
            fpComponentFrameWindow->toolBar().floatingFrame()->hide();
        }

        // request the frame space from the frame window
        if( lpBorderwidths->top != fpComponentFrameWindow->dummy().size().height() )
        {
            fpComponentFrameWindow->setExtensionSize( &fpComponentFrameWindow->dummy(),
                                                      (unsigned long)lpBorderwidths->top );
        }

        // tell in-place pres site about the client area size change
        assert( fpComponentFrameWindow->fpActivePresSite );
        RECT border;
        GetBorder( &border );
        fpComponentFrameWindow->fpActivePresSite->UpdateClipRect( &border );

    }

    return ar;

}


