/*******************************************************************************
* FILE NAME: idocstor.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IDocumentStorage                                                         *
*     IStructuredStorage                                                       *
*     IFlatFileStorage                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Lotus.       *
*   (C) Copyright Lotus Development Corporation. 1996 - All Rights Reserved.   *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "idocstor.hpp"

#include <imetatyp.hpp>
#include <istatnry.hpp>
#include <icompnen.hpp>
#include <iembmod.hpp>
#include <iecmpimp.hpp>
#include <iolestrm.hpp>
#include <ifilstrm.hpp>
#include <icompimp.hpp>
#include <inotifev.hpp>


////////////////////////////////////////////////////////////////////////////////
// IDocumentStorage methods
////////////////////////////////////////////////////////////////////////////////

IDocumentStorage::IDocumentStorage( IComponent& comp )
: fpComponent( &comp ), fpStorage( NULL )
{
    setFileName( "" );
}


IDocumentStorage::~IDocumentStorage()
{
}


void
IDocumentStorage::setFileName( const IString& name )
{
    // don't do anything if the file is not different
    if ( name == fFileName && name.length() != 0 )
        return;

    // set the file name
    fFileName = name;
    
    // set the document name as well
    component().setDocumentName( name.subString( name.lastIndexOf('\\')+1 ) );

    // Tell OLE that our name has changed
    IString monikerName = fFileName;
    if( name.length() == 0 )
        // changing to "untitled"
        monikerName = component().uniqueDocumentName();

    component().model()->fileRenamed( monikerName );

}


void
IDocumentStorage::setStorage( CAswStorage* stor )
{
    fpStorage = stor;
}


void
IDocumentStorage::loadStorage( CAswStorage* stor, CAswStream* strm )
{
    CAswStream  tempStream( IComponentStationery::stationery() );
    CAswStream* theStream;

    if ( strm == NULL )
    {
        CAswResult result = tempStream.Setup();
        if ( ASWFAILED(result) )
            throw IAswException( result );
        
        // Open the stream...
        result = stor->OpenChildStream( &tempStream,
                                        OLESTR("CONTENTS"),
                                        STGM_DIRECT | STGM_READWRITE |
                                        STGM_SHARE_EXCLUSIVE );
        if ( ASWFAILED(result) )
            throw IAswException( result );

        theStream = &tempStream;
    }
    else
        theStream = strm;

    // try to read our object's data
    IOLEStream stream( *theStream );
    IModel* newModel = NULL;
    ::resurrect( newModel, stream );

    // if successful, adopt the new model
    if ( newModel )
        component().adoptModel( newModel );

    if ( strm == NULL )
        tempStream.Shutdown();

}


void
IDocumentStorage::saveStorage( CAswStorage* stor,
                               CAswStream* strm,
                               Boolean sameAsLoad,
                               Boolean remember ) const
{
    CAswStream  tempStream( IComponentStationery::stationery() );
    CAswStream* theStream;

    if ( strm == NULL )
    {
        CAswResult result = tempStream.Setup();
        if ( ASWFAILED(result) )
            throw IAswException( result );
        
        // Create a new stream...
        result = stor->CreateChildStream( &tempStream,
                                          OLESTR("CONTENTS"),
                                          STGM_CREATE | STGM_DIRECT |
                                          STGM_WRITE  | STGM_SHARE_EXCLUSIVE );
        if ( ASWFAILED(result) )
            throw IAswException( result );

        theStream = &tempStream;
    }
    else
        theStream = strm;

    // write our data to the stream
    IOLEStream stream( *theStream );
    ::flatten(fpComponent->model(), stream);

    if ( strm == NULL ) {
        tempStream.Shutdown();
    }

}


void
IDocumentStorage::resetFile()
{
    component().adoptModel( IComponentStationery::stationery().createModel() );
    
    setFileName( "" );
    
    component().setInitialSize(); // reset extents
    component().objectInstance()->SetIsDirty( FALSE );
    component().notifier().notifyObservers(
        INotificationEvent( IComponent::ContentsReplacedId, component().notifier() ) );
    
}


////////////////////////////////////////////////////////////////////////////////
// IFlatFileStorage methods
////////////////////////////////////////////////////////////////////////////////

IFlatFileStorage::IFlatFileStorage( IComponent& comp )
: IDocumentStorage( comp )
{
}


IFlatFileStorage::~IFlatFileStorage()
{
}


void
IFlatFileStorage::loadFile(const IString& name)
{
    IModel* newModel = NULL;
    
    // orphan and delete the old model
    delete component().orphanModel();

    if ( name.length() == 0 )
    {
        resetFile();
    }
    else
    {
        try {
            // try to read our object's data
            IFileStream stream( name );
            ::resurrect( newModel, stream );
        }
        catch ( IException& e ) {
            // load failed, adopt an empty model
            resetFile();
            // rethrow the exception
            throw e;
        }
    
        // adopt the new model
        component().adoptModel( newModel );
        
        setFileName( name );
        
        component().objectInstance()->SetIsDirty( FALSE );
        component().notifier().notifyObservers(
            INotificationEvent( IComponent::ContentsReplacedId, component().notifier() ) );
        
    }
    
}


void
IFlatFileStorage::saveFile(const IString& name,
                           Boolean sameAsLoad,
                           Boolean remember)
{
    // write our data to the stream
    IFileStream stream( name );
    ::flatten( component().model(), stream );
    stream.setLogicalEndOfStream( stream.position() );
    
    if ( sameAsLoad || remember )    // "Save" or "Save as"
    { 
        component().objectInstance()->SetIsDirty( FALSE );
        if ( ! sameAsLoad )          // "Save as"
        {                    
            setFileName( name );
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
// IStructuredStorage methods
////////////////////////////////////////////////////////////////////////////////

IStructuredStorage::IStructuredStorage( IComponent& comp,
                                        IEmbedderModel* embModel )
    : IDocumentStorage( comp ),
      fpDocFile( NULL ),
      fpRootStorage( NULL )
{
    // check if we are embedded or stand-alone
    if ( ! comp.isEmbedded() )
    {
        // we are not embedded, so create a doc file storage
        fpDocFile = new CAswDocFileStorage( IComponentStationery::stationery() );
        assert( fpDocFile );
        CAswResult result = fpDocFile->Setup();
        IASWASSERTSTATE( result );

        // now we create a CAswStorage to act as the root storage
        fpRootStorage = new CAswStorage( IComponentStationery::stationery() );
        result = fpRootStorage->Setup();
        IASWASSERTSTATE( result );

        // if we are an embedder, pass the root file storage to our container
        if ( embModel != NULL)
        {
            // we know that rootContainer() is not NULL because we are
            // not embedded (i.e. ! comp.isEmbedded() above)
            embModel->rootContainer()->SetRootStorage( fpDocFile );
        }

        //FB Create an unnamed document
        DOCFILESTGCREATE  DFSC;
        DFSC.lpstrName  = NULL;
        DFSC.dwgrfMode  = STGM_DIRECT | STGM_READWRITE | STGM_CREATE |
                          STGM_SHARE_EXCLUSIVE;
        DFSC.dwReserved = 0UL;
        result = fpDocFile->Create(fpRootStorage, &DFSC);
        IASWASSERTSTATE( result );

        // set the root storage to be the one to use
        IDocumentStorage::setStorage( fpRootStorage );
    }

}


IStructuredStorage::~IStructuredStorage()
{
    if (fpRootStorage) {
        fpRootStorage->Shutdown( FALSE );
        delete fpRootStorage;
    }
    
    if (fpDocFile) {
        fpDocFile->Shutdown( FALSE );
        delete fpDocFile;
    }
}


void
IStructuredStorage::setStorage( CAswStorage* stor )
{
    // we must be running stand-alone, i.e. no root storage
    assert( fpRootStorage == NULL ); 
    IDocumentStorage::setStorage( stor );
}


// -----------------------------------------------------------------------------
// resetFile()
// -----------------------------------------------------------------------------
// Open a new temp file and reset the model
void
IStructuredStorage::resetFile()
{
    assert( ! component().isEmbedded() && fpRootStorage && fpDocFile );

    // shutdown the old storage and doc file storage
    fpRootStorage->Shutdown();
    fpDocFile->Shutdown();
    
    // and set up again
    CAswResult result = fpDocFile->Setup();
    IASWASSERTSTATE( result );
    result = fpRootStorage->Setup();
    IASWASSERTSTATE( result );

    // Create a new unnamed doc file
    //FB TODO: figure out exactly what to set here
    DOCFILESTGCREATE  DFSC;
    DFSC.lpstrName  = NULL;
    DFSC.dwgrfMode  = STGM_DIRECT | STGM_READWRITE | STGM_CREATE |
                      STGM_SHARE_EXCLUSIVE;
    DFSC.dwReserved = 0UL;
    result = fpDocFile->Create(fpRootStorage, &DFSC);
    IASWASSERTSTATE( result );

    // call the base class to reset the model and file name
    IDocumentStorage::resetFile();
    
}

void
IStructuredStorage::loadFile(const IString& name)
{
    // orphan and delete the old model
    delete component().orphanModel();

    if ( name.length() == 0 )
    {
        // if file-new open a temp file and reset the model
        resetFile();
        return;
    }

    CAswResult          result(S_OK, ASD_OK);
    CAswOLEString       lpszOLEFile( name );

    CAswDocFileStorage* tDocFile        = fpDocFile;
    CAswStorage*        tRootStorage    = storage();
    CAswDocFileStorage* delDocFile      = NULL;
    CAswStorage*        delRootStorage  = NULL;

    Boolean             differentFile   = name != fileName() || name.length() == 0;
    
    do {
        if ( differentFile )
        {
            delDocFile = tDocFile = new CAswDocFileStorage( IComponentStationery::stationery() );
            result = tDocFile->Setup();
            if ( ASWFAILED(result) ) break;

            delRootStorage = tRootStorage = new CAswStorage( IComponentStationery::stationery() );
            result = tRootStorage->Setup();
            if ( ASWFAILED(result) ) break;
        }

        //FB TODO: figure out exactly what to set here
        DOCFILESTGOPEN    DFSO;
        DFSO.pAswStgPriority = NULL;
        DFSO.snbExclude = NULL;
        DFSO.dwReserved = NULL;
        DFSO.lpstrName = lpszOLEFile.GetOLEPtr();
        DFSO.dwgrfMode = STGM_TRANSACTED | STGM_READWRITE;

        result = tDocFile->Open(tRootStorage, &DFSO);
        if ( ASWFAILED(result) ) break;

        try {
            loadStorage( tRootStorage, NULL );
        }
        catch ( IException& e ) {
            // load failed, adopt an empty model
            resetFile();
            throw e;
        }

    } while ( FALSE );

    if ( differentFile )
    {
        if ( ASWSUCCEEDED(result) )
        {
            // switch to new doc file and storage
            delRootStorage = fpRootStorage;
            fpRootStorage = tRootStorage;
            IDocumentStorage::setStorage( fpRootStorage );

            delDocFile = fpDocFile;
            fpDocFile = tDocFile;

            IEmbedderModel* embModel;
            ::dynamicCastTo( embModel, component().model() );

            // if we are an embedder, pass the new root file storage to our container
            if ( embModel != NULL && ! component().isEmbedded() )
            {
                embModel->rootContainer()->SetRootStorage( fpDocFile );
            }
        }
             
        if ( delRootStorage ) {
            delRootStorage->Shutdown();
            delete delRootStorage;
        }
        if ( delDocFile ) {
            delDocFile->Shutdown();
            delete delDocFile;
        }

    }

    if ( ASWFAILED(result) )
    {
        resetFile();
        throw IAswException( result );
    }
    
    setFileName( name );
    component().objectInstance()->SetIsDirty( FALSE );
    component().notifier().notifyObservers(
        INotificationEvent( IComponent::ContentsReplacedId, component().notifier() ) );

}


void
IStructuredStorage::loadStorage( CAswStorage* stor, CAswStream* strm )
{
    // call base class
    IDocumentStorage::loadStorage( stor, strm );

    loadEmbeddedComponents( stor );
    
}


void
IStructuredStorage::loadEmbeddedComponents( CAswStorage* fromStorage )
{
    IEmbedderModel* embModel;
    ::dynamicCastTo( embModel, component().model() );

    if ( embModel != NULL) {
        CAswResult  result(S_OK, ASD_OK);

        const IEmbedderModel::ComponentList&
            compList = embModel->componentList();

        ICursor& cursor = *compList.newCursor();
    
        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext()) {
            IEmbeddedComponent* embeddedComp = compList.elementAt(cursor);

            embeddedComp->preinitialize( component().bundle(), fromStorage );
            embeddedComp->initialize( false, false );

        }

        delete &cursor;

        if ( ASWFAILED(result) )
            throw IAswException( result );

    }

}


void
IStructuredStorage::saveFile(const IString& name,
                             Boolean sameAsLoad,
                             Boolean remember)
{
    CAswResult          result(S_OK, ASD_OK);
    CAswOLEString       lpszOLEFile( name );
    
    CAswDocFileStorage* tDocFile        = fpDocFile;
    CAswStorage*        tRootStorage    = storage();
    CAswDocFileStorage* delDocFile      = NULL;
    CAswStorage*        delRootStorage  = NULL;

    do {
        if ( ! sameAsLoad )
        {
            delDocFile = tDocFile = new CAswDocFileStorage( IComponentStationery::stationery() );
            result = tDocFile->Setup();
            if ( ASWFAILED(result) ) break;

            delRootStorage = tRootStorage = new CAswStorage( IComponentStationery::stationery() );
            result = tRootStorage->Setup();
            if ( ASWFAILED(result) ) break;

            DOCFILESTGCREATE  DFSC;
            DFSC.lpstrName  = lpszOLEFile.GetOLEPtr();
            DFSC.dwgrfMode  = STGM_TRANSACTED | STGM_READWRITE | STGM_CREATE |
                              STGM_SHARE_EXCLUSIVE;
            DFSC.dwReserved = 0UL;
            result = tDocFile->Create(tRootStorage, &DFSC);
            if ( ASWFAILED(result) ) break;
        }

        try {
            saveStorage( tRootStorage, NULL, sameAsLoad, remember );
        }
        catch ( IException& ) {
            result.Set(STG_E_WRITEFAULT, ASD_UNCHANGED);
        }
//FB should we catch IAswException also??
        if ( ASWFAILED(result) ) break;

        if ( remember ) {
            delRootStorage = fpRootStorage;
            fpRootStorage = tRootStorage;
            IDocumentStorage::setStorage( fpRootStorage );

            delDocFile = fpDocFile;
            fpDocFile = tDocFile;

            IEmbedderModel* embModel;
            ::dynamicCastTo( embModel, component().model() );

            // if we are an embedder, pass the new root file storage to our container
            if ( embModel != NULL && ! component().isEmbedded() )
            {
                embModel->rootContainer()->SetRootStorage( fpDocFile );
            }
        }

    } while ( FALSE );


    if ( ASWSUCCEEDED(result) )
        tRootStorage->Commit();

    // shutdown and delete the ones we're not keeping
    if ( delRootStorage ) {
        delRootStorage->Shutdown();
        delete delRootStorage;
    }
    if ( delDocFile ) {
        delDocFile->Shutdown();
        delete delDocFile;
    }

    if ( ASWFAILED(result) )
        throw IAswException( result );

    if ( sameAsLoad || remember )    // "Save" or "Save as"
    { 
        component().objectInstance()->SetIsDirty( FALSE );
        if ( ! sameAsLoad )          // "Save as"
        {                    
            setFileName( name );
        }
    }
    
}


void
IStructuredStorage::saveStorage( CAswStorage* stor,
                                 CAswStream* strm,
                                 Boolean sameAsLoad,
                                 Boolean remember ) const
{
    // call base class
    IDocumentStorage::saveStorage( stor, strm, sameAsLoad, remember );

    saveEmbeddedComponents( stor, sameAsLoad, remember );

}


void
IStructuredStorage::saveEmbeddedComponents( CAswStorage* toStorage,
                                            Boolean sameAsLoad,
                                            Boolean remember) const
{
    IEmbedderModel* embModel;
    ::dynamicCastTo( embModel, component().model() );

    if ( embModel != NULL) {
        CAswResult  result(S_OK, ASD_OK);

        const IEmbedderModel::ComponentList&
            compList = embModel->componentList();

        ICursor& cursor = *compList.newCursor();
    
        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext()) {
            IEmbeddedComponent* embeddedComp = compList.elementAt(cursor);
        
            if ( sameAsLoad )
                result = embeddedComp->containedObject()->SaveToCurrent();
            else if ( remember )
                result = embeddedComp->containedObject()->SaveToNewStorage( toStorage );
            else
                result = embeddedComp->containedObject()->SaveToDifferentStorage( toStorage );

            if ( ASWFAILED(result) ) break;
        }

        delete &cursor;
    
        if ( ASWFAILED(result) )
            throw IAswException( result );

    }

}


