/*******************************************************************************
* FILE NAME: iguibndl.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IGUIBundle                                                               *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Taligent.    *
*   (C) Copyright Taligent, Inc. 1996 - All rights reserved.                   *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "iguibndl.hpp"

#include <icmpfram.hpp>
#include <icompnen.hpp>
#include <icompimp.hpp>
#include <iecmpimp.hpp>
#include <iembmod.hpp>
#include <iview.hpp>
#include <istatnry.hpp>
#include <idataex.hpp>

// ICLUI classes
#include <ifiledlg.hpp>
#include <imenubar.hpp>
#include <itbar.hpp>
#include <itbarfrm.hpp>
#include <imnitem.hpp>
#include <imenuhdr.hpp>
#include <isubmenu.hpp>
#include <ipopmenu.hpp>
#include <isysmenu.hpp>
#include <icoordsy.hpp>

/*******************************************************************************
*                                                                              *
*   Class IMainMenuHandler                                                     *
*                                                                              *
*******************************************************************************/

class IMainMenuHandler : public IMenuHandler
{
public:
                            IMainMenuHandler( IGUIBundle& bundle );
    virtual                 ~IMainMenuHandler();

    virtual Boolean         makePopUpMenu ( IMenuEvent& menuEvent );
    virtual Boolean         menuShowing   ( IMenuEvent& menuEvent,
                                            ISubmenu&   submenu );

protected:
    IGUIBundle&             fBundle;

    virtual Boolean         dispatchHandlerEvent ( IEvent& event );

private:
    IString                 fMenuTextCut;
    IString                 fMenuTextCopy;
    IString                 fMenuTextPaste;
    IString                 fMenuTextPasteSpecial;
    IString                 fMenuTextPasteLink;
    IString                 fMenuTextAll;
    IString                 fMenuTextReturn;
    IString                 fMenuTextDelete;
    IString                 fMenuTextCtrlX;
    IString                 fMenuTextCtrlC;
    IString                 fMenuTextCtrlV;

    HMENU                   fhPopupMenu;   // current popup menu
    HMENU                   fhEditMenu;

    // Helper functions
    IString                 getStringFromResourceFile( unsigned long stringId, IString strDefault );
    Boolean                 itemExists( ISubmenu& menu, unsigned long id );
};


// the text instance variables are used to cache values from the resource file
// they MUST have a value, so we initialize them with hardcoded English text 
// if there is no corresponding value in the user-supplied resource file
IMainMenuHandler::IMainMenuHandler( IGUIBundle& bundle ) :
    fBundle( bundle ),
    fMenuTextCut( getStringFromResourceFile( IC_CDF_MENU_TEXT_CUT, "Cut" ) + " " ),
    fMenuTextCopy( getStringFromResourceFile( IC_CDF_MENU_TEXT_COPY, "Copy" ) + " " ),
    fMenuTextPaste( getStringFromResourceFile( IC_CDF_MENU_TEXT_PASTE, "Paste" ) ),
    fMenuTextPasteSpecial( getStringFromResourceFile( IC_CDF_MENU_TEXT_PASTE_SPECIAL, "Paste Special" ) ),
    fMenuTextPasteLink( getStringFromResourceFile( IC_CDF_MENU_TEXT_PASTE_LINK, "Paste Link" ) ),
    fMenuTextAll( getStringFromResourceFile( IC_CDF_MENU_TEXT_ALL, "All" ) ),
    fMenuTextReturn( getStringFromResourceFile( IC_CDF_MENU_TEXT_RETURN, "Exit and return to" ) + " " ),
    fMenuTextDelete( getStringFromResourceFile( IC_CDF_MENU_TEXT_DELETE, "Delete" ) + " " ),
    fMenuTextCtrlX( getStringFromResourceFile( IC_CDF_MENU_TEXT_CTRLX, "Ctrl+X" ) ),
    fMenuTextCtrlC( getStringFromResourceFile( IC_CDF_MENU_TEXT_CTRLC, "Ctrl+C" ) ),
    fMenuTextCtrlV( getStringFromResourceFile( IC_CDF_MENU_TEXT_CTRLV, "Ctrl+V" ) ),
    fhPopupMenu( 0 ),
    fhEditMenu( 0 )
{
    handleEventsFor( &bundle.frameWindow() );
}

IMainMenuHandler::~IMainMenuHandler()
{
    stopHandlingEventsFor( &fBundle.frameWindow() );
}

// -----------------------------------------------------------------------------
// getStringFromResourceFile
// -----------------------------------------------------------------------------
// This method tries to load a string from the resource file using the supplied
// resource ID. If found, then that text is returned, otherwise the default text
// is returned.
IString IMainMenuHandler::getStringFromResourceFile( unsigned long stringId, IString strDefault )
{
    IString strTemp( IApplication::current().userResourceLibrary().tryToLoadString( stringId ) );
    if( strTemp.length() >0 )
        return strTemp;
    else
        return strDefault;
}

// -----------------------------------------------------------------------------
// itemExists
// -----------------------------------------------------------------------------
// helper function to check to see if a menu item exists
Boolean
IMainMenuHandler::itemExists( ISubmenu& menu, unsigned long id )
{
    Boolean bFound = false;

    // iterate through items on the menu until found
    ISubmenu::Cursor cursor( menu );
    for( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
    {
        if( menu.elementAt( cursor ).id() == id )
        {
            bFound = true;
            break;
        }

    } // next cursor

    return bFound;

}

// -----------------------------------------------------------------------------
// dispatchHandlerEvent
// -----------------------------------------------------------------------------
// Override ICLUI behavior when in-place active because ICLUI can't deal with
// the container (i.e. non IFrameWindow) menus.
Boolean
IMainMenuHandler::dispatchHandlerEvent ( IEvent& evt )
{
    // Use standard ICLUI code if not in-place active
    if ( &fBundle.objectView() == &fBundle.view() )
        return IMenuHandler::dispatchHandlerEvent( evt );

    switch (evt.eventId())
    {
    case WM_CONTEXTMENU:
      {
        IMenuEvent mnEvt(evt);
        return makePopUpMenu(mnEvt);
      }

    case WM_INITMENUPOPUP:
      {
        IMenuEvent mnEvt(evt);
        ISubmenu submenu( IMenuHandle( evt.parameter1() ) );
        return menuShowing(mnEvt, submenu);
      }

    default:
        // ignore everything else
        break;

    }

    return false;
}

// -----------------------------------------------------------------------------
// makePopUpMenu
// -----------------------------------------------------------------------------
Boolean
IMainMenuHandler::makePopUpMenu( IMenuEvent& event )
{
    IMenu* pPopupMenu;

    // create a popup menu
    if ( &fBundle.objectView() == &fBundle.view() )
    {
        // note: menu will automatically be deleted since the "set auto delete" is on
        pPopupMenu = new IPopUpMenu( event.window(), kPopupMenu );
    }
    else
    {
        fhPopupMenu = CreatePopupMenu();
        pPopupMenu = new ISubmenu( fhPopupMenu );
    }
    
    // see if this is an embedder model, and if so see if there is
    // a selected component
    IEmbedderModel* pEmbModel = NULL;
    IEmbeddedComponent* pComponent = NULL;
    ::dynamicCastTo( pEmbModel, fBundle.component().model() );
    if( pEmbModel )
        pComponent = pEmbModel->selectedComponent();

    // get object name
    // if this is an embedder model, and an object is selected, then
    // cut/copy apply to the selected object. Otherwise, if selectAll
    // is active they apply to the model as a whole.
    IString strObject = "";

    if( pComponent )
    {
        // get the object's name (for the menu)
        CAswOLEString str( 255 );
        pComponent->containedObject()->GetUserPresentableName( 2, str.GetOLEPtr(), 255 );
        CAswANSIString astr( str.GetOLEPtr() );
        strObject = IString(astr);

    }

    else
    {
        if( fBundle.fbSelectAll )
            strObject = fMenuTextAll;
    }


    // cut and copy
    // ------------
    pPopupMenu->addText( kEditCut, IString( fMenuTextCut + strObject ) );
    pPopupMenu->addText( kEditCopy, IString( fMenuTextCopy + strObject ) );

    // paste
    // ------------
    WORD wResults;
    IComponentStationery::stationery().pasteTarget().PasteInitMenuPopup( &wResults );
    Boolean bIsUs = IComponentStationery::stationery().clipSource().IsObjectOnClipboard();

    if( wResults & 1 /*PASTE_EDIT*/ )        pPopupMenu->addText( kEditPaste, fMenuTextPaste );
    if( pEmbModel && !bIsUs )
    {
        if( wResults & 2 /*PASTE_SPECIAL*/ ) pPopupMenu->addText( kEditPasteSpecial, fMenuTextPasteSpecial );
        if( wResults & 4 /*PASTE_LINK*/ )    pPopupMenu->addText( kEditPasteLink, fMenuTextPasteLink );

    } // if an embedder model

    // delete
    // ------------
    pPopupMenu->addText( kEditDelete, IString( fMenuTextDelete + strObject ) );

    // active component
    // ----------------
    if( pComponent )
    {
        pPopupMenu->addSeparator();
        pPopupMenu->addText( kVerbMin, "xxx" ); // xxx is a placeholder; it will be replaced by the object name

        // "object" -> use Aswan method to create popup menu
        pComponent->containedObject()->AddVerbMenu( 
            (HMENU)(void*)pPopupMenu->menuHandle(),
            kVerbMin, kVerbMax, TRUE, kConvert, NULL );
    }

        
    // need to try/catch since menu will not show if any other menu is currently showing
    try
    {
        if ( fhPopupMenu ) {
            HWND hOwner = (HWND)(void*)event.window()->handle();
            HWND hFrame = (HWND)(void*)IComponentStationery::stationery()
                                            .bundle().frameWindow().handle();
            IPoint nativePoint = ICoordinateSystem::convertToNative(
                           event.mousePosition(), event.window()->size() );
            POINTL scrPoint = nativePoint.asPOINTL();
            ClientToScreen( hOwner, (LPPOINT)&scrPoint );
            // Show the popup - Align topleft corner of menu with mouse pointer
            if (!TrackPopupMenu( fhPopupMenu,
                                 TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RIGHTBUTTON,
                                 (int)scrPoint.x, (int)scrPoint.y,
//                               0, hOwner, NULL ))
                                 0, hFrame, NULL ))
            {
                ITHROWGUIERROR("TrackPopupMenu");
            }
            delete pPopupMenu;
        }
        else
        {
            ((IPopUpMenu*)pPopupMenu)->show( event.mousePosition() );
            ((IPopUpMenu*)pPopupMenu)->setAutoDeleteObject();
        }
    }
       
    catch( IException& )
    {
        delete pPopupMenu;
        pPopupMenu = NULL;
    }
    
    return true;
}

// -----------------------------------------------------------------------------
// menuShowing
// -----------------------------------------------------------------------------
// called when a popup menu is about to display. 
Boolean
IMainMenuHandler::menuShowing( IMenuEvent& menuEvent, ISubmenu& submenu )
{
    HMENU hSubmenu = (HMENU)(void*)submenu.menuHandle();

    // if this is the popup menu just exit - it does not need any preprocessing
    if( fhPopupMenu == hSubmenu )
    {
        fhPopupMenu = 0;
        return false;
    }
    
    // "File" menu
    if( menuEvent.menuItemId() == kFileMenu )
    {
        // say "Exit and return to ..."
        if( itemExists( submenu, kFileExit ) )
        {
        
            // if we are embedded then we MUST be launched out of place
            // (since the File menu comes from containers during in-place activation)
            if( fBundle.component().isEmbedded() )
            {
                const IString& strName = fBundle.component().objectInstance()->hostName();
                submenu.setText( kFileExit, fMenuTextReturn + strName );
            }

            // yes, handled exit option
            return true;
        }
   
        // this was the file menu but the "Exit" option was removed
        return true;

    } // this was the File Menu

    Boolean bInPlace = &fBundle.objectView() != &fBundle.view();

    // "Edit" menu (or in-place submenu which may be the "Edit" menu)
    if( menuEvent.menuItemId() == kEditMenu ||
        ( bInPlace && ( !fhEditMenu || fhEditMenu == hSubmenu ) ) )
    {
        // in-place submenus have no id's so we have to test the items to
        // find out if we're working with the "Edit" menu or not.
        Boolean foundEditItem = false;

        // see if this is an embedder model, and if so see if there is
        // a selected component
        IEmbedderModel* pEmbModel = NULL;
        IEmbeddedComponent* pComponent = NULL;

        ::dynamicCastTo( pEmbModel, fBundle.component().model() );
        
        if( pEmbModel )
            pComponent = pEmbModel->selectedComponent();

        // ---------------------------------
        // Cut/Copy/Delete menu items
        // ---------------------------------
        // if this is an embedder model, and an object is selected, then
        // cut/copy/delete apply to the selected object. Otherwise, if
        // selectAll is active they apply to the model as a whole.
        IString strObject = "";
        if( pComponent )
        {
            // get the object's name (for the menu)
            CAswOLEString str( 255 );
            pComponent->containedObject()->GetUserPresentableName( 2, str.GetOLEPtr(), 255 );
            CAswANSIString astr( str.GetOLEPtr() );
            strObject = IString(astr);
        }
        else
        {
            if( fBundle.fbSelectAll )
                strObject = fMenuTextAll;
        }

        // Cut
        if( itemExists( submenu, kEditCut ) )
        {
            submenu.setText( kEditCut,
                             IString( fMenuTextCut + strObject + fMenuTextCtrlX ) );
            if( strObject.length() )
                submenu.enableItem( kEditCut );
            else
                submenu.disableItem( kEditCut );
            foundEditItem = true;
        }

        // Copy
        if( itemExists( submenu, kEditCopy ) )
        {
            submenu.setText( kEditCopy,
                             IString( fMenuTextCopy + strObject+ fMenuTextCtrlC ) );
            if( strObject.length() )
                submenu.enableItem( kEditCopy );
            else
                submenu.disableItem( kEditCopy );
            foundEditItem = true;
        }

        // ---------------------------------
        // Paste/Paste Special/Paste Link
        // ---------------------------------
        // check the clipboard to see what is available
        WORD wResults;
        IComponentStationery::stationery().pasteTarget().PasteInitMenuPopup( &wResults );

        // check to see if "we" are the app that placed the data on the clipboard
        // if so,then we cannot paste link or paste special (it does not make sense)
        Boolean bIsUs = IComponentStationery::stationery().clipSource().IsObjectOnClipboard();

        // enable paste?
        if( itemExists( submenu, kEditPaste ) )
        {
            if( wResults & 1 /*PASTE_EDIT*/ )
                submenu.enableItem( kEditPaste );
            else
                submenu.disableItem( kEditPaste );
            foundEditItem = true;
        }    
        
        // enable paste special?
        if( itemExists( submenu, kEditPasteSpecial ) )
        {
            if( wResults & 2 /*PASTE_SPECIAL*/ && !bIsUs )
                submenu.enableItem( kEditPasteSpecial );
            else
                submenu.disableItem( kEditPasteSpecial );
            foundEditItem = true;
        }

        // enable paste link?
        if( itemExists( submenu, kEditPasteLink ) )
        {
            if( wResults & 4 /*PASTE_LINK*/ && !bIsUs )
                submenu.enableItem( kEditPasteLink );
            else
                submenu.disableItem( kEditPasteLink );
            foundEditItem = true;
        }

        // Delete
        if( itemExists( submenu, kEditDelete ) )
        {
            submenu.setText( kEditDelete,
                             IString( fMenuTextDelete + strObject ) );
            if( strObject.length() )
                submenu.enableItem( kEditDelete );
            else
                submenu.disableItem( kEditDelete );
            foundEditItem = true;
        }

        // Select all
        if( itemExists( submenu, kEditSelectAll ) )
        {
            if( fBundle.fbSelectAll )
                submenu.checkItem( kEditSelectAll );
            else
                submenu.uncheckItem( kEditSelectAll );
            foundEditItem = true;
        }

        // stop here if we havn't found any edit menu items
        if( !foundEditItem )
            return false;

        // "Object" menu item
        if( pComponent )
        {
            // "object" -> use Aswan method to create popup menu
            pComponent->containedObject()->AddVerbMenu( 
                                            hSubmenu, kVerbMin, kVerbMax,
                                            TRUE, kConvert, NULL );
        }
        else
        {
            // no selected component - show the word "Object" greyed
            CAswContainedObject::AddVerbMenuNoSelectedObject(
                                            hSubmenu, kVerbMin, kVerbMax,
                                            kConvert );
        }

        // this was the Edit Menu
        fhEditMenu = hSubmenu;

        return true;

    }

    return false;

}


/*******************************************************************************
*                                                                              *
*   Class IGUIBundle                                                           *
*                                                                              *
*   Main frame window for the component                                        *
*                                                                              *
*******************************************************************************/

// Constuction/Destruction
IGUIBundle::IGUIBundle() :
    fpFrameWindow( NULL ),
    fpComponent( NULL ),
    fpView( NULL ),
    fpInPlaceView( NULL ),
    fpInPlaceBar( NULL ),
    fpInPlaceTools( NULL ),
    fpInPlaceFrame( NULL ),
    fpMenuHandler( NULL ),
    fpMainMenuHandler( NULL ),
    fbSelectAll( false ),
    fbUsingInPlace( false )
{
}

IGUIBundle::~IGUIBundle()
{
    fpMenuHandler->stopHandlingEventsFor( fpFrameWindow );

    delete fpMenuHandler;
    delete fpMainMenuHandler;
    delete fpView;
    delete fpFrameWindow;
    delete fpComponent;

}

// -----------------------------------------------------------------------------
// adoptFrameWindow
// -----------------------------------------------------------------------------
BOOL
IGUIBundle::adoptFrameWindow( IComponentFrameWindow* pFrameWindow )
{
    assert( fpFrameWindow == NULL );
    
    fpFrameWindow = pFrameWindow;
    
    // create menu (command) handler
    fpMenuHandler = new ICommandConnectionTo<IGUIBundle>( *this, handleDefaultMenuCommands );
    fpMenuHandler->handleEventsFor( pFrameWindow );
    fpMenuHandler->handleEventsFor( &pFrameWindow->toolBar() );

    // create menu preprocessor handler
    fpMainMenuHandler = new IMainMenuHandler( *this );

    return TRUE;
}

// -----------------------------------------------------------------------------
// adoptComponent
// -----------------------------------------------------------------------------
BOOL
IGUIBundle::adoptComponent( IComponent* pComponent )
{
    assert( fpComponent == NULL );
    
    fpComponent = pComponent;

    return TRUE;
}


// -----------------------------------------------------------------------------
// adoptView
// -----------------------------------------------------------------------------
// adopt an out-of-place view. Attaches the view as the client area of the 
// frame window
BOOL
IGUIBundle::adoptView( IView* pView )
{
    assert( fpView == NULL );

    fpView = pView;
    
    // attach the view as the client area of the frame window
    assert( fpFrameWindow != NULL ); // the frame window MUST be adopted before the view
    frameWindow().setClient( pView );

    // initialize the view (starts the notifier and paint handlers)
    pView->initialize();

    return TRUE;
}


// -----------------------------------------------------------------------------
// initializeInPlaceWindows
// -----------------------------------------------------------------------------
void
IGUIBundle::createInPlaceWindows()
{
    
    // create an in-place (always hidden) frame window
    const char* name = IBase::messageText(IC_CDF_IN_PLACE_FRAME_NAME,
                                          IComponentStationery::stationery().appName());
    fpInPlaceFrame = new IFrameWindow( name );

    fpInPlaceBar = new IMenuBar( kInplaceMenu, fpInPlaceFrame );
    try {
        fpInPlaceBar->disableItem( kEditUndo );
    }
    catch( IException& )
    {
    }
    fpInPlaceTools = new IComponentToolBar( *fpInPlaceFrame );

    // work around (ICLUI bug?) because the frame window never seems
    // to get the toolbar events when we use a floating toolbar
    fpMenuHandler->handleEventsFor( fpInPlaceTools );

    // switch to in-place windows during in-place view construction
    // and initialization
    useInPlaceWindows();

    // create in-place view
    fpInPlaceView = IComponentStationery::stationery().createView();
    fpInPlaceFrame->setClient( fpInPlaceView );

    // initialize the view (starts the notifier and paint handlers)
    fpInPlaceView->initialize();
// SMY12
    fpMainMenuHandler->handleEventsFor( fpInPlaceView );

    // switch back to out-of-place windows
    useInPlaceWindows( false );

}

// -----------------------------------------------------------------------------
// destroyInPlaceWindows
// -----------------------------------------------------------------------------
Boolean
IGUIBundle::destroyInPlaceWindows()
{
    
    if ( fpInPlaceView != NULL )
    {
// SMY12
        fpMainMenuHandler->stopHandlingEventsFor( fpInPlaceView );
        fpInPlaceView->finalize();

        useInPlaceWindows( false );

        // work around (ICLUI bug?) because the frame window never seems
        // to get the toolbar events when we use a floating toolbar
        fpMenuHandler->stopHandlingEventsFor( fpInPlaceTools );

        // delete the internally created menu bar, tool bar, and frame window
        delete fpInPlaceBar;
        fpInPlaceBar = NULL;
        delete fpInPlaceTools;
        fpInPlaceTools = NULL;
        delete fpInPlaceFrame;
        fpInPlaceFrame = NULL;

        delete fpInPlaceView;
        fpInPlaceView = NULL;

        return true;
    }

    return false;
    
}

// -----------------------------------------------------------------------------
// frameWindow
// -----------------------------------------------------------------------------
IFrameWindow&
IGUIBundle::frameWindow() const
{
    return *fpFrameWindow;
}

// -----------------------------------------------------------------------------
// objectFrame
// -----------------------------------------------------------------------------
// returns either the in-place frame window (if present it means we are
// in-place), or the main frame window (out-of-place)
IFrameWindow&
IGUIBundle::objectFrame() const
{

    if( fbUsingInPlace )
        return *fpInPlaceFrame;
    else
        return *fpFrameWindow;

}

// -----------------------------------------------------------------------------
// objectView
// -----------------------------------------------------------------------------
// returns either the in-place view (if present it means we are in place), or
// the main frame client view
IView&
IGUIBundle::objectView() const
{

    if( fbUsingInPlace )
        return *fpInPlaceView;
    else
        return *fpView;

}

// -----------------------------------------------------------------------------
// objectMenuBar
// -----------------------------------------------------------------------------
// returns either the in-place menu bar, or the main frame window menu bar
IMenuBar&
IGUIBundle::objectMenuBar()
{

    if( fbUsingInPlace )
        return *fpInPlaceBar;
    else
        return fpFrameWindow->menuBar();

}

// -----------------------------------------------------------------------------
// objectToolBar
// -----------------------------------------------------------------------------
// returns either the in-place tool bar, or the main frame window tool bar
IToolBar&
IGUIBundle::objectToolBar()
{

    if( fbUsingInPlace )
        return *fpInPlaceTools;
    else
        return fpFrameWindow->toolBar();

}

// -----------------------------------------------------------------------------
// handleDefaultMenuCommands
// -----------------------------------------------------------------------------
Boolean
IGUIBundle::handleDefaultMenuCommands( ICommandEvent& theEvent )
{

    Boolean bHandled = false;


    // check for OLE verbs
    if( theEvent.commandId() >= kVerbMin && theEvent.commandId() <= kVerbMax )
    {
        handleDoVerb( theEvent.commandId() - kVerbMin );
        bHandled = true;
    }

    else if( theEvent.commandId() == ISystemMenu::idClose )
    {
#ifdef IC_SCREEN_DEBUG            
cout << "inside handle close" << endl;
#endif
        if ( theEvent.window() == fpFrameWindow )
        {
            handleExit();
            bHandled = true;
        }
        else
            bHandled = false;
    }

    else
    switch( theEvent.commandId() )
    {
        case kFileNew:
            handleFileNew();
            bHandled = true;
            break;
        case kFileOpen:
            handleFileOpen();
            bHandled = true;
            break;
        case kFileSave:
            try {
                handleFileSave();
            }
            catch ( IException& ) {
                const char* msg = IBase::messageText(IC_CDF_SAVE_FAILED);
                IComponentStationery::stationery().messageBox( msg,
                                                  IMessageBox::moveable  |
                                                  IMessageBox::applicationModal|
                                                  IMessageBox::errorIcon |
                                                  IMessageBox::okButton);
            }
            bHandled = true;
            break;
        case kFileSaveAs:
            try {
                handleFileSaveAs();
            }
            catch ( IException& ) {
                const char* msg = IBase::messageText(IC_CDF_SAVE_FAILED);
                IComponentStationery::stationery().messageBox( msg,
                                                  IMessageBox::moveable  |
                                                  IMessageBox::applicationModal|
                                                  IMessageBox::errorIcon |
                                                  IMessageBox::okButton);
            }
            bHandled = true;
            break;
        case kFileUpdate:
            bHandled = false;
            break;
        case kFileSaveCopyAs:
            try {
                handleFileSaveAs( true );
            }
            catch ( IException& ) {
                const char* msg = IBase::messageText(IC_CDF_SAVE_FAILED);
                IComponentStationery::stationery().messageBox( msg,
                                                  IMessageBox::moveable  |
                                                  IMessageBox::applicationModal|
                                                  IMessageBox::errorIcon |
                                                  IMessageBox::okButton);
            }
            bHandled = true;
            break;
        case kFileExit:
            handleExit();
            bHandled = true;
            break;

        case kEditUndo: 
            bHandled = false;
            break;
        case kEditCut:
            handleCut();
            bHandled = true;
            break;
        case kEditCopy:
            handleCopy();
            bHandled = true;
            break;
        case kEditPaste:
            handlePaste();
            bHandled = true;
            break;
        case kEditPasteSpecial:
            handlePasteSpecial();
            bHandled = true;
            break;
        case kEditInsertObject:
            handleInsertObject();
            bHandled = true;
            break;
        case kEditPasteLink:
            handlePasteLink();
            bHandled = true;
            break;
        case kEditLinks:
            handleEditLinks();
            bHandled = true;
            break;
        case kConvert:
            handleConvert();
            bHandled = true;
            break;
        case kEditDelete:
            handleDelete();
            bHandled = true;
            break;
        case kEditSelectAll:
            handleSelectAll( !fbSelectAll );
            bHandled = true;
            break;
        case kViewToolbar:
        {
            Boolean checked =
                fpFrameWindow->menuBar().isItemChecked( kViewToolbar );
            if ( !checked )
                fpFrameWindow->showToolBar( true );
                // fpFrameWindow->toolBar().setLocation( IToolBar::aboveClient );
            else
                fpFrameWindow->showToolBar( false );
                // fpFrameWindow->toolBar().setLocation( IToolBar::hidden );
            bHandled = true;
            break;
        }
        case kHelpAbout:
            bHandled = false;
            break;
        default:
            bHandled = false;
            break;

    }

    return bHandled;
}


// -----------------------------------------------------------------------------
// okToDiscardFile
// -----------------------------------------------------------------------------
Boolean
IGUIBundle::okToDiscardFile()
{

    if ( fpComponent->model()->hasChanged() )
    {

        IString docName = fpComponent->documentName();
        //   "Save changes to %1 ?"
        const char* msg = IBase::messageText(IC_CDF_SAVE, docName);
        IMessageBox::Response ret = IComponentStationery::stationery().
                                         messageBox( msg,
                                                     IMessageBox::moveable         |
                                                     IMessageBox::applicationModal |
                                                     IMessageBox::queryIcon        |
                                                     IMessageBox::yesNoCancelButton);
        if ( ret == IMessageBox::cancel )
            return false;

        if ( ret == IMessageBox::yes )
        {
            try
            {
                handleFileSave();
            }
            catch ( IException& )
            {
                const char* msg = IBase::messageText(IC_CDF_DISCARD);
                ret = IComponentStationery::stationery().messageBox( msg,
                                                  IMessageBox::moveable         |
                                                  IMessageBox::applicationModal |
                                                  IMessageBox::queryIcon        |
                                                  IMessageBox::yesNoButton );
                if ( ret == IMessageBox::cancel )
                    return false;
                
            }
        }

    }

    return true;
}

// -----------------------------------------------------------------------------
// handleFileNew
// -----------------------------------------------------------------------------
void
IGUIBundle::handleFileNew()
{

    if ( okToDiscardFile() ) {
        // call the component to load a null file name
        fpComponent->storage().loadFile( "" );
    }

}

// -----------------------------------------------------------------------------
// handleFileOpen
// -----------------------------------------------------------------------------
void
IGUIBundle::handleFileOpen()
{

    IString strFilter =
        IString( '*' ) + IComponentStationery::stationery().fileExt();

    // display file open

    // create settings object
    IFileDialog::Settings settings;
    settings.setOpenDialog();
    settings.setFileName( strFilter );

    // create the file dialog object
    IFileDialog dialog( &frameWindow(), &frameWindow(), settings );

    if ( dialog.pressedOK() &&
         ( dialog.fileName() != fpComponent->storage().fileName() ||
           dialog.fileName().length() == 0 ) &&
         okToDiscardFile() )
    {
        try {
            fpComponent->storage().loadFile( dialog.fileName() );
        }
        catch ( IException& ) {
            // "Unable to open: %1"
            const char* msg = IBase::messageText(IC_CDF_OPEN_FAILED,
                                                 dialog.fileName());
            IComponentStationery::stationery().messageBox( msg,
                                              IMessageBox::moveable         |
                                              IMessageBox::applicationModal |
                                              IMessageBox::errorIcon        |
                                              IMessageBox::okButton );
        }

    }

}

// -----------------------------------------------------------------------------
// handleFileSave
// -----------------------------------------------------------------------------
void
IGUIBundle::handleFileSave()
{

    IString name = fpComponent->storage().fileName();

    if ( name.length() != 0 ) {
        // if there is a current file name, then just do a save
        fpComponent->storage().saveFile( name, true, false );
    }
    else
    {
        // no current file name - do a save as
        handleFileSaveAs();
    }

}

// -----------------------------------------------------------------------------
// handleFileSaveAs
// -----------------------------------------------------------------------------
void
IGUIBundle::handleFileSaveAs( Boolean copy )
{

    IString strFilter =
        IString( '*' ) + IComponentStationery::stationery().fileExt();

    // display file save as dialog box

    // create settings object
    IFileDialog::Settings settings;
    settings.setSaveAsDialog();
    settings.setFileName( strFilter );

    // get file name from the file dialog
    IFileDialog dialog( &frameWindow(), &frameWindow(), settings );

    if ( dialog.pressedOK() )
    {
        fpComponent->storage().saveFile( dialog.fileName(), false, !copy );

    }
    
}

// -----------------------------------------------------------------------------
// handleExit
// -----------------------------------------------------------------------------
void
IGUIBundle::handleExit(Boolean close )
{
#ifdef IC_SCREEN_DEBUG
cout << "handleExit, embedded: " << fpComponent->isEmbedded() << endl;
#endif

    // close will be set to TRUE if a SC_CLOSE command is received
    // and there is no need to bring up any dialog to save content
    if ( close || fpComponent->isEmbedded() || okToDiscardFile() )
    {
        // hide the frame window
        fpFrameWindow->hide();
        fpComponent->objectInstance()->AdviseOnHideWindow();
        
        // close component (and embedded components if we're a container)
        DWORD dwSaveOption =
            fpComponent->isEmbedded() ? OLECLOSE_SAVEIFDIRTY : OLECLOSE_NOSAVE;
        fpComponent->objectInstance()->CloseObjectInstance( dwSaveOption );

        // shutdown of in-place windows and finalize on view is required 
        // before the frame window is closed
        destroyInPlaceWindows();
        fpView->finalize();
    
        // close the frame window
        fpFrameWindow->close();

        // static destruction of the stationery object will kick in next
        
    }
    
}

// -----------------------------------------------------------------------------
// handleInsertObject
// -----------------------------------------------------------------------------
// displays the Insert Object dialog box. Then, if OK was pressed, calls on the 
// embedder model to create a new object
void
IGUIBundle::handleInsertObject()
{
    // show Lotus insert object dialog
    OLEUILTSCREATEOBJECT co;
    memset( &co, 0 , sizeof(co) );

    co.IO.cbStruct = sizeof( OLEUILTSCREATEOBJECT );
    co.IO.dwFlags = IOF_SELECTCREATENEW;
    co.IO.hWndOwner = (HWND)(void*)objectView().handle();

    CAswOLEString oleFile( MAX_PATH );
    oleFile = "";
    co.IO.lpszFile = oleFile.GetOLEPtr(); 
    co.IO.cchFile = MAX_PATH; 
 
    if( OleUILtsCreateObject( &co ) == OLEUI_OK )
    {
        
        
        // if they pressed "OK" create the object
        IEmbedderModel* pModel = NULL;
        ::dynamicCastTo( pModel, component().model() );
        if( pModel )
        {
            IEmbeddedComponent* pComponent = pModel->createEmbeddedComponent();
            pModel->adopt( pComponent );

            // must massage flags from OleUILtsCreateObject - MS/Lotus bug
            DWORD dwFlags = 0;
            BOOL bActivate = true;
            if ( co.IO.dwFlags & IOF_SELECTCREATEFROMFILE )
            {
                bActivate = false;
                dwFlags = (co.IO.dwFlags & IOF_CHECKLINK) ? IOF_CREATELINKOBJECT : IOF_CREATEFILEOBJECT;
            }
            else
            {
                ASSERT( co.IO.dwFlags & IOF_SELECTCREATENEW );
                dwFlags = IOF_CREATENEWOBJECT;
            }

            if ( co.IO.dwFlags & IOF_CHECKDISPLAYASICON )
            {
                pComponent->setAspect( IComponent::kIcon );

                dwFlags |= IOF_CHECKDISPLAYASICON; // not sure what this does?
            }
            
            CAswResult ar = pComponent->containedObject()->
                    CreateFromInsertObject( dwFlags,
                                            co.IO.clsid,
                                            co.IO.lpszFile,
                                            co.IO.hMetaPict );
            if( ASWFAILED( ar ) )
            {
                const char* msg = IBase::messageText(IC_CDF_INSERT_FAILED);
                IComponentStationery::stationery().messageBox( msg,
                                                  IMessageBox::moveable  |
                                                  IMessageBox::applicationModal|
                                                  IMessageBox::errorIcon |
                                                  IMessageBox::okButton);
//>>>>>>                ShowAswResult( ar, msg );
                delete pModel->orphan( *pComponent );
            }

            else
            {
                // set up a presentation area for the object and launch it
                pComponent->initialize( true, bActivate );
            }

        } // we are an embedder model

    } // they pressed "OK"


}


// -----------------------------------------------------------------------------
// handleCut
// -----------------------------------------------------------------------------
// a cut is a copy, then a delete of the source
// if an embedded object is currently selected then cut it, else call out to
// the model to do a cut
void
IGUIBundle::handleCut()
{

    // first do a copy
    handleCopy();

    // and then do a delete
    handleDelete();

}


// -----------------------------------------------------------------------------
// handleDelete
// -----------------------------------------------------------------------------
// if one or more embedded objects are currently selected then delete it/them
void
IGUIBundle::handleDelete()
{
    IEmbedderModel* pEmbedderModel = NULL;
    IEmbeddedComponent* pComponent = NULL;

    // see if we have an embedder model or a model
    ::dynamicCastTo( pEmbedderModel, component().model() );
    if( !pEmbedderModel )
        // applications must override this method if they want to cut
        // native model data
        return;
        
    pComponent = pEmbedderModel->selectedComponent();
    
    if( pComponent )
        // delete the selected component
        delete pEmbedderModel->orphan( *pComponent );
        
    else if( fbSelectAll )
    {
        // delete all embedded components
        const IEmbedderModel::ComponentList& componentList =
            pEmbedderModel->componentList();

        while ( componentList.numberOfElements() != 0 )
        {
            IEmbeddedComponent* pCurrentComp = componentList.firstElement();
            delete pEmbedderModel->orphan( *pCurrentComp );
        }
    }
    
}

// -----------------------------------------------------------------------------
// handleCopy
// -----------------------------------------------------------------------------
// copy the embedded component onto the clipboard (if a component is selected)
// else copy the model data onto the clipboard
void
IGUIBundle::handleCopy()
{

    IEmbedderModel* pEmbedderModel = NULL;
    IEmbeddedComponent* pComponent = NULL;

    // see if we have an embedder model or a model
    ::dynamicCastTo( pEmbedderModel, component().model() );
    if( pEmbedderModel )
        pComponent = pEmbedderModel->selectedComponent();
    
    if( pComponent || fbSelectAll )
    {
        // if pComponent is NULL this method will put the whole object
        // on the clipboard
        IComponentStationery::stationery().clipSource().PutItemOnClipboard( pComponent );
    }

}

// -----------------------------------------------------------------------------
// handlePaste
// -----------------------------------------------------------------------------
void
IGUIBundle::handlePaste()
{

    // call on the paste target to handle the paste
    IComponentStationery::stationery().pasteTarget().ProcessPaste();

}


// -----------------------------------------------------------------------------
// handlePasteSpecial
// -----------------------------------------------------------------------------
void
IGUIBundle::handlePasteSpecial()
{

    // call on the paste target to handle the paste
    IComponentStationery::stationery().pasteTarget().
        ProcessPasteSpecial( (HWND)(void*)objectView().handle() );

}

// -----------------------------------------------------------------------------
// handlePasteLink
// -----------------------------------------------------------------------------
void
IGUIBundle::handlePasteLink()
{

    // call on the paste target to handle the paste link
    IComponentStationery::stationery().pasteTarget().ProcessPaste( TRUE );

}

// -----------------------------------------------------------------------------
// handleEditLinks
// -----------------------------------------------------------------------------
void
IGUIBundle::handleEditLinks()
{

    // call on the CAswUILinkContainer to show dialog
    IEmbedderModel* pEmbModel = NULL;
    ::dynamicCastTo( pEmbModel, component().model() );

    pEmbModel->container()->
        DoManageLinksDlg( (HWND)(void*)objectView().handle() );

}



// -----------------------------------------------------------------------------
// handleDoVerb
// -----------------------------------------------------------------------------
// do a verb on the currently selected component
void
IGUIBundle::handleDoVerb( unsigned iVerb )
{

    // see if we have an embedder model or a model
    IEmbedderModel* pEmbedderModel = NULL;
    IEmbeddedComponent* pComponent = NULL;
    ::dynamicCastTo( pEmbedderModel, component().model() );
    if( pEmbedderModel )
        pComponent = pEmbedderModel->selectedComponent();

    if( pComponent ) {
        pComponent->showGrabHandles();

        try {
            pComponent->goActive( iVerb );

            // turn off grab handles if in-place active
            if ( pComponent->isInPlaceActive() )
                pComponent->showGrabHandles( false );
        
        }
        catch ( IException& ) {
            // "Unable to activate server"
            const char* msg = IBase::messageText(IC_CDF_ACTIVATE_FAILED);
            IComponentStationery::stationery().messageBox( msg,
                                              IMessageBox::moveable         |
                                              IMessageBox::applicationModal |
                                              IMessageBox::errorIcon        |
                                              IMessageBox::okButton );
        }

    }

}

// -----------------------------------------------------------------------------
// handleConvert
// -----------------------------------------------------------------------------
void
IGUIBundle::handleConvert()
{
    // not implemented - this should be a user callout
    return;
}


// -----------------------------------------------------------------------------
// handleSelectAll
// -----------------------------------------------------------------------------
void
IGUIBundle::handleSelectAll( Boolean selectAll )
{
    if ( fbSelectAll == selectAll )
        return;
    
    // toggle select all
    fbSelectAll = selectAll;
  
    if( fbSelectAll )
    {
        IEmbedderModel* pEmbedderModel = NULL;
        ::dynamicCastTo( pEmbedderModel, component().model() );

        if ( pEmbedderModel && pEmbedderModel->selectedComponent() != NULL ) 
        {
            // there is no "current selection" when all are selected
            pEmbedderModel->selectComponent( NULL );

            objectView().refresh();
        }
        
    }

}

