/*******************************************************************************
* FILE NAME: iview.cpp                                                         *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the classes:                                             *
*     IView                                                                    *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1996             *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*   Portions or this component are based on work originating from Taligent.    *
*   (C) Copyright Taligent, Inc. 1996 - All rights reserved.                   *
*                                                                              *
*******************************************************************************/
#include <precomp.hpp>

#include "iview.hpp"

#include <istatnry.hpp>
#include <icompnen.hpp>
#include <iguibndl.hpp>
#include <icmpfram.hpp>
#include <iembmod.hpp>
#include <iecmpimp.hpp>
#include <idataex.hpp>
#include <icmpfram.hpp>

#include <iobservr.hpp>
#include <ipainhdr.hpp>
#include <imoushdr.hpp>
#include <imenubar.hpp>
#include <isizehdr.hpp>
#include <ihandle.hpp>

class IViewResizeHandler;
class IViewImplementation;
class IViewHelpHandler;

/*******************************************************************************
*                                                                              *
*   Class IViewResizeHandler                                                   *
*                                                                              *
*******************************************************************************/

class IViewResizeHandler : public IResizeHandler
{
public:
                            IViewResizeHandler( IViewImplementation*, IView& );
    virtual                 ~IViewResizeHandler();

protected:
    virtual Boolean         windowResize(IResizeEvent& event);

private:
    IView&                  fView;
    IViewImplementation*    fViewImp;

};


/*******************************************************************************
*                                                                              *
*   Class IViewHelpHandler                                                     *
*                                                                              *
*******************************************************************************/

class IViewHelpHandler : public IHandler {
public:
    IViewHelpHandler( IView* view );
    ~IViewHelpHandler();

protected:
    virtual Boolean         dispatchHandlerEvent(IEvent& event);

private:
    IView*                  fpView;
};


/*******************************************************************************
*                                                                              *
*   Class IViewImplementation                                                  *
*                                                                              *
*   Implementation class for IView                                             *
*                                                                              *
*******************************************************************************/

class IViewImplementation
{
public:
    IViewImplementation( IView* view );
    ~IViewImplementation();

    // paint handling
    Boolean                 handlePaintWindow( IPaintEvent& event );

    // mouse handling events
    Boolean                 handleMouseClicked( IMouseClickEvent& event );
//  Boolean                 handleMouseMoved( IMouseEvent& event );

    // window resize event
    Boolean                 handleWindowResized( IResizeEvent& event);

public:
    IView*                                   fpView;
    IAswDropTarget*                          fpDropTarget;

    IObserverConnectionTo<IView>             fConnection;
    IPaintConnectionTo<IViewImplementation>  fPaintHandler;
    IMouseConnectionTo<IViewImplementation>  fMouseHandler;
    IViewResizeHandler                       fResizeHandler;
    IViewHelpHandler*                        fpHelpHandler;

};


// -----------------------------------------------------------------------------
// IViewResizeHandler
// -----------------------------------------------------------------------------
IViewResizeHandler::IViewResizeHandler( IViewImplementation* viewImp, IView& view )
                   :fView( view )
                   ,fViewImp( viewImp )
{
    handleEventsFor( &fView );
    if ( &fView.bundle().objectFrame() == &fView.bundle().frameWindow() )
    {
        handleEventsFor( &fView.bundle().frameWindow() );
    }
}

IViewResizeHandler::~IViewResizeHandler()
{
    if ( &fView.bundle().objectFrame() == &fView.bundle().frameWindow() )
    {
        stopHandlingEventsFor( &fView.bundle().frameWindow() );
    }
    stopHandlingEventsFor( &fView );
}

Boolean IViewResizeHandler::windowResize(IResizeEvent& event)
{
   fViewImp->handleWindowResized(event);
}


// -----------------------------------------------------------------------------
// IViewHelperHandler
// -----------------------------------------------------------------------------
IViewHelpHandler::IViewHelpHandler( IView* view ) :
    fpView( view )
{
    handleEventsFor( view );
}

// -----------------------------------------------------------------------------
// ~IViewHelperHandler
// -----------------------------------------------------------------------------
IViewHelpHandler::~IViewHelpHandler()
{
    stopHandlingEventsFor( fpView );
}

// -----------------------------------------------------------------------------
// dispatchHandlerEvent
// -----------------------------------------------------------------------------
Boolean IViewHelpHandler::dispatchHandlerEvent(IEvent& event)
{
    Boolean returnCode = false;
    if (event.eventId() == WM_HELP)
    {
        // route the help message to the frame window
        event.setResult( fpView->bundle().frameWindow().sendEvent( event ) );
        returnCode = true;
    }
    return returnCode;
}


// -----------------------------------------------------------------------------
// IViewImplementation
// -----------------------------------------------------------------------------
IViewImplementation::IViewImplementation( IView* view ) :
    fpView( view ),
    fpDropTarget( NULL ),
    fConnection  ( *view, IView::handleNotification ),
    fPaintHandler( *this, IViewImplementation::handlePaintWindow ),
    fMouseHandler( *this, IViewImplementation::handleMouseClicked ),
//                        IViewImplementation::handleMouseMoved )
    fResizeHandler( this, *view ),
    fpHelpHandler( NULL )
{
    if ( &view->bundle().objectFrame() != &view->bundle().frameWindow() )
        fpHelpHandler = new IViewHelpHandler( view );
}

// -----------------------------------------------------------------------------
// ~IViewImplementation
// -----------------------------------------------------------------------------
IViewImplementation::~IViewImplementation()
{
    delete fpHelpHandler;
}

// -----------------------------------------------------------------------------
// handleWindowResized
// -----------------------------------------------------------------------------
Boolean
IViewImplementation::handleWindowResized( IResizeEvent& event )
{

    if (event.oldSize() != event.newSize())
    {
        if ( event.window() == fpView )
        {
            // since the view window (esp. the in-place view) may have been
            // resized directly in response to Windows messages we have to call
            // the ICLUI sizeTo to keep the wrappers in synch.
            fpView->sizeTo( event.newSize() );
        }
        else
        {
            // the frame window was resized manually
            ((IComponentFrameWindow*)&fpView->bundle().frameWindow())->
                changedNegotiatedSpace();
            return false;
        }

    }
    return true;
}

// -----------------------------------------------------------------------------
// handleMouseClicked
// -----------------------------------------------------------------------------
// We are only interested in clicks on an embedder model
// in a view we are interested in single clicks and double clicks
Boolean
IViewImplementation::handleMouseClicked( IMouseClickEvent& event )
{

    // we are only interested in left mouse clicks (for now)
    if( event.mouseButton() != IMouseClickEvent::button1 )
        return false;

    // see if this is an embedder model
    IEmbedderModel* pEmbModel = NULL;
    ::dynamicCastTo( pEmbModel, fpView->model() );
    if( !pEmbModel )
        return false;

    // see if they clicked on an embedded item

    IEmbeddedComponent* pEmbeddedComp = NULL;

    IEmbeddedComponent* selected = pEmbModel->selectedComponent();
    if ( selected && selected->area().contains( event.mousePosition() ) )
    {
        pEmbeddedComp = selected;
    }
    else
    {
        const IEmbedderModel::ComponentList& componentList = pEmbModel->componentList();
        ICursor& cursor = *componentList.newCursor();

        // iterate through the items.
        for( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
        {

            IEmbeddedComponent* pCurrentComp = componentList.elementAt( cursor );
            if( pCurrentComp )
            {

                // see if they clicked on this one
                BOOL bClickedOnThisOne = pCurrentComp->area().contains( event.mousePosition() );

                if( bClickedOnThisOne )
                    pEmbeddedComp = pCurrentComp;

            }

        } // next component

    }

    if ( pEmbeddedComp != NULL )
    {
        switch( event.mouseAction() )
        {
            case IMouseClickEvent::down:
            {
                // for single clicks just select the component

                // delay adjusting tool bar until mouse up-clicks
                if ( selected && selected->isInPlaceActive() )
                    ((IComponentFrameWindow&)fpView->bundle().frameWindow())
                    .delayAdjustment();

                pEmbModel->selectComponent( pEmbeddedComp );

                // tell grab handles that the mouse is down
                // (no grab handles for in-place objects)
                IGrabHandles* handles = pEmbeddedComp->grabHandles();
                if( handles )
                    handles->mouseDown( event.mousePosition() );
                fpView->refresh();
                break;
            }
            case IMouseClickEvent::doubleClick:
            {
                // for double clicks, activate the item as well

//              pEmbModel->selectComponent( pEmbeddedComp );
                try
                {
                    pEmbeddedComp->goActive();

                    // turn off grab handles if in-place active
                    if ( pEmbeddedComp->isInPlaceActive() )
                        pEmbeddedComp->showGrabHandles( false );

                }
                catch( IException& )
                {
                    IGrabHandles* handles = pEmbeddedComp->grabHandles();
                    if( handles )
                        handles->mouseDown( event.mousePosition() );
                    fpView->refresh();

                    // "Unable to activate server"
                    const char* msg = IBase::messageText(IC_CDF_ACTIVATE_FAILED);
                    IComponentStationery::stationery().messageBox( msg,
                                              IMessageBox::moveable         |
                                              IMessageBox::applicationModal |
                                              IMessageBox::errorIcon        |
                                              IMessageBox::okButton );
                }
                break;
            }

        }
        
    }
    
    // clicked in an area where there is no embedded component
    else
    {
        if ( event.mouseAction() != IMouseClickEvent::up )
        {
            // deselect any embedded component
            if ( pEmbModel->selectedComponent() != NULL )
            {
                pEmbModel->selectComponent( NULL );
                fpView->refresh();
            }
        }

    }

    return true;
}

// -----------------------------------------------------------------------------
// handlePaintWindow
// -----------------------------------------------------------------------------
Boolean
IViewImplementation::handlePaintWindow( IPaintEvent& event )
{
    IPresSpaceHandle hdl = event.presSpaceHandle();
    IRectangle rect = event.rect();

    event.clearBackground( fpView->backgroundColor() );
    fpView->draw( hdl, rect );

    return true;
}


/*******************************************************************************
*                                                                              *
*   Class IView                                                                *
*                                                                              *
*******************************************************************************/

// -----------------------------------------------------------------------------
// IView
// -----------------------------------------------------------------------------
// public constructor
IView::IView( IGUIBundle& bundle ) :
    ICanvas( kDefaultView,
        &bundle.objectFrame(),
        &bundle.frameWindow(),
        IRectangle( IPoint(0,0), bundle.component().contentSize() ),
        ICanvas::defaultStyle() | IWindow::clipChildren ),
    fpBundle( &bundle ),
    fpComponent( &bundle.component() ),
    fpImplementation( new IViewImplementation( this ) ),
    fpViewClient( NULL )
{
}


// -----------------------------------------------------------------------------
// IView
// -----------------------------------------------------------------------------
// copy constructor
IView::IView( const IView& other )
{
    // no copying allowed
    assert( false );
}

// -----------------------------------------------------------------------------
// ~IView
// -----------------------------------------------------------------------------
// destructor
IView::~IView()
{
    delete fpImplementation->fpDropTarget;
    delete fpImplementation;

}

// -----------------------------------------------------------------------------
// initialize
// -----------------------------------------------------------------------------
void
IView::initialize ()
{
    // start event handlers
    fpImplementation->fConnection.handleNotificationsFor( fpComponent->notifier() );
    fpImplementation->fPaintHandler.handleEventsFor( this );
    fpImplementation->fMouseHandler.handleEventsFor( this );

    if ( fpComponent->isContainer() && isDropTarget() ) {
        // enable the view as a drop target
        fpImplementation->fpDropTarget = new IAswDropTarget( IComponentStationery::stationery() );
        fpImplementation->fpDropTarget->Setup( (HWND)(void*)handle(), &IComponentStationery::stationery().dragSource() );
    }

}

// -----------------------------------------------------------------------------
// finalize
// -----------------------------------------------------------------------------
void
IView::finalize ()
{
    fpImplementation->fConnection.stopHandlingNotificationsFor( fpComponent->notifier() );
    fpImplementation->fPaintHandler.stopHandlingEventsFor( this );
    fpImplementation->fMouseHandler.stopHandlingEventsFor( this );

    if( fpImplementation->fpDropTarget )
        fpImplementation->fpDropTarget->Shutdown();

}

// -----------------------------------------------------------------------------
// handleNotification
// -----------------------------------------------------------------------------
// respond to changes in the model
void
IView::handleNotification( const INotificationEvent& )
{

    refresh();
}

// -----------------------------------------------------------------------------
// isDropTarget
// -----------------------------------------------------------------------------
Boolean
IView::isDropTarget() const
{
    return true;
}

// -----------------------------------------------------------------------------
// isDragSource
// -----------------------------------------------------------------------------
Boolean
IView::isDragSource() const
{
    return true;
}

// -----------------------------------------------------------------------------
// model
// -----------------------------------------------------------------------------
IModel*
IView::model() const
{
    return fpComponent->model();
}

// -----------------------------------------------------------------------------
// draw
// -----------------------------------------------------------------------------
// called from the paint handler
// first we draw our native data, then call on all embedded components to draw themselves
void
IView::draw( IPresSpaceHandle& hdl,
             const IRectangle& invalidArea,
             Boolean metaFile )
{

    // draw our contents first so that the embedded components will go on top
    drawContents( hdl, invalidArea, metaFile );

    // Call on any embedded objects to draw themselves
    IEmbedderModel* pEmbModel = NULL;
    ::dynamicCastTo( pEmbModel, model() );

    if( pEmbModel )
    {

        IEmbeddedComponent* pSelectedComp = pEmbModel->selectedComponent();

        const IEmbedderModel::ComponentList& componentList = pEmbModel->componentList();
        ICursor& cursor = *componentList.newCursor();

        // iterate through the items.
        for( cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
        {

            IEmbeddedComponent* pEmbeddedComp = componentList.elementAt( cursor );

            // we want to draw the "selected" component (if any) last
            // so it will appear "on top" of any other windows
            if ( pEmbeddedComp != pSelectedComp )
                pEmbeddedComp->draw( hdl, invalidArea, metaFile );

        }

        delete &cursor;

        // draw the selected component last (if any)
        // (we are not responsible for drawing in-place active objects)
        if( pSelectedComp && (!pSelectedComp->isInPlaceActive() || metaFile) )
        {
            // draw the object
            pSelectedComp->draw( hdl, invalidArea, metaFile );

        } // there is a currently selected component

    } // is an embedder model

}

// -----------------------------------------------------------------------------
// drawContents
// -----------------------------------------------------------------------------
// default implementation does nothing
void
IView::drawContents( IPresSpaceHandle&, const IRectangle&, Boolean )
{
    // Subclasses override this method
}

// -----------------------------------------------------------------------------
// setViewClient
// -----------------------------------------------------------------------------
IView&
IView::setViewClient(IControl* clientControl)
{
   fpViewClient = clientControl;
   if (fpViewClient)   fpViewClient->sizeTo( size() );
   return *this;
}

// -----------------------------------------------------------------------------
// viewClient
// -----------------------------------------------------------------------------
IControl*
IView::viewClient()
{
   return fpViewClient;
}

// -----------------------------------------------------------------------------
// sizeTo
// -----------------------------------------------------------------------------
IView&
IView::sizeTo(const ISize& newSize)
{

    // call the base class to resize
   ICanvas::sizeTo(newSize);

   // size the client control to view
   // do not use newSize, in case it was altered
   if (fpViewClient)   fpViewClient->sizeTo( size() );
   return *this;
}

// -----------------------------------------------------------------------------
// moveSizeTo
// -----------------------------------------------------------------------------
IView&
IView::moveSizeTo(const IRectangle& newRect)
{

   // call the base class to resize
   ICanvas::moveSizeTo(newRect);

   // size the client control to view
   // do not use newRect, in case it was altered
   if (fpViewClient)   fpViewClient->sizeTo( size() );
   return *this;
}


