/*******************************************************************************
* FILE NAME: i32to16.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of functions to allow user messages  *
*   to be sent properly under Win32s                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <windows.h>

#ifndef _ICLCCL_
 #include <iclccl.h>
#endif

#include <itrace.hpp>
#include <iplatfrm.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _I32TO16_CPP_
  #include <ipagetun.h>
#endif

extern "C" {
LRESULT WINAPI S16to32Message(HWND hWnd, unsigned int * Msg,
                              DWORD * wParam, LPARAM * lParam);

LRESULT WINAPI S32to16SendMessage(HWND hWnd, UINT Msg,
                                  WPARAM wParam, LPARAM lParam);

LRESULT WINAPI S32to16PostMessage(HWND hWnd, UINT Msg,
                                  WPARAM wParam, LPARAM lParam);
WORD getFreeCacheIndex(VOID);
}

#define  CACHE_SIZE    1000

struct s32to16 {
   DWORD wparam;
   DWORD lparam;
   UINT  Msg;
};

DWORD     wCacheTableIndex = 0;

// static to set to all  0x00
static s32to16 dw32to16CacheTable[CACHE_SIZE];


/********************************************************************/
/*  Table lookup to resolve message                                 */
/*                                                                  */
/********************************************************************/
LRESULT WINAPI S16to32Message(HWND hWnd, unsigned int * Msg,
                              DWORD * wParam, LPARAM * lParam)
{
  wCacheTableIndex = LOWORD(*wParam);

  *wParam = dw32to16CacheTable[wCacheTableIndex].wparam;
  *lParam = dw32to16CacheTable[wCacheTableIndex].lparam;
  *Msg = dw32to16CacheTable[wCacheTableIndex].Msg;

  ITRACE_DEVELOP(IString("Win32s: S16to32Message: Msg Change, table index is: ")
               +IString(wCacheTableIndex).d2x()
               +IString(" wParam: ")
               +IString(*wParam).d2x()
               +IString(" lParam: ")
               +IString((unsigned long)*lParam).d2x()
               +IString(" Msg: ")
            +IString(*Msg).d2x() );

  dw32to16CacheTable[wCacheTableIndex].wparam = 0;
  dw32to16CacheTable[wCacheTableIndex].lparam = 0;
  dw32to16CacheTable[wCacheTableIndex].Msg = 0;

  return(0);
}


/********************************************************************/
/*  Place message in table                                          */
/*                                                                  */
/*  Note that a check is also done to see if the message is a CCL   */
/*  messages.  Any CCL message destined for a CCL control never     */
/*  makes it into this function.  However, any CCL message destined */
/*  for a non-CCL control (such as Native container) will be        */
/*  processed here.                                                 */
/*                                                                  */
/*  The notification codes send from the non-CCL control            */
/*  (native container) are sent from a handle with SendEvent        */
/*  These messages pass through here, and must be processed in the  */
/*  case where the notification code is a CCL notification code     */
/*  with something other than a windows handle in the lParam        */
/*                                                                  */
/********************************************************************/
LRESULT WINAPI S32to16SendMessage(HWND hWnd, UINT Msg,
                                  WPARAM wParam, LPARAM lParam)
{
  if ( !(IPlatform::isWin32S()) ||
    ((Msg < CCL_MESSAGE_BASE) || (wParam == 0) || (HIWORD(wParam) == 0))
       && !( ((Msg == WM_CONTROL) || (Msg == WM_COMMAND)) &&
             (!IsWindow((HWND)lParam)) &&
             (HIWORD(wParam) >= CCL_NOTIFY_CODE_BASE) &&
             (HIWORD(wParam) <= CCL_NOTIFY_CODE_TOP) )  )
    {
     ITRACE_ALL(IString("S32to16SendMessage: no change to Msg") );

     // This message does not make use of the high-order word of
     // wParam, so it's OK that it will be lost on a SendMessage()
     return(SendMessage(hWnd, Msg, wParam, lParam));
    }
  else
    {
     //return first available slot in table
     wCacheTableIndex = getFreeCacheIndex();

     //Put wParam value in table
     //Put lParam in table (special for WM_COMMAND with CCL notification code)
     //Put Msg in table
     dw32to16CacheTable[wCacheTableIndex].wparam = wParam;
     dw32to16CacheTable[wCacheTableIndex].lparam = lParam;
     dw32to16CacheTable[wCacheTableIndex].Msg = Msg;

     ITRACE_DEVELOP(IString("Win32s: S32to16SendMessage: Msg Change, table index is: ")
               +IString(wCacheTableIndex).d2x()
               +IString(" wParam: ")
               +IString(dw32to16CacheTable[wCacheTableIndex].wparam).d2x()
               +IString(" lParam: ")
               +IString(dw32to16CacheTable[wCacheTableIndex].lparam).d2x()
               +IString(" Msg: ")
               +IString(dw32to16CacheTable[wCacheTableIndex].Msg).d2x() );

     //Create new wParam with table index (LOWORD,HIWORD)
     wParam = MAKEWPARAM(wCacheTableIndex,0);

     //Call API with new wParam containing index, and new Msg
     return(SendMessage(hWnd, IC_UM_32TO16, wParam, lParam));
    }
}

/********************************************************************/
/*  Place message in table                                          */
/*                                                                  */
/*  Note that a check is also done to see if the message is a CCL   */
/*  messages.  Any CCL message destined for a CCL control never     */
/*  makes it into this function.  However, any CCL message destined */
/*  for a non-CCL control (such as Native container) will be        */
/*  processed here.                                                 */
/*                                                                  */
/*  The notification codes send from the non-CCL control            */
/*  (native container) are sent from a handle with SendEvent        */
/*  These messages pass through here, and must be processed in the  */
/*  case where the notification code is a CCL notification code     */
/*  with something other than a windows handle in the lParam        */
/*                                                                  */
/********************************************************************/
LRESULT WINAPI S32to16PostMessage(HWND hWnd, UINT Msg,
                                  WPARAM wParam, LPARAM lParam)
{
  if ( !(IPlatform::isWin32S()) ||
    ((Msg < CCL_MESSAGE_BASE) || (wParam == 0) || (HIWORD(wParam) == 0))
       && !( ((Msg == WM_CONTROL) || (Msg == WM_COMMAND)) &&
             (!IsWindow((HWND)lParam)) &&
             (HIWORD(wParam) >= CCL_NOTIFY_CODE_BASE) &&
             (HIWORD(wParam) <= CCL_NOTIFY_CODE_TOP) )  )
    {
     ITRACE_ALL(IString("S32to16PostMessage: no change") );

     // This message does not make use of the high-order word of
     // wParam, so it's OK that it will be lost on a SendMessage()
     return(PostMessage(hWnd, Msg, wParam, lParam));
    }
  else
    {
     //return first available slot in table
     wCacheTableIndex = getFreeCacheIndex();

     //Put wParam value in table
     //Put lParam in table (special for WM_COMMAND with CCL notification code)
     //Put Msg in table
     dw32to16CacheTable[wCacheTableIndex].wparam = wParam;
     dw32to16CacheTable[wCacheTableIndex].lparam = lParam;
     dw32to16CacheTable[wCacheTableIndex].Msg = Msg;

     ITRACE_DEVELOP(IString("Win32s: S32to16PostMessage: Msg Change, table index is: ")
               +IString(wCacheTableIndex).d2x()
               +IString(" wParam: ")
               +IString(dw32to16CacheTable[wCacheTableIndex].wparam).d2x()
               +IString(" lParam: ")
               +IString(dw32to16CacheTable[wCacheTableIndex].lparam).d2x()
               +IString(" Msg: ")
               +IString(dw32to16CacheTable[wCacheTableIndex].Msg).d2x() );

     //Create new wParam with table index (LOWORD,HIWORD)
     wParam = MAKEWPARAM(wCacheTableIndex,0);

     //Call API with new wParam containing index
     return(PostMessage(hWnd, IC_UM_32TO16, wParam, lParam));
    }
}


WORD getFreeCacheIndex(VOID)
{
  //Search the table until you find an empty entry.
  //An entry with a 0 value is considered empty, because a 0 value
  //would never be put into the table.
  WORD i;

  for (i=0; ((i <= CACHE_SIZE) && (dw32to16CacheTable[i].wparam != 0)); i++) ;

  return(i);
}
