/*******************************************************************************
* FILE NAME: iapp.cpp                                                          *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IApplication                                                             *
*     ICurrentApplication                                                      *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_DOSPROCESS
  #define INCL_DOSERRORS
  #include <iwindefs.h>
  #include <stdlib.h>
}

#include <iapp.hpp>

#ifndef _ITHREAD_
  #include <ithread.hpp>
#endif

#ifndef _ISTRING_
  #include <istring.hpp>
#endif

#ifndef _IRESLIB_
  #include <ireslib.hpp>
#endif

#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _IPLATFRM_
  #include <iplatfrm.hpp>
#endif

#ifndef _ITRACE_
  #include <itrace.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IAPP_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
  Static objects:
    IApplication::pCurrent - initially zero.
------------------------------------------------------------------------------*/
ICurrentApplication
 *IApplication::pCurrent = 0;

static const char
 resLibVar[]  = "ICLUI RESLIB",
#ifdef IC_WIN
 defaultLib[] = "CPPWOR3U";
#else
 defaultLib[] = "CPPOOR3U";
#endif

// The following are added to stop the compiler statically
// generating them.
IApplication :: ~IApplication ( ) {}
ICurrentApplication :: ~ICurrentApplication ( ) {
   /* delete pCurrentApplicationData when it is defined. */}


/*------------------------------------------------------------------------------
  asString - Returns "IApplication(id)".
------------------------------------------------------------------------------*/
IString IApplication :: asString ( ) const
{
  return "IApplication(" + IString( this->pid.asString() ) + ")";
}

/*------------------------------------------------------------------------------
  asDebugInfo - Returns "IApplication(IVBase(@xxxx),id=xx)".
------------------------------------------------------------------------------*/
IString IApplication :: asDebugInfo ( ) const
{
  IString result( "IApplication(" );
  result += this->Inherited::asDebugInfo();
  result += ",id=";
  result += IString( this->pid.asString() );
  result += ")";
  return( result );
}

/*------------------------------------------------------------------------------
   id - return the pid data member.
------------------------------------------------------------------------------*/
IProcessId IApplication :: id ( ) const
{
  return( pid );
}

/*------------------------------------------------------------------------------
  setId - set the pid data member to the argument value.
------------------------------------------------------------------------------*/
IApplication &IApplication :: setId ( const IProcessId &id )
{
  this->pid = id;
  return( *this );
}

/*----------------------------------------------------------------------------
  setPriority - Issue DosSetPriority using specified arguments.
----------------------------------------------------------------------------*/
IApplication &IApplication :: setPriority ( PriorityClass priorityClass,
                                            long          priorityLevel,
                                            Boolean       setDescendents )
{
  IMODTRACE_DEVELOP("IApplication::setPriority");

#ifdef IC_WIN
  if (IPlatform::isWin32S())
  {
    ITRACE_DEVELOP("Win32s:  setPriority has no effect");
    return( *this );
  }
  else
  {
    ITRACE_DEVELOP("Win32:  setPriority LEVEL ARGUMENT has no effect");
  }
#endif

  // default to current priorityClass if no change requested...
  if ( priorityClass == noChange )
    priorityClass = IThread::current().priorityClass();

#ifdef IC_PM
  unsigned long
    pclass;
  switch ( priorityClass ) {
    case idleTime:
      pclass = PRTYC_IDLETIME;
      break;
    case regular:
      pclass = PRTYC_REGULAR;
      break;
    case timeCritical:
      pclass = PRTYC_TIMECRITICAL;
      break;
    case foregroundServer:
      pclass = PRTYC_FOREGROUNDSERVER;
      break;
  }

  unsigned long
    pscope = ( setDescendents ) ? PRTYS_PROCESSTREE : PRTYS_PROCESS,
    rc     = DosSetPriority( pscope, pclass, priorityLevel, this->id() );

  if (rc != NO_ERROR)
  {
    ITHROWSYSTEMERROR( rc,
                       "DosSetPriority",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif
#ifdef IC_WIN
//------------------------------------------------------------
// DEFECT 8421/7546 : correct thread prioritie handling on NT
//------------------------------------------------------------
  int pclass(0);
  switch ( priorityClass ) {
    case idleTime:
      pclass = IDLE_PRIORITY_CLASS;
      break;
    case regular:
      pclass = NORMAL_PRIORITY_CLASS;
      break;
    case timeCritical:
      pclass = REALTIME_PRIORITY_CLASS	;
      break;
    case foregroundServer:
      pclass = HIGH_PRIORITY_CLASS;
      break;
  }

  if ( !SetPriorityClass( GetCurrentProcess(), pclass) )
  {
    ITHROWGUIERROR2( "SetPriorityClass",
                     IBaseErrorInfo::accessError,
                     IException::recoverable );
  }
//------------------------------------------------------------
#endif

  return( *this );
}

/*----------------------------------------------------------------------------
  adjustPriority - Issue DosSetPriority to adjust the priority level.
----------------------------------------------------------------------------*/
IApplication &IApplication :: adjustPriority ( long    adjustment,
                                               Boolean setDescendents )
{
  IMODTRACE_DEVELOP("IApplication::adjustPriority");

//------------------------------------------------------------
// DEFECT 8421/7546 : correct thread prioritie handling on NT
//------------------------------------------------------------
#ifdef IC_WIN
  ITRACE_DEVELOP("Windows:  adjustPriority has no effect");
  return( *this );
#endif
//------------------------------------------------------------

#ifdef IC_PM
  unsigned long
    pscope = ( setDescendents ) ? PRTYS_PROCESSTREE : PRTYS_PROCESS,
    rc     = DosSetPriority( pscope, PRTYC_NOCHANGE, adjustment, this->id() );
  if (rc != NO_ERROR)
  {
    ITHROWSYSTEMERROR( rc,
                       "DosSetPriority",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
#endif //IC_PM

  return( *this );
}

/*----------------------------------------------------------------------------
 currentPID -  Return current process's id from DosGetInfoBlocks.
----------------------------------------------------------------------------*/
IProcessId IApplication :: currentPID ( )
{
#ifdef IC_PM
  PPIB ppib;
  PTIB ptib;
  DosGetInfoBlocks( &ptib, &ppib );
  return( IProcessId( ppib->pib_ulpid ) );
#endif
#ifdef IC_WIN
  return( IProcessId( (void*)GetCurrentProcessId() ) );
#endif
}


/*------------------------------------------------------------------------------
  Construct current application with pib/pid set for the current process.
------------------------------------------------------------------------------*/
ICurrentApplication :: ICurrentApplication ( )
  : IApplication ( 0 )
  , userLib      ( 0 )
  , icluiLib     ( 0 )
  , exeResource  ( 0 )
  , argCount     ( 0 )
  , argValue     ( 0 )
  , pCurrentApplicationData ( 0 )
#ifdef IC_PM
  , pPIB         ( 0 )
#endif
{
  IMODTRACE_DEVELOP("ICurrentApplication::ICurrentApplication");

#ifdef IC_PM
  this->setId( this->pib().pib_ulpid );
#endif
#ifdef IC_WIN
  this->setId( this->currentPID() );
#endif
}

#ifdef IC_PM
/*------------------------------------------------------------------------------
  pib - If not set previously, get pointer to PIB for this process.  It is
  "cached" in the ICurrentApplication object.
------------------------------------------------------------------------------*/
struct pib_s &ICurrentApplication :: pib ( )
{
  if ( !this->pPIB )
  {
    PTIB ptib;
    DosGetInfoBlocks( &ptib, &this->pPIB );
  }

  return( *this->pPIB );
}
#endif

/*------------------------------------------------------------------------------
  resourceLibrary - Return ICLUI resource library data member.
------------------------------------------------------------------------------*/
IResourceLibrary &ICurrentApplication :: resourceLibrary ( ) const
{
  IMODTRACE_DEVELOP("ICurrentApplication::resourceLibrary");

  if ( !this->icluiLib && !this->exeResource )
  {
    const char* pLibName = getenv ( resLibVar );
    if ( !pLibName )
      pLibName = defaultLib;

    ((ICurrentApplication*)this)->setResourceLibrary( pLibName );
  }

  return( *(this->icluiLib) );
}

/*------------------------------------------------------------------------------
  setResourceLibrary - Delete existing resource library (if there is one)
  and create a new one from input .dll name.
------------------------------------------------------------------------------*/
ICurrentApplication &ICurrentApplication ::
                             setResourceLibrary ( const char* resLibName )
{
  IMODTRACE_DEVELOP("ICurrentApplication::setResourceLibrary");

  if (this->icluiLib)
    delete this->icluiLib;

  if (resLibName)
  {
    this->icluiLib = new IDynamicLinkLibrary( resLibName );
    this->exeResource = false;
  }
  else
  {
    this->icluiLib = new IResourceLibrary;
    this->exeResource = true;
  }

  return( *this );
}

/*------------------------------------------------------------------------------
  userResourceLibrary - Return user resource library data member.
------------------------------------------------------------------------------*/
IResourceLibrary &ICurrentApplication :: userResourceLibrary ( ) const
{
  IMODTRACE_DEVELOP("ICurrentApplication::userResourceLibrary");

  if ( !this->userLib )
    ((ICurrentApplication*)this)->setUserResourceLibrary( 0 );

  return( *(this->userLib) );
}

/*------------------------------------------------------------------------------
  setUserResourceLibrary - Attempt to load the new resource library.
  If it succeeds, then free the old one if there was one and reset
  the userLib data member to point to the new one.
------------------------------------------------------------------------------*/
ICurrentApplication &ICurrentApplication ::
                             setUserResourceLibrary ( const char* resLibName )
{
  IMODTRACE_DEVELOP("ICurrentApplication::setUserResourceLibrary");

  IResourceLibrary
   *newLib = 0,
   *oldLib = this->userLib;

  if (resLibName)
    newLib = new IDynamicLinkLibrary( resLibName );
  else
    newLib = new IResourceLibrary;

  this->userLib = newLib;
  if (oldLib)
    delete oldLib;

  return( *this );
}

/*------------------------------------------------------------------------------
  argc - Return arg count data member.
------------------------------------------------------------------------------*/
int ICurrentApplication :: argc ( ) const
{
  return( this->argCount );
}

/*------------------------------------------------------------------------------
  argv - Return argument array data member.
------------------------------------------------------------------------------*/
IString ICurrentApplication :: argv ( int argNo ) const
{
  IString result;
  if (argNo < argCount)
    result = this->argValue[ argNo ];

  return( result );
}

/*------------------------------------------------------------------------------
  setArgs - If current data member isn't empty, delete old array of args.
  Allocate new array data member and initialize each element from input
  char* array.
------------------------------------------------------------------------------*/
ICurrentApplication &ICurrentApplication ::
                             setArgs ( int argc, const char* const argv[] )
{
  if (this->argCount)
  {
    // free the old array
    delete [] this->argValue;
  }
  if (argc)
  {
    // Allocate new array of pointers and allocate buffers for args.
    this->argValue = new IString[ argc ];
    this->argCount = 0;
    while (this->argCount < argc)
    {
      this->argValue[ this->argCount ] = argv[ this->argCount ];
      this->argCount++;
    }
  }

  return( *this );
}

/*------------------------------------------------------------------------------
  run - Process PM messages on current thread.
------------------------------------------------------------------------------*/
ICurrentApplication &ICurrentApplication :: run ( )
{
  IThread::current().processMsgs();
  return( *this );
}

/*------------------------------------------------------------------------------
  exit - Exit current thread (which will exit application in a single-threaded
  app).
------------------------------------------------------------------------------*/
ICurrentApplication &ICurrentApplication :: exit ( )
{
  IThread::current().exit(0);
  return( *this );
}

/*------------------------------------------------------------------------------
  current - return pointer to current application.
------------------------------------------------------------------------------*/
ICurrentApplication &IApplication :: current ( )
{
  if ( ! IApplication::pCurrent )
    IApplication::pCurrent = new ICurrentApplication;

  return( *IApplication::pCurrent );
}

/*------------------------------------------------------------------------------
  Constructor, which only initializes the pid.
------------------------------------------------------------------------------*/
IApplication :: IApplication ( const IProcessId &id )
  : pid( id )
{
  IMODTRACE_DEVELOP("IApplication::IApplication");
}

/*------------------------------------------------------------------------------
  asDebugInfo - Generate string with current application contents.
------------------------------------------------------------------------------*/
IString ICurrentApplication :: asDebugInfo ( ) const
{
  IString result( "ICurrentApplication(" );
  result += this->Inherited::asDebugInfo();
  result += ",icluiLib=" + IString( (unsigned long) this->icluiLib );
  result += ",userLib=" + IString( (unsigned long) this->userLib );
  result += ",argCount=" + IString( this->argCount );
  result += ",argValue=" + IString( (unsigned long) this->argValue );
  result += "->[";
  IString comma;
  for ( int n = 0; n < this->argCount; n++ )
  {
    result += comma;
    result += this->argValue[ n ].asDebugInfo();
    comma = ",";
  }

  result += "]";
#ifdef IC_PM
  result += ",pPIB=" + IString( (unsigned long) this->pPIB );
#endif
  result += ")";
  return( result );
}
