/*******************************************************************************
* FILE NAME: ibidiset.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions used for the    *
*   bidirectional support.                                                     *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#define INCL_WINCOUNTRY
#define INCL_WINHOOKS
extern "C" {
  #include <iwindefs.h>
  #include <stdlib.h>
}

#include <ibidiset.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <igrafctx.hpp>
#include <ireslib.hpp>
#include <itbpriv.hpp>
#include <iwindow.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IBIDISET_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| The following are defined as part of the OS/2 bidectional-support toolkit    |
| (pmbidi.h), and are included here for use in portable code.                  |
------------------------------------------------------------------------------*/
#define WM_SETBIDIATTR          0x0BD0
#define WM_QUERYBIDIATTR        0x0BD1
#define BDAM_INIT               0x80000000
#define BDAM_LEVEL              0x70000000
#define BDAM_TEXTTYPE           0x01000000
#define BDAM_TEXT_ORIENTATION   0x00030000
#define BDAM_WND_ORIENTATION    0x00100000
#define BDAM_NUMERALS           0x00003000
#define BDAM_SYM_SWAP           0x00000100
#define BDAM_WORD_BREAK         0x00000200
#define BDAM_TEXT_SHAPE         0x000000FF
#define BDA_INIT                0x80000000
#define BDA_LEVEL               0x30000000
#define BDA_TEXTTYPE_VISUAL     0x00000000
#define BDA_TEXTTYPE_IMPLICIT   0x01000000
#define BDA_WND_ORIENT_LTR      0x00000000
#define BDA_WND_ORIENT_RTL      0x00100000
#define BDA_TEXT_ORIENT_LTR     0x00000000
#define BDA_TEXT_ORIENT_RTL     0x00010000
#define BDA_TEXT_ORIENT_DYNAMIC 0x00020000
#define BDA_TEXT_ORIENT_CONTEXT 0x00020000
#define BDA_NUMERALS_NOMINAL    0x00000000
#define BDA_NUMERALS_PASSTHRU   0x00001000
#define BDA_NUMERALS_NATIONAL   0x00002000
#define BDA_NUMERALS_CONTEXTUAL 0x00003000
#define BDA_SYM_SWAP_OFF        0x00000000
#define BDA_SYM_SWAP_ON         0x00000100
#define BDA_WORD_BREAK_OFF      0x00000000
#define BDA_WORD_BREAK_ON       0x00000200
#define BDA_TEXT_DISPLAY_SHAPED 0x00000000
#define BDA_TEXT_SAVE_SHAPED    0x00000001
#define BDA_TEXT_NOMINAL        0x00000010
#define BDA_TEXT_INITIAL        0x00000011
#define BDA_TEXT_MIDDLE         0x00000012
#define BDA_TEXT_FINAL          0x00000013
#define BDA_TEXT_ISOLATED       0x00000014
#define LI_BD_WND_ATTR          0x00000020
#define LIF_CHILD_INHERIT       0x00000002
#define LIF_WND_REFRESH         0x00000004


/*------------------------------------------------------------------------------
| class IBidiData                                                              |
|                                                                              |
| Private class used for implementation of bidi support.                       |
| Only a single instance of this class is needed per process, which is         |
| managed by IBidiSettingsData.                                                |
------------------------------------------------------------------------------*/
class IBidiData : public IBase {
typedef IBase
  Inherited;
public:
  IBidiData ( );
Boolean
  support;
unsigned long
  codepage;
}; // IBidiData

IBidiData::IBidiData ( )
  : support( false ),
    codepage( 0 )
{ }


/*------------------------------------------------------------------------------
| class IBidiSettingsData                                                      |
|                                                                              |
| Private data class for IBidiSettings.                                        |
------------------------------------------------------------------------------*/
class IBidiSettingsData : public IBase {
typedef IBase
  Inherited;
public:
  IBidiSettingsData ( );
IBidiSettingsData
 &operator=         ( const IBidiSettingsData& data );
Boolean
  operator==        ( const IBidiSettingsData& data ) const;
static IBidiData
 *bidi;
unsigned long
  data;
#ifndef IC_PM
unsigned long
  fExplicitlySetFlags;
enum ESetFlags {
  kTextType             = 1,
  kTextOrientation      = 2,
  kWindowLayout         = 4,
  kNumeralDisplay       = 8,
  kSymmetricSwapping    = 16,
  kWordByWordReordering = 32,
  kTextShape            = 64
  };
#endif
}; // IBidiSettingsData

IBidiSettingsData::IBidiSettingsData ( )
  : data( 0 )
#ifndef IC_PM
  , fExplicitlySetFlags( 0 )
#endif
{ }

IBidiSettingsData&
  IBidiSettingsData::operator= ( const IBidiSettingsData& data )
{
  if ( this != &data )
  {  // Not copying self.
     this->data = data.data;
#ifndef IC_PM
     this->fExplicitlySetFlags = data.fExplicitlySetFlags;
#endif
  }
  return *this;
}

Boolean IBidiSettingsData::operator== ( const IBidiSettingsData& data ) const
{
  return ( this->data == data.data
#ifndef IC_PM
           &&
           this->fExplicitlySetFlags == data.fExplicitlySetFlags
#endif
         );
}

/*------------------------------------------------------------------------------
| static data                                                                  |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
#ifndef IC_PM
IBidiSettings
  IBidiSettings::fgApplicationDefaults;
#endif
IBidiData
 *IBidiSettingsData::bidi( 0 );
#pragma data_seg()

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IBidiSettings::isBidiSupported ()
{
#ifdef IC_WIN
  if ( ! IBidiSettingsData::bidi )
  {
     IBidiSettingsData::bidi = new IBidiData;
     IBidiSettingsData::bidi->support = false;

     // Check if the system is mideast enabled.
     if ( GetSystemMetrics( SM_MIDEASTENABLED ) )
     {
        IBidiSettingsData::bidi->support = true;
     }
#ifdef IC_DEVELOP
     else      // Code for development testing.
     {
        const char* bidiVar = getenv( "ICLUI_BIDI" );
        if ( bidiVar )
        {
           IBidiSettingsData::bidi->support = true;
        }
     }
#endif
  }
  return IBidiSettingsData::bidi->support;
#endif // IC_WIN

}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
void IBidiSettings::setApplicationDefaults
                      ( const IBidiSettings& applicationValues )
{
  IASSERTSTATE( IBidiSettings::isBidiSupported() == true );

#ifdef IC_PM
  IBidiSettingsData::bidi->fSetInfo( 0, LI_BD_PROCESS_ATTR,
                                     applicationValues.fBidiSettingsData->data |
                                       BDA_INIT | BDA_LEVEL,
                                       BDAM_INIT | BDAM_LEVEL |
                                       BDAM_TEXTTYPE | BDAM_TEXT_ORIENTATION |
                                       BDAM_WND_ORIENTATION | BDAM_NUMERALS |
                                       BDAM_SYM_SWAP | BDAM_WORD_BREAK |
                                       BDAM_TEXT_SHAPE,
                                     0, 0 );
#else
  fgApplicationDefaults = applicationValues;
#endif
}

IBidiSettings IBidiSettings::applicationDefaults ( )
{
  IASSERTSTATE( IBidiSettings::isBidiSupported() == true );

#ifdef IC_PM
  IBidiSettings
    applicationSettings;
  applicationSettings.fBidiSettingsData->data =
      IBidiSettingsData::bidi->fQueryInfo( 0, LI_BD_PROCESS_ATTR, 0, 0 );
  return applicationSettings;
#else
  return fgApplicationDefaults;
#endif
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::IBidiSettings ( const IWindow& window )
  : fBidiSettingsData( 0 )
{
  this->initialize( window.handle() );
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::IBidiSettings ( const IWindowHandle windowHandle )
  : fBidiSettingsData( 0 )
{
  this->initialize( windowHandle );
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::IBidiSettings ( const IGraphicContext& graphicContext )
  : fBidiSettingsData( 0 )
{
  this->initialize( graphicContext.handle() );
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::IBidiSettings ( const IPresSpaceHandle presSpace )
  : fBidiSettingsData( 0 )
{
  this->initialize( presSpace );
}

/*------------------------------------------------------------------------------
| Default constructor (used for application-default functions only, for now).  |
------------------------------------------------------------------------------*/
IBidiSettings::IBidiSettings ( )
  : fBidiSettingsData( new IBidiSettingsData )
{ }

/*------------------------------------------------------------------------------
| Copy constructor.                                                            |
------------------------------------------------------------------------------*/
IBidiSettings::IBidiSettings ( const IBidiSettings& settings )
  : fBidiSettingsData( new IBidiSettingsData )
{
  *( this->fBidiSettingsData ) = *( settings.fBidiSettingsData );
}

/*------------------------------------------------------------------------------
| Assignment operator.                                                         |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::operator= ( const IBidiSettings& settings )
{
  *( this->fBidiSettingsData ) = *( settings.fBidiSettingsData );
  return *this;
}

/*------------------------------------------------------------------------------
| Comparison operator.                                                         |
------------------------------------------------------------------------------*/
Boolean IBidiSettings::operator== ( const IBidiSettings& settings ) const
{
  return ( *( this->fBidiSettingsData ) ==
                       *( settings.fBidiSettingsData ) );
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::~IBidiSettings ( )
{
  if ( fBidiSettingsData )
  {
     delete fBidiSettingsData;
  }
}

/*------------------------------------------------------------------------------
| Initialization from an IWindowHandle.                                        |
------------------------------------------------------------------------------*/
void IBidiSettings::initialize ( const IWindowHandle& windowHandle )
{
  IASSERTSTATE( IBidiSettings::isBidiSupported() == true );

  fBidiSettingsData = new IBidiSettingsData;
  // Check for RTL extended styles, and set flags in fBidiSettingsData->data
  // accordingly.
  unsigned long
    style = GetWindowLong( windowHandle, GWL_EXSTYLE );
  ( *this )
    .setTextOrientation( ( style & WS_EX_RTLREADING ) ?
                            IBidiSettings::textRightToLeft :
                            IBidiSettings::textLeftToRight );
  ( *this )
    .setWindowLayout( ( style & WS_EX_RIGHT ) ?
                         IBidiSettings::layoutRightToLeft :
                         IBidiSettings::layoutLeftToRight );
  ( *this )
   .setTextType( IBidiSettings::implicit )
   .setNumeralDisplay( IBidiSettings::contextual )
   .enableSymmetricSwapping( false )
   .enableWordByWordReordering( false );
}

/*------------------------------------------------------------------------------
| Initialization from an IPresSpaceHandle.                                     |
------------------------------------------------------------------------------*/
void IBidiSettings::initialize ( const IPresSpaceHandle& handle )
{
  IASSERTSTATE( IBidiSettings::isBidiSupported() == true );

  fBidiSettingsData = new IBidiSettingsData;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::apply ( const IWindow& window,
                                      Boolean childInherit,
                                      Boolean refresh )
{
  if ( window.isFrameWindow() )
  {
     this->applyToFrameWindow( (IWindow&)window, childInherit, refresh );
  }
  else
  {
     IWindowHandle
       nonConstWindowHandle = window.handle();
     this->applyToWindow( nonConstWindowHandle, childInherit, refresh );
  }

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
void IBidiSettings::applyToFrameWindow ( IWindow& frameWindow,
                                         Boolean  childInherit,
                                         Boolean  refresh ) const
{
  //-----------------------------------------------------------------
  // Let applyToWindow do most of the work.
  //-----------------------------------------------------------------
  IWindowHandle
    nonConstWindowHandle = frameWindow.handle();
  this->applyToWindow( nonConstWindowHandle, childInherit, refresh );

  //-----------------------------------------------------------------
  // If the window is a frame, we need to update any toolbars as well
  // in case they are floating or hidden.  We do this by sending a
  // private message to the frame and letting the tool bar handler
  // (attached to the frame automatically by the tool bar) do the
  // work of updating BIDI attributes.
  //-----------------------------------------------------------------
  if ( childInherit )
  {
     frameWindow.sendEvent( IC_UM_TOOLBAR, IC_UM_TBAR_BIDI_CHANGE );
  }
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::apply ( const IWindowHandle windowHandle,
                                      Boolean childInherit,
                                      Boolean refresh )
{
  IWindow
   *bidiWindow = IWindow::windowWithHandle( windowHandle );
  if ( bidiWindow  &&  bidiWindow->isFrameWindow() )
  {
     this->applyToFrameWindow( *bidiWindow, childInherit, refresh );
  }
  else
  {
     IWindowHandle
       nonConstWindowHandle = windowHandle;
     this->applyToWindow( nonConstWindowHandle, childInherit, refresh );
  }

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
void IBidiSettings::applyToWindow ( IWindowHandle& windowHandle,
                                    Boolean        childInherit,
                                    Boolean        refresh ) const
{
  // Window orientation and text direction are the only attributes
  // we can control.
  long
    extendedStyle = GetWindowLong( windowHandle, GWL_EXSTYLE );

  if ( this->windowLayout() == IBidiSettings::layoutLeftToRight )
  {  // Need left-to-right window orientation.
     extendedStyle &= (unsigned long) ~( WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR );
     extendedStyle |= ( WS_EX_LEFT | WS_EX_RIGHTSCROLLBAR );
  }
  else
  {  // Need right-to-left window orientation.
     extendedStyle &= (unsigned long) ~( WS_EX_LEFT | WS_EX_RIGHTSCROLLBAR );
     extendedStyle |= ( WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR );
  }

  if ( this->textOrientation() == IBidiSettings::textLeftToRight )
  {  // Need left-to-right text.
     extendedStyle &= (unsigned long) ~WS_EX_RTLREADING;
     extendedStyle |= WS_EX_LTRREADING;
  }
  else
  {  // Need right-to-left text.
     extendedStyle &= (unsigned long) ~WS_EX_LTRREADING;
     extendedStyle |= WS_EX_RTLREADING;
  }

  SetWindowLong( windowHandle, GWL_EXSTYLE, extendedStyle );

  IWindow
   *bidiWindow = IWindow::windowWithHandle( windowHandle );
  if ( bidiWindow )
  {
     bidiWindow->setLayoutDistorted( IWindow::layoutChanged, 0 );

     // Allow for control-specific processing to support the change.
     bidiWindow->sendEvent( IC_UM_BIDI_CHANGE );
  }

  if ( refresh )
  {
     if ( bidiWindow )
     {
        bidiWindow->refresh();
     }
  }

  if ( childInherit )
  {
     if ( bidiWindow )
     {
        // Propagate new bidi attributes to all child windows.
        IBidiSettings
          *nonConstBidiSettings = (IBidiSettings*) this;
        IWindow::ChildCursor
          childCursor( *bidiWindow );
        for ( childCursor.setToFirst();
              childCursor.isValid();
              childCursor.setToNext() )
        {
           IWindowHandle
             childWindow( bidiWindow->childAt( childCursor ) );
           ( *nonConstBidiSettings )
            .apply( childWindow, childInherit, refresh );
        }
     }
  }
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::apply ( const IGraphicContext& graphicContext )
{
  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::apply ( const IPresSpaceHandle presSpace )
{
  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::setTextType
                 ( IBidiSettings::BidiTextType textType )
{
  switch ( textType )
  {
     case visual:
       fBidiSettingsData->data &= (unsigned long) ~BDA_TEXTTYPE_IMPLICIT;
       fBidiSettingsData->data |= BDA_TEXTTYPE_VISUAL;
       break;
     case implicit:
       fBidiSettingsData->data &= (unsigned long) ~BDA_TEXTTYPE_VISUAL;
       fBidiSettingsData->data |= BDA_TEXTTYPE_IMPLICIT;
       break;
     default:
       ITHROWLIBRARYERROR( IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable );
       break;
  }

#ifndef IC_PM
  fBidiSettingsData->fExplicitlySetFlags |= IBidiSettingsData::kTextType;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::setTextOrientation
                 ( IBidiSettings::BidiTextOrientation textOrientation )
{
  switch ( textOrientation )
  {
     case textLeftToRight:
       fBidiSettingsData->data &= (unsigned long)
                            ~( BDA_TEXT_ORIENT_RTL | BDA_TEXT_ORIENT_CONTEXT );
       fBidiSettingsData->data |= BDA_TEXT_ORIENT_LTR;
       break;
     case textRightToLeft:
       fBidiSettingsData->data &= (unsigned long)
                            ~( BDA_TEXT_ORIENT_LTR | BDA_TEXT_ORIENT_CONTEXT );
       fBidiSettingsData->data |= BDA_TEXT_ORIENT_RTL;
       break;
     case textContextual:
       fBidiSettingsData->data &= (unsigned long)
                            ~( BDA_TEXT_ORIENT_LTR | BDA_TEXT_ORIENT_RTL );
       fBidiSettingsData->data |= BDA_TEXT_ORIENT_CONTEXT;
       break;
     default:
       ITHROWLIBRARYERROR( IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable );
       break;
  }

#ifndef IC_PM
  fBidiSettingsData->fExplicitlySetFlags |=
                        IBidiSettingsData::kTextOrientation;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::setWindowLayout
                 ( IBidiSettings::BidiLayout windowLayout )
{
  switch ( windowLayout )
  {
     case layoutLeftToRight:
       fBidiSettingsData->data &= (unsigned long) ~BDA_WND_ORIENT_RTL;
       fBidiSettingsData->data |= BDA_WND_ORIENT_LTR;
       break;
     case layoutRightToLeft:
       fBidiSettingsData->data &= (unsigned long) ~BDA_WND_ORIENT_LTR;
       fBidiSettingsData->data |= BDA_WND_ORIENT_RTL;
       break;
     default:
       ITHROWLIBRARYERROR( IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable );
       break;
  }

#ifndef IC_PM
  fBidiSettingsData->fExplicitlySetFlags |=
                        IBidiSettingsData::kWindowLayout;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::setNumeralDisplay
                 ( IBidiSettings::BidiNumeralType numeralDisplay )
{
  fBidiSettingsData->data &= (unsigned long)
                          ~( BDA_NUMERALS_NOMINAL | BDA_NUMERALS_PASSTHRU |
                             BDA_NUMERALS_NATIONAL | BDA_NUMERALS_CONTEXTUAL );

  switch ( numeralDisplay )
  {
     case arabic:
       fBidiSettingsData->data |= BDA_NUMERALS_NOMINAL;
       break;
     case asStored:
       fBidiSettingsData->data |= BDA_NUMERALS_PASSTHRU;
       break;
     case national:
       fBidiSettingsData->data |= BDA_NUMERALS_NATIONAL;
       break;
     case contextual:
       fBidiSettingsData->data |= BDA_NUMERALS_CONTEXTUAL;
       break;
     default:
       ITHROWLIBRARYERROR( IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable );
       break;
  }

#ifndef IC_PM
  fBidiSettingsData->fExplicitlySetFlags |=
                        IBidiSettingsData::kNumeralDisplay;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::enableSymmetricSwapping ( Boolean enable )
{
  if ( enable )
  {
     fBidiSettingsData->data &= (unsigned long) ~BDA_SYM_SWAP_OFF;
     fBidiSettingsData->data |= BDA_SYM_SWAP_ON;
  }
  else
  {
     fBidiSettingsData->data &= (unsigned long) ~BDA_SYM_SWAP_ON;
     fBidiSettingsData->data |= BDA_SYM_SWAP_OFF;
  }

#ifndef IC_PM
  fBidiSettingsData->fExplicitlySetFlags |=
                        IBidiSettingsData::kSymmetricSwapping;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::disableSymmetricSwapping ( )
{
  return this->enableSymmetricSwapping( false );
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::enableWordByWordReordering ( Boolean enable )
{
  if ( enable )
  {
     fBidiSettingsData->data &= (unsigned long) ~BDA_WORD_BREAK_OFF;
     fBidiSettingsData->data |= BDA_WORD_BREAK_ON;
  }
  else
  {
     fBidiSettingsData->data &= (unsigned long ) ~BDA_WORD_BREAK_ON;
     fBidiSettingsData->data |= BDA_WORD_BREAK_OFF;
  }

#ifndef IC_PM
  fBidiSettingsData->fExplicitlySetFlags |=
                        IBidiSettingsData::kWordByWordReordering;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::disableWordByWordReordering ( )
{
  return this->enableWordByWordReordering( false );
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings& IBidiSettings::setTextShape
                 ( IBidiSettings::BidiTextShape textShape )
{
  fBidiSettingsData->data &= (unsigned long)
                          ~( BDA_TEXT_DISPLAY_SHAPED | BDA_TEXT_SAVE_SHAPED |
                             BDA_TEXT_NOMINAL | BDA_TEXT_MIDDLE | BDA_TEXT_FINAL |
                             BDA_TEXT_ISOLATED );

  switch ( textShape )
  {
     case displayShaped:
       fBidiSettingsData->data |= BDA_TEXT_DISPLAY_SHAPED;
       break;
     case saveShaped:
       fBidiSettingsData->data |= BDA_TEXT_SAVE_SHAPED;
       break;
     case nominalShape:
       fBidiSettingsData->data |= BDA_TEXT_NOMINAL;
       break;
     case initialShape:
       fBidiSettingsData->data |= BDA_TEXT_INITIAL;
       break;
     case middleShape:
       fBidiSettingsData->data |= BDA_TEXT_MIDDLE;
       break;
     case finalShape:
       fBidiSettingsData->data |= BDA_TEXT_FINAL;
       break;
     case isolatedShape:
       fBidiSettingsData->data |= BDA_TEXT_ISOLATED;
       break;
     default:
       ITHROWLIBRARYERROR( IC_INVALIDENUMVALUE,
                           IBaseErrorInfo::invalidParameter,
                           IException::recoverable );
       break;
  }

#ifndef IC_PM
  fBidiSettingsData->fExplicitlySetFlags |= IBidiSettingsData::kTextShape;
#endif

  return *this;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::BidiTextType IBidiSettings::textType ( ) const
{
#ifndef IC_PM
  if ( ! ( fBidiSettingsData->fExplicitlySetFlags &
                        IBidiSettingsData::kTextType ) )
  {
     return implicit;
  }
  else
#endif
  if ( fBidiSettingsData->data & BDA_TEXTTYPE_IMPLICIT )
  {
     return implicit;
  }
  return visual;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::BidiTextOrientation IBidiSettings::textOrientation ( ) const
{
#ifndef IC_PM
  if ( ! ( fBidiSettingsData->fExplicitlySetFlags &
                        IBidiSettingsData::kTextOrientation ) )
  {
     return textLeftToRight;
  }
  else
#endif
  if ( fBidiSettingsData->data & BDA_TEXT_ORIENT_CONTEXT )
  {
     return textContextual;
  }
  else if ( fBidiSettingsData->data & BDA_TEXT_ORIENT_RTL )
  {
     return textRightToLeft;
  }
  return textLeftToRight;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::BidiLayout IBidiSettings::windowLayout ( ) const
{
#ifndef IC_PM
  if ( ! ( fBidiSettingsData->fExplicitlySetFlags &
                        IBidiSettingsData::kWindowLayout ) )
  {
     return layoutLeftToRight;
  }
  else
#endif
  if ( fBidiSettingsData->data & BDA_WND_ORIENT_RTL )
  {
     return layoutRightToLeft;
  }
  return layoutLeftToRight;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::BidiNumeralType IBidiSettings::numeralDisplay ( ) const
{
#ifndef IC_PM
  if ( ! ( fBidiSettingsData->fExplicitlySetFlags &
                        IBidiSettingsData::kNumeralDisplay ) )
  {
     return contextual;
  }
  else
#endif
  if ( ( fBidiSettingsData->data & BDA_NUMERALS_CONTEXTUAL ) ==
                BDA_NUMERALS_CONTEXTUAL )
  {
     return contextual;
  }
  else if ( ( fBidiSettingsData->data & BDA_NUMERALS_PASSTHRU ) ==
                BDA_NUMERALS_PASSTHRU )
  {
     return asStored;
  }
  else if ( ( fBidiSettingsData->data & BDA_NUMERALS_NATIONAL ) ==
                BDA_NUMERALS_NATIONAL )
  {
     return national;
  }
  return arabic;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IBidiSettings::isSymmetricSwappingEnabled ( ) const
{
#ifndef IC_PM
  if ( ! ( fBidiSettingsData->fExplicitlySetFlags &
                        IBidiSettingsData::kSymmetricSwapping ) )
  {
     return false;
  }
  else
#endif
  if ( fBidiSettingsData->data & BDA_SYM_SWAP_ON )
  {
     return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IBidiSettings::isWordByWordReorderingEnabled ( ) const
{
#ifndef IC_PM
  if ( ! ( fBidiSettingsData->fExplicitlySetFlags &
                        IBidiSettingsData::kWordByWordReordering ) )
  {
     return false;
  }
  else
#endif
  if ( fBidiSettingsData->data & BDA_WORD_BREAK_ON )
  {
     return true;
  }
  return false;
}

/*------------------------------------------------------------------------------
|                                                                              |
------------------------------------------------------------------------------*/
IBidiSettings::BidiTextShape IBidiSettings::textShape ( ) const
{
#ifndef IC_PM
  if ( ! ( fBidiSettingsData->fExplicitlySetFlags &
                        IBidiSettingsData::kTextShape ) )
  {
     return displayShaped;
  }
  else
#endif
  if ( ( fBidiSettingsData->data & BDA_TEXT_ISOLATED ) == BDA_TEXT_ISOLATED )
  {
     return isolatedShape;
  }
  else if ( ( fBidiSettingsData->data & BDA_TEXT_FINAL ) == BDA_TEXT_FINAL )
  {
     return finalShape;
  }
  else if ( ( fBidiSettingsData->data & BDA_TEXT_MIDDLE ) == BDA_TEXT_MIDDLE )
  {
     return middleShape;
  }
  else if ( ( fBidiSettingsData->data & BDA_TEXT_INITIAL ) == BDA_TEXT_INITIAL )
  {
     return initialShape;
  }
  else if ( ( fBidiSettingsData->data & BDA_TEXT_NOMINAL ) == BDA_TEXT_NOMINAL )
  {
     return nominalShape;
  }
  else if ( ( fBidiSettingsData->data & BDA_TEXT_SAVE_SHAPED ) == BDA_TEXT_SAVE_SHAPED )
  {
     return saveShaped;
  }
  return displayShaped;
}
