/*******************************************************************************
* FILE NAME: iBrshMgr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iBrshMgr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512 - 20
#pragma priority( -2147482132 )
// Priority is 10 more negative than iwindow0.cpp and 20 more negative than
// most of the rest of the class library.  This ensures the static collection,
// IBrushManager::BrushList, is destructed after any static objects that need
// it, such as IBaseSpinDefaultHandler in ispinszh.cpp.

#include <ibrshmgr.hpp>

/*------------------------------------------------------------------------------
| Brush Element class.                                                         |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IBrushElem::IBrushElem                                                       |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IBrushElem::IBrushElem( )
   : lColor(0), hBrush(0), lUseCount(0)
{
   // Default object only.
}

/*------------------------------------------------------------------------------
| IBrushElem::IBrushElem                                                       |
|                                                                              |
| Normal constructor. Used by IBrushManager for creating a brush.              |
------------------------------------------------------------------------------*/
IBrushElem::IBrushElem( long lClr )
   : lColor(lClr), hBrush(0), lUseCount(1)
{
   // Create the brush object and set use count to 1, because at least
   // one person is using it if brush manager created it.
   hBrush = CreateSolidBrush( lColor );

   // Note: lClr assumed to be an RGB value, adjusted via NEARESTPALETTECOLOR().
}

/*------------------------------------------------------------------------------
| IBrushElem::~IBrushElem                                                      |
|                                                                              |
| Destructor.  Cleans up the resource object.                                  |
| Should only be invoked as a result of a freeBrush() method call.             |
------------------------------------------------------------------------------*/
IBrushElem::~IBrushElem( )
{
   // Delete the object because no one is using it anymore.
   if ( hBrush )
      DeleteObject( hBrush );
}

/*------------------------------------------------------------------------------
| IBrushElem::useBrush                                                         |
|                                                                              |
| Increase use count, tracking references to this brush resource.              |
------------------------------------------------------------------------------*/
long IBrushElem::useBrush( )
{
   lUseCount++;
   return lUseCount;
}

/*------------------------------------------------------------------------------
| IBrushElem::freeBrush                                                        |
|                                                                              |
| Decrease use count, tracking references to this brush resource.              |
------------------------------------------------------------------------------*/
long IBrushElem::freeBrush( )
{
   lUseCount--;

   // If no one is referring to the brush, clean it up.
   if ( lUseCount == 0 )
      delete this;

   return lUseCount;
}

/*------------------------------------------------------------------------------
| IBrushElem::operator==                                                       |
|                                                                              |
| Equality between IBrushElems.                                                |
------------------------------------------------------------------------------*/
Boolean IBrushElem::operator==(IBrushElem& br)
{
   return( hBrush == br.hBrush );
}

/*------------------------------------------------------------------------------
| Brush Manager class. (list of Brush Elements)                                |
------------------------------------------------------------------------------*/

// Initialize static obects.
//
IBrushManager::BrushList IBrushManager::brushList;
long IBrushManager::lBrushCount = 0;

/*------------------------------------------------------------------------------
| IBrushElem::useBrush (public)                                                |
|                                                                              |
| Return a brush handle for the specified color, and increase the usage        |
| count for that brush handle. (Re-Use existing, if available.)                |
------------------------------------------------------------------------------*/
HBRUSH IBrushManager::useBrush( long lClr )
{
   // Convert and store via NEARESTPALETTECOLOR, so all
   // close colors will be using the same brush handle.
   long lRealColor = NEARESTPALETTECOLOR(lClr);

   IBrushElem* pElem(0);
   pElem = findBrush( lRealColor );

   if ( !pElem ) {
      // New entry required...
      // Create, add, set useCount to 1, and return handle.
      pElem = new IBrushElem( lRealColor );
      brushList.add( pElem );
      lBrushCount++;
   }

   // Existing entry... just return the handle.
   return( pElem->getHandle() );
}

/*------------------------------------------------------------------------------
| IBrushElem::freeBrush (public)                                               |
|                                                                              |
| Locate the specified brush handle, and release this use of it.               |
------------------------------------------------------------------------------*/
void IBrushManager::freeBrush( HBRUSH hBr )
{
   long lUseCount(0);
   IBrushElem* pElem(0);

   if (pElem = findBrush(hBr)) {
      // Save cursor location, in case elem must be deleted.
      BOOL bFound;
      long myKey = brushList.key(pElem);
      BrushList::Cursor myCursor(brushList);
      bFound = brushList.locateElementWithKey(myKey, myCursor);

      // Decrease access count on brush elem (may delete the elem).
      lUseCount = pElem->freeBrush();

      // Remove from list, if no longer used.
      if ( !lUseCount && bFound ) {
            brushList.removeAt(myCursor);
            lBrushCount--;
      }
   }
}

/*------------------------------------------------------------------------------
| IBrushElem::findBrush (protected)                                            |
|                                                                              |
| Locate the specified brush handle by color.                                  |
------------------------------------------------------------------------------*/
IBrushElem* IBrushManager::findBrush( long lClr )
{
   IBrushElem* pElem(0);
   BrushList::Cursor myCursor(brushList);

   BOOL bFound;
   bFound = brushList.locateElementWithKey(lClr, myCursor);
   if ( bFound ) {
      pElem = brushList.elementAt(myCursor);
      if ( pElem )
         return pElem;
   }

   return 0;
}

/*------------------------------------------------------------------------------
| IBrushElem::findBrush (protected)                                            |
|                                                                              |
| Locate the specified brush handle by handle.                                 |
------------------------------------------------------------------------------*/
IBrushElem* IBrushManager::findBrush( HBRUSH hBr )
{
   IBrushElem* pElem(0);
   BrushList::Cursor myCursor(brushList);

   if ( brushList.setToFirst(myCursor) ) {
      do {
         pElem = brushList.elementAt(myCursor);
         if ( pElem && (pElem->getHandle() == hBr) )
            return pElem;
      } while( myCursor.setToNext() );
   }

   return 0;
}

