/*******************************************************************************
* FILE NAME: icheckbx.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icheckbx.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINBUTTONS
  #define INCL_WINERRORS
  #define INCL_GPIBITMAPS
  #define INCL_WINPOINTERS
  #define INCL_WINSYS
  #define INCL_GPIPRIMITIVES
  #include <iwindefs.h>
}

#include <icheckbx.hpp>
#include <ibidiset.hpp>
#include <icolor.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <ihandle.hpp>
#include <ihandler.hpp>
#include <inotifev.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <istring.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICHECKBX_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public check box styles                                                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const ICheckBox::Style
  ICheckBox::autoSelect        ( 0, IBS_AUTOCHECKBOX ),
  ICheckBox::classDefaultStyle ( WS_VISIBLE, IBS_AUTOCHECKBOX );
#pragma data_seg()


/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
  ICheckBox::Style
        ICheckBox::currentDefaultStyle ( WS_VISIBLE, IBS_AUTOCHECKBOX );
#pragma data_seg()


/*------------------------------------------------------------------------------
| ICheckBox::ICheckBox                                                         |
|                                                                              |
| Constructor to create a check box on a standard window.                      |
------------------------------------------------------------------------------*/
ICheckBox :: ICheckBox(unsigned long ulId,
                       IWindow* pwndParent,
                       IWindow* pwndOwner,
                       const IRectangle& rectInit,
                       const Style& chbsStyle)
                     : fCheckBoxData( 0 )
{

  IASSERTPARM(pwndParent!=0);          //Assertions on input parms

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | chbsStyle.asExtendedUnsignedLong() );

  //Create the check box control
  IWindowHandle ownerHandle = IWindowHandle(0);
  if (pwndOwner)
     ownerHandle = pwndOwner->handle();

  unsigned long
    createStyle = this->convertToGUIStyle( chbsStyle );

  // Check if we need to supply bidi support to turn the check box
  // into a right-to-left one.
  if ( IBidiSettings::isBidiSupported() )
  {
     // The check box inherits bidi attributes from its parent
     // window.
     IBidiSettings
       bidiValues( *pwndParent );
     if ( bidiValues.windowLayout() == IBidiSettings::layoutRightToLeft )
     {
        // Change the control into a right-to-left one.
        createStyle |= BS_RIGHTBUTTON;
     }
  }

  //Create the check box control
  IWindowHandle whCheckBox =
      this -> create( ulId,
                      0,
                      createStyle,
                      WC_BUTTON,
                      pwndParent->handle(),
                      ownerHandle,
                      rectInit,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( chbsStyle, true ) );
  startHandlingEventsFor(whCheckBox);
}


/*------------------------------------------------------------------------------
| ICheckBox::ICheckBox                                                         |
|                                                                              |
| Constructor to instantiate an object for a dialog template.                  |
------------------------------------------------------------------------------*/
ICheckBox :: ICheckBox(unsigned long ulId,
                       IWindow* pdlgwndParent)
                     : fCheckBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(ulId, pdlgwndParent);
}


/*------------------------------------------------------------------------------
| ICheckBox::ICheckBox                                                         |
|                                                                              |
| Constructor to instantiate an object from an existing check box control.     |
------------------------------------------------------------------------------*/
ICheckBox :: ICheckBox(const IWindowHandle& wh)
                     : fCheckBoxData( 0 )
{
  setAutoDestroyWindow(false);
  startHandlingEventsFor(wh);
}


/*------------------------------------------------------------------------------
| ICheckBox::~ICheckBox                                                        |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
ICheckBox :: ~ICheckBox()
{ }


/*------------------------------------------------------------------------------
| ICheckBox::defaultStyle                                                      |
|                                                                              |
| Return the default style for new check box objects.                          |
------------------------------------------------------------------------------*/
ICheckBox::Style  ICheckBox :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| ICheckBox::setDefaultStyle                                                   |
------------------------------------------------------------------------------*/
void  ICheckBox :: setDefaultStyle(const Style& chbsStyle)
{
  currentDefaultStyle = chbsStyle;
}


/*------------------------------------------------------------------------------
| ICheckBox::convertToGUIStyle                                                 |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ICheckBox::convertToGUIStyle(const IBitFlag& guiStyle,
                                           Boolean bExtOnly) const
{
  // Obtain the style from the class (ISettingButton) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK;
  }
  else
  {
    // Let the clean BS_ styles, upper 12 bits of the lower word of the
    // style, flow thru
    ulStyle |= guiStyle.asUnsignedLong() & IBS_MASK;

    // BS_ styles in the lower 4 bits of the style overlap.  Therefore, map
    // our extended style bits, back to these values.
    unsigned long ulTmp = guiStyle.asExtendedUnsignedLong() & IBS_EXTCHECKBX;

    // Handle overlapping styles
    if (ulTmp & IBS_AUTOCHECKBOX)
      ulStyle |= BS_AUTOCHECKBOX;
    else
      ulStyle |= BS_CHECKBOX;

#ifdef IC_WIN
    ulStyle |= WS_CHILD;
#endif
  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| ICheckBox::enableAutoSelect                                                  |
|                                                                              |
| Sets or removes the autoSelect style for the check box control.              |
------------------------------------------------------------------------------*/
ICheckBox&  ICheckBox :: enableAutoSelect(Boolean bTurnOn)
{
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;

  if (bTurnOn)
  {
    ulStyle &= ~BS_PRIMARYSTYLES;  //Use masking flag from pmwin.h
    ulStyle |= BS_AUTOCHECKBOX;
  }
  else
  {
    ulStyle &= ~BS_PRIMARYSTYLES;  //Use masking flag from pmwin.h
    ulStyle |= BS_CHECKBOX;
  }

  if (ulStyle != ulOldStyle)
  {
    setStyle(ulStyle);
    refresh();                        //Force repaint of control
  }
  return *this;
}


/*------------------------------------------------------------------------------
| ICheckBox::disableAutoSelect                                                 |
------------------------------------------------------------------------------*/
ICheckBox&  ICheckBox :: disableAutoSelect()
{
  enableAutoSelect(false);
  return *this;
}


/*------------------------------------------------------------------------------
| ICheckBox::isAutoSelect                                                      |
------------------------------------------------------------------------------*/
Boolean  ICheckBox :: isAutoSelect() const
{
  return( (style() & BS_AUTOCHECKBOX & ~(BS_CHECKBOX)) ? true : false );
}


/*------------------------------------------------------------------------------
| ICheckBox::calcMinimumSize                                                   |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The minimum size   |
| is based on the maximum height of the font and the font width of the         |
| check box text (plus the size of the check box itself).                      |
------------------------------------------------------------------------------*/
ISize  ICheckBox :: calcMinimumSize() const
{
#ifdef IC_PM
   BITMAPINFOHEADER pbmpData;
   pbmpData.cbFix=sizeof(BITMAPINFOHEADER);
#endif
#ifdef IC_WIN
   BITMAP           pbmpData;
#endif
   HBITMAP hbit = (HBITMAP) IGETSYSBITMAP(SBMP_CHECKBOXES);
   IQUERYBITMAPPARAMETERS(hbit, &pbmpData);    //Retrieve size of check
   IDELETEBITMAP(hbit);                        //box bitmap

   ISize size = displaySize(text());
   unsigned long ulWidth = size.width();
   unsigned long i=textLength();
#ifdef IC_PM
   long bmpWidth = pbmpData.cx;
#endif
#ifdef IC_WIN
   long bmpWidth = pbmpData.bmWidth;
#endif

   if (i==0)
     ulWidth += ((bmpWidth/4) + 2 + (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   else
     ulWidth += (ulWidth/i + (bmpWidth/4) + 2 +
                                    (2 * IQUERYSYSVALUE(SV_CXBORDER)));
   unsigned long ulHeight = size.height() +
                                     (2 * IQUERYSYSVALUE(SV_CYBORDER));
   return ISize(ulWidth, ulHeight);
}
