/*******************************************************************************
* FILE NAME: igroupbx.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igroupbx.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINERRORS
  #define INCL_WINSTATICS
  #define INCL_GPIPRIMITIVES
  #define INCL_WINSYS
  #include <iwindefs.h>
}

#include <igroupbx.hpp>
#ifdef IC_WIN
  #include <igbpriv.hpp>
#endif
#include <icconst.h>
#include <icolor.hpp>
#include <iexcept.hpp>
#include <ifont.hpp>
#include <ihandle.hpp>
#include <iplatfrm.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <istring.hpp>
#include <itrace.hpp>
#include <ibrshmgr.hpp>    // 25227

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGROUPBX_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public group box styles                                                      |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IGroupBox::Style
  IGroupBox::classDefaultStyle (WS_VISIBLE);
#pragma data_seg()


/*------------------------------------------------------------------------------
| Default style for new group box objects (initial value).                     |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
IGroupBox::Style IGroupBox::currentDefaultStyle = WS_VISIBLE;
#pragma data_seg()

#ifdef IC_WIN
class IGroupBoxData : public IBase {
/*------------------------------------------------------------------------------
| The IGroupBoxData class encapsulates private data and functions used by      |
| the IGroupBox class.  An object of this class is created in the IGroupBox    |
| constructors.                                                                |
------------------------------------------------------------------------------*/
public:
  IGroupBoxData ( );

virtual
 ~IGroupBoxData ( );

HBRUSH
  bkgrndbrush;
}; // IGroupBoxData

/*------------------------------------------------------------------------------
| IGroupBoxData::IGroupBoxData                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
IGroupBoxData::IGroupBoxData ( )
  : bkgrndbrush( 0 )
{ }

/*------------------------------------------------------------------------------
| IGroupBoxData::~IGroupBoxData                                                |
|                                                                              |
------------------------------------------------------------------------------*/
IGroupBoxData::~IGroupBoxData ( )
{
  if ( bkgrndbrush != 0 )
     IBrushManager::freeBrush( bkgrndbrush );    // 25227
}
/*------------------------------------------------------------------------------
| IGroupHandler::IGroupHandler                                                 |
|                                                                              |
| Empty constructor here for page tuning.                                      |
------------------------------------------------------------------------------*/
IGroupHandler::IGroupHandler()
{ }

/*------------------------------------------------------------------------------
| IGroupHandler::~IGroupHandler                                                |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IGroupHandler::~IGroupHandler()
{ }

/*------------------------------------------------------------------------------
| IGroupHandler::dispatchHandlerEvent                                          |
|                                                                              |
| Intercepts control color events                                              |
| This handler returns false for all other events.                             |
------------------------------------------------------------------------------*/
IBase::Boolean IGroupHandler::dispatchHandlerEvent( IEvent& event )
{
  IMODTRACE_ALL("IGroupHandler::dispatchHandlerEvent");
  Boolean
    result = false;

  switch( event.eventId() )
  {
    case WM_CTLCOLORBTN:
    case WM_CTLCOLORSTATIC:
    {
      IGroupBox* ctlwin = (IGroupBox*)
         IWindow::windowWithHandle(IWindowHandle(event.parameter2()) );

      IColor clrFgnd = ctlwin->foregroundColor();
      IColor clrBgnd = ctlwin->backgroundColor();

      // Set foreground and background text colors in device context
      SetBkColor( (HDC)event.parameter1(), clrBgnd.asRGBLong() );
      SetTextColor( (HDC)event.parameter1(), clrFgnd.asRGBLong() );

      // Return HBRUSH for painting controls background
      if ( ctlwin->fGroupBoxData->bkgrndbrush != 0 )     // 25227
         IBrushManager::freeBrush( ctlwin->fGroupBoxData->bkgrndbrush );
      ctlwin->fGroupBoxData->bkgrndbrush =
         IBrushManager::useBrush( clrBgnd.asRGBLong() );

      event.setResult( ctlwin->fGroupBoxData->bkgrndbrush );
      result = true;
      break;
    }

    default:
      break;
  }
  return result;
}

/*------------------------------------------------------------------------------
| DefaultHandler                                                               |
|                                                                              |
| This class wrappers an object pointer and destroys the pointer to the object |
| when the wrapper is destructed.                                              |
------------------------------------------------------------------------------*/
static struct DefaultHandler
{
  operator IGroupHandler* ();
IGroupHandler
 *operator -> ()
    {
    return *this;
    }
 ~DefaultHandler ( )
    {
    if (ptr)
      {
      delete ptr;
      ptr = 0;
      }
    }
IGroupHandler
 *ptr;
} defaultHandler;

DefaultHandler::operator IGroupHandler* ( )
  {
  if ( !ptr )
    ptr = new IGroupHandler;
  return ptr;
  }
#endif // IC_WIN


/*------------------------------------------------------------------------------
| IGroupBox::IGroupBox                                                         |
|                                                                              |
| Constructor to create a group box control on a standard window.              |
------------------------------------------------------------------------------*/
IGroupBox :: IGroupBox(unsigned long ulId,
                       IWindow* pwndParent,
                       IWindow* pwndOwner,
                       const IRectangle& rectInit,
                       const Style& gbsStyle)
#ifdef IC_WIN
  : fGroupBoxData( new IGroupBoxData )
#else
  : fGroupBoxData( 0 )
#endif
{
  IASSERTPARM(pwndParent!=0);          //Assertions on input parms

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | gbsStyle.asExtendedUnsignedLong() );

  //Create group box control
  IWindowHandle ownerHandle = IWindowHandle(0);
  if (pwndOwner)
     ownerHandle = pwndOwner->handle();
  IWindowHandle whGroupBox =
      this -> create( ulId,
                      0,
                      convertToGUIStyle( gbsStyle ),
#ifdef IC_MOTIFPM
                      WC_STATIC,
#endif
#ifdef IC_WIN
                      WC_BUTTON,
#endif
                      pwndParent->handle(),
                      ownerHandle,
                      rectInit,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( gbsStyle, true ) );

#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
  startHandlingEventsFor(whGroupBox);
}


/*------------------------------------------------------------------------------
| IGroupBox::IGroupBox                                                         |
|                                                                              |
| Constructor to create instance for a template window.                        |
------------------------------------------------------------------------------*/
IGroupBox :: IGroupBox(unsigned long ulId,
                       IWindow* pdlgwndParent)
#ifdef IC_WIN
  : fGroupBoxData( new IGroupBoxData )
#else
  : fGroupBoxData( 0 )
#endif
{
  setAutoDestroyWindow(false);
#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
  startHandlingEventsFor(ulId, pdlgwndParent);
}


/*------------------------------------------------------------------------------
| IGroupBox::IGroupBox                                                         |
|                                                                              |
| Constructor to instantiate from an existing group box control.               |
------------------------------------------------------------------------------*/
IGroupBox :: IGroupBox(const IWindowHandle& wh)
#ifdef IC_WIN
  : fGroupBoxData( new IGroupBoxData )
#else
  : fGroupBoxData( 0 )
#endif
{
  setAutoDestroyWindow(false);
#ifdef IC_WIN
  this->addHandler( defaultHandler );
#endif
  startHandlingEventsFor(wh);
}


/*------------------------------------------------------------------------------
| IGroupBox::~IGroupBox                                                        |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IGroupBox :: ~IGroupBox()
{
#ifdef IC_WIN
  removeHandler(defaultHandler);
  if ( fGroupBoxData != 0 )
    delete fGroupBoxData;
#endif
}


/*------------------------------------------------------------------------------
| IGroupBox::defaultStyle                                                      |
|                                                                              |
| Return the default style for new group box objects.                          |
------------------------------------------------------------------------------*/
IGroupBox::Style  IGroupBox :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| IGroupBox::setDefaultStyle                                                   |
|                                                                              |
| Replace the default style for new group box objects.                         |
------------------------------------------------------------------------------*/
void  IGroupBox :: setDefaultStyle(const IGroupBox::Style& gbsStyle)
{
  currentDefaultStyle = gbsStyle;
}


/*------------------------------------------------------------------------------
| IGroupBox::convertToGUIStyle                                                 |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IGroupBox::convertToGUIStyle(const IBitFlag& guiStyle,
                                           Boolean bExtOnly) const
{
  // Obtain the style from the class (IControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & IS_EXTMASK ;
  }
  else
  {
    // Let the clean styles flow thru and add the groupbox style
    ulStyle |= guiStyle.asUnsignedLong();
#ifdef IC_MOTIFPM
    ulStyle |= SS_GROUPBOX;
#endif
#ifdef IC_WIN
    ulStyle |= (BS_GROUPBOX | WS_CHILD);
#endif
  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| IGroupBox::calcMinimumSize                                                   |
------------------------------------------------------------------------------*/
ISize  IGroupBox :: calcMinimumSize() const
{
  ISize siz = displaySize(text());

  /*----------------------------------------------------------------------------
  | We need to pad the width of 2 characters to allow for the border           |
  ----------------------------------------------------------------------------*/
  int pad = IFont(this).avgCharWidth() << 1;
  siz.setWidth(siz.width() + pad);

  return siz;
}

/*------------------------------------------------------------------------------
| WIN32 Color Support                                                          |
|                                                                              |
| Windows has no API equivalent for Presentation Parameters.  The following    |
| group box color function is NOPed for Windows.  The query function returns   |
| an appropriate default IColor object.                                        |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| IGroupBox::foregroundColor                                                   |
|                                                                              |
| Returns the foreground color of the IGroupBox.                               |
------------------------------------------------------------------------------*/
IColor IGroupBox::foregroundColor () const
{
#ifdef IC_PM
  return (IWindow::color(PP_FOREGROUNDCOLOR,
                         IGUIColor(IGUIColor::windowStaticText)));
#else
//--------------------------------------------------------------------------
// DEFECT 23404 :
// The following behaves correctly on win32, either picking up the default
// static text color, or frame foreground color if set, or an explicit 
// groupbox color. Therefore, no reason to NOP this code.
//--------------------------------------------------------------------------
  return (IWindow::color(PP_FOREGROUNDCOLOR,
                         IGUIColor(IGUIColor::windowStaticText)));
//--------------------------------------------------------------------------
//  return IGUIColor( IGUIColor::windowStaticText );
#endif //IC_PM
}

/*------------------------------------------------------------------------------
| IGroupBox::setText                                                           |
|                                                                              |
| Changes the text used by the group box.                                      |
------------------------------------------------------------------------------*/
IGroupBox &  IGroupBox :: setText ( const char* pszText )
{
  Inherited::setText(pszText);
  setLayoutDistorted(IWindow::minimumSizeChanged, 0);
  return *this;
}


/*------------------------------------------------------------------------------
| IGroupBox::setText                                                           |
|                                                                              |
| Changes the group box text from a string table.                              |
------------------------------------------------------------------------------*/
IGroupBox &  IGroupBox :: setText ( const IResourceId & residText )
{
  Inherited::setText(residText);
  return *this;
}

/*------------------------------------------------------------------------------
| IGroupBox::visibleRectangle                                                  |
|                                                                              |
| Returns the area painted by the group box.                                   |
------------------------------------------------------------------------------*/
IRectangle IGroupBox :: visibleRectangle () const
{
  return IRectangle(position(),ISize(0,0));
}

