/*******************************************************************************
* FILE NAME: igstring.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in igstring.hpp                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
{
   #include <iwindefs.h>
}

#include <igstring.hpp>
#include <igrafctx.hpp>
#include <igbundle.hpp>
#include <ihandle.hpp>
#include <ireslib.hpp>
#include <ifont.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IGSTRING_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IGString::IGString                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::IGString( const IGString& graphicString )
  : IGString::Inherited( graphicString ), fString( graphicString.fString ),
    fLocation( graphicString.fLocation ), fFont(0), fClip(0)
{
  if (graphicString.fFont)
    fFont = new IFont(*(graphicString.fFont));
}

/*------------------------------------------------------------------------------
| IGString::IGString                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::IGString( const IString& string, const IPoint& point )
  : IGString::Inherited(), fString( string ), fLocation( point ), fFont(0), fClip(0)
{
}

/*------------------------------------------------------------------------------
| IGString::IGString                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::IGString( const IString& string, const IPoint& point,
                    const IFont& font )
  : IGString::Inherited(), fString( string ), fLocation( point ),
    fFont( new IFont( font )), fClip(0)
{
}

/*------------------------------------------------------------------------------
| IGString::~IGString                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString::~IGString()
{
  if (fFont)
  {
    delete fFont;
  }
}


/*------------------------------------------------------------------------------
| IGString::drawOn                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::drawOn(IGraphicContext& graphicContext)
{
  graphicContext.draw( *this );
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::font                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IFont IGString::font() const
{
  if ( fFont )
    return IFont( *fFont );
  else
    return IFont();
}

/*------------------------------------------------------------------------------
| IGString::setFont                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::setFont( const IFont& font )
{
  if ( fFont )
  {
    delete fFont;
  }
  fFont = new IFont(font);
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::hasFont                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGString::hasFont() const
{
  return ((fFont == 0) ? false : true);
}

/*------------------------------------------------------------------------------
| IGString::removeFont                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::removeFont()
{
  if ( fFont )
  {
    delete fFont;
    fFont = 0;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::moveTo                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::moveTo ( const IPoint& point )
{
  fLocation = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::position                                                           |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint IGString::position( ) const
{
  return fLocation;
}

IGString& IGString::setClippingRect   ( const IRectangle& rect )
{
  if (fClip)
    delete fClip;
  fClip = new IRectangle( rect );
  return *this;
}

IRectangle IGString::clippingRect       ( ) const
{
  if (fClip)
    return *fClip;

  return IRectangle();

}

IGString& IGString::resetClippingRect ( )
{
  if (fClip)
  {
    delete fClip;
    fClip = 0;
  }
  return *this;
}

IBase::Boolean IGString::isClippingRectSet  ( ) const
{
  return ((fClip == 0) ? false : true);
}

/*------------------------------------------------------------------------------
| IGString::operator=                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::operator=( const IGString& graphicString )
{
  if ( this == &graphicString )
    return *this;

  this->Inherited::operator=( graphicString );
  fString = graphicString.fString;
  fLocation = graphicString.fLocation;
  if ( fFont )
  {
    delete fFont;
    fFont = 0;
  }
  if ( graphicString.fFont )
  {
    fFont = new IFont(*graphicString.fFont);
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::operator==                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGString::operator==( const IGString& graphicString ) const
{
  if ( fFont && graphicString.fFont )
  {
    return ( fString   == graphicString.fString    &&
             fLocation == graphicString.fLocation  &&
             this->Inherited::operator==( graphicString ));
  }
  else if (!fFont && !graphicString.fFont)
  {
    return ( fString    == graphicString.fString    &&
             fLocation  == graphicString.fLocation  &&
             this->Inherited::operator==( graphicString ));
  }
  else
  {
    return false;
  }
}

/*------------------------------------------------------------------------------
| IGString::operator!=                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IBase::Boolean IGString::operator!=( const IGString& graphicString ) const
{
  return !( *this == graphicString );
}

/*------------------------------------------------------------------------------
| IGString::setText                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::setText( const char* text )
{
  fString = text;
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::setText                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IGString& IGString::setText( const IResourceId& residText )
{
  IString str = residText.resourceLibrary().loadString(residText);
  setText((char*)str);
  return *this;
}

/*------------------------------------------------------------------------------
| IGString::text                                                               |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IString IGString::text( ) const
{
  return fString;
}
