/*******************************************************************************
* FILE NAME: ihandle.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ihandle.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_WINPOINTERS
  #define INCL_GPIBITMAPS
  #define INCL_WINMESSAGEMGR
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
  #include <malloc.h>
  }

#include <ihandle.hpp>

#ifndef _IRESLIB_
  #include <ireslib.hpp>
#endif

#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _ISTRING_
  #include <istring.hpp>
#endif

#ifndef _ISTREAM_
  #include <istream.hpp>
#endif

#ifndef _ITHREAD_
  #include <ithread.hpp>
#endif
#ifndef _IWINDOW_
  #include <iwindow.hpp>
#endif

#ifndef _IWCNAME_
  #include <iwcname.hpp>
#endif

#ifndef _ICCONST_
  #include <icconst.h>
#endif

#ifndef _IPLATFRM_
  #include <iplatfrm.hpp>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IHANDLE_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_WIN
  #ifndef _ICLCCL_
   #include <iclccl.h>
  #endif
  #include <icctlsta.hpp>

// Define a structure used on the EnumWindows/EnumChildWindows callback function
typedef struct _CallbackParms {
  Boolean                    fPost;
  unsigned long              fEventId;
  IEventData                 fParameter1;
  IEventData                 fParameter2;
  IWindowHandle::BroadcastTo fValue;
} CallbackParms;

// Callback function for EnumWindows/EnumChildWindows
static BOOL CALLBACK iwindowhandleEnumCB( HWND hwnd, LPARAM callbackParms);

#endif

#include <itrace.hpp>

/*------------------------------------------------------------------------------
| IThreadHandle::noHandle - Used in IThread constructor to represent the       |
|                           default value for the threadHandle parameter,      |
|                           which indicates that the thread is already         |
|                           accessible to the Open Class Library or is the     |
|                           current thread.                                    |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IThreadHandle IThreadHandle::noHandle ((void*)-1);
#pragma data_seg()

/*------------------------------------------------------------------------------
 isValid - Is this a valid window handle for the presentation system?
------------------------------------------------------------------------------*/
Boolean IWindowHandle::isValid ( ) const
{
#ifdef IC_PM
  Boolean valid = IQUERYANCHOR( this->handle );
  if ( ! valid )
  {
     // Handle the slim chance that the application is relying on this
     // function to call WinInitialize via ICurrentThread::anchorBlock.
     valid = IISWINDOW( IThread::current().anchorBlock(), this->handle );
  }
  return valid;
#endif
#ifdef IC_WIN
  return IISWINDOW( 1, this->handle );
#endif
#ifdef IC_MOTIF
  return IISWINDOW( IThread::current().anchorBlock(), this->handle );
#endif
}


/*------------------------------------------------------------------------------
 sendEvent - send a window system message/event to the window
 represented by this handle, using the specified parameters.
------------------------------------------------------------------------------*/
IEventResult IWindowHandle::sendEvent  ( unsigned long eventId,
            const IEventParameter1 &parm1, const IEventParameter2 &parm2 ) const
{
  MRESULT rs;

#ifdef IC_WIN

  if ( (IPlatform::isWin32S()) &&
       (HIWORD(parm1) != 0) &&
       (eventId >= CCL_MESSAGE_BASE) &&
       (eventId <= CCL_MESSAGE_TOP) &&
       (IWindowClassName(this->handle) != WC_NATIVECONTAINER) )
   {
    /*****************************************************************/
    /* At this point the message could be destined for a CCl control */
    /* or for a native control.  If its for a native control, then   */
    /* the message is processed by a native control handler, not by  */
    /* the CCL control window proc.  We cannot guarentee that the    */
    /* CCL .DLL will be loaded.  Therefore, if the message is for a  */
    /* a native control, let it pass to ISENDMESSAGE, which will     */
    /* process the message in i32to16.cpp                            */
    /*****************************************************************/

    ITRACE_DEVELOP("Win32s: Send Msg to CCL control, calling .DLL 32to16");
    ITRACE_ALL(IString("HWND=") + IString((unsigned long)this->handle).d2x()+
               IString(", Msg=") + IString(eventId).d2x()+
               IString(", MP1=") + IString((unsigned long)parm1).d2x()+
               IString(", MP2=") + IString((unsigned long)parm2).d2x());

    rs = (*IControlStatics::lpfn32to16SendMessage)(this->handle,
                                                   (unsigned int)eventId,
                                                   parm1,
                                                   parm2);

    ITRACE_DEVELOP(IString("Win32s: CCL returned DLL 32to16 result is:")
                 + IString((long)rs) );
   }
  else
#endif
   rs = ISENDMESSAGE(this->handle, (unsigned int)eventId, parm1, parm2);

   return IEventResult((int)rs); //JJB NEED TO ADD A LONG CONSTRUCTOR
}

/*------------------------------------------------------------------------------
 sendEvents -  send a window system message/event to the children
 or decendents of the window represented by this handle,
 using the specified parameters.
------------------------------------------------------------------------------*/
void IWindowHandle::sendEvents ( unsigned long eventId,
         const IEventParameter1 &parm1, const IEventParameter2 &parm2,
         BroadcastTo value ) const
{
#ifdef IC_PM
   unsigned long ulCmd = BMSG_SEND;
   switch(value)
   {
     case descendants:
       ulCmd |= BMSG_DESCENDANTS;
       break;
     case frames:
       ulCmd |= BMSG_FRAMEONLY;
       break;
     default:
       ulCmd |= (BMSG_FRAMEONLY | BMSG_DESCENDANTS);
       break;
   } /* endSwitch */
   if (!WinBroadcastMsg(this->handle, eventId, parm1, parm2, ulCmd))
      ITHROWGUIERROR2("WinBroadcastMsg", IBaseErrorInfo::invalidParameter, IException::recoverable);
#endif
#ifdef IC_WIN
  // Enumerate windows and send the event
  CallbackParms callbackParms;
  callbackParms.fPost       = false;
  callbackParms.fEventId    = eventId;
  callbackParms.fParameter1 = parm1;
  callbackParms.fParameter2 = parm2;
  callbackParms.fValue      = value;
  switch(value)
  {
    case frames:
      EnumWindows( (WNDENUMPROC)iwindowhandleEnumCB,
                   (LPARAM)(&callbackParms));
      break;
    default:
      EnumChildWindows(this->handle,
                       (WNDENUMPROC)iwindowhandleEnumCB,
                       (LPARAM)(&callbackParms));
      break;
  } /* endSwitch */
#endif
}


/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to the window
 represented by this handle, using the specified parameters.
------------------------------------------------------------------------------*/
void IWindowHandle::postEvent  ( unsigned long eventId,
         const IEventParameter1 &parm1, const IEventParameter2 &parm2 ) const
{
  MRESULT rs;

#ifdef IC_WIN

  if ( (IPlatform::isWin32S()) &&
       (HIWORD(parm1) != 0) &&
       (eventId >= CCL_MESSAGE_BASE) &&
       (eventId <= CCL_MESSAGE_TOP) )
   {
    ITRACE_DEVELOP("Win32s: Post Msg to CCL control, calling .DLL 32to16");

    ITRACE_ALL(IString("HWND=") + IString((unsigned long)this->handle).d2x()+
               IString(", Msg=") + IString(eventId).d2x()+
               IString(", MP1=") + IString((unsigned long)parm1).d2x()+
               IString(", MP2=") + IString((unsigned long)parm2).d2x());

    rs = (*IControlStatics::lpfn32to16PostMessage)(this->handle,
                                                   (unsigned int)eventId,
                                                   parm1,
                                                   parm2);

    ITRACE_DEVELOP(IString("Win32s result is:") + IString((long)rs) );
    ITRACE_DEVELOP("Win32s: CCL! returned DLL 32to16 post");
   }
  else
#endif
   rs = IPOSTMESSAGE(this->handle, (unsigned int)eventId, parm1, parm2);

   if(!rs)
      ITHROWGUIERROR2("WinPostMsg", IBaseErrorInfo::invalidParameter, IException::recoverable);
}

/*------------------------------------------------------------------------------
 postEvents -  post (asynchronously)  a window system message/event to
 the  children/decendents of the window represented by this handle,
 using the specified parameters.
------------------------------------------------------------------------------*/
void IWindowHandle::postEvents ( unsigned long eventId,
      const IEventParameter1 &parm1, const IEventParameter2 &parm2,
      BroadcastTo value) const
{
#ifdef IC_PM
   unsigned long ulCmd = BMSG_POST;
   switch(value)
   {
     case descendants:
       ulCmd |= BMSG_DESCENDANTS;
       break;
     case frames:
       ulCmd |= BMSG_FRAMEONLY;
       break;
     default:
       ulCmd |= (BMSG_FRAMEONLY | BMSG_DESCENDANTS);
       break;
   } /* endSwitch */

   if (!WinBroadcastMsg(this->handle, eventId, parm1, parm2, ulCmd))
      ITHROWGUIERROR2("WinBroadcastMsg", IBaseErrorInfo::invalidParameter, IException::recoverable);
#endif
#ifdef IC_WIN
  // Enumerate windows and post the event
  CallbackParms callbackParms;
  callbackParms.fPost       = true;
  callbackParms.fEventId    = eventId;
  callbackParms.fParameter1 = parm1;
  callbackParms.fParameter2 = parm2;
  callbackParms.fValue      = value;
  switch(value)
  {
    case frames:
      EnumWindows( (WNDENUMPROC)iwindowhandleEnumCB,
                   (LPARAM)(&callbackParms));
      break;
    default:
      EnumChildWindows(this->handle,
                       (WNDENUMPROC)iwindowhandleEnumCB,
                       (LPARAM)(&callbackParms));
      break;
  } /* endSwitch */
#endif
}

/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to this
 message queue , using the specified parameters.
------------------------------------------------------------------------------*/
void IMessageQueueHandle::postEvent  ( unsigned long eventId,
         const IEventParameter1 &parm1, const IEventParameter2 &parm2) const
{
   if (!IPOSTQUEUEMESSAGE(this->asUnsigned(), eventId, parm1, parm2))
      ITHROWGUIERROR2("WinPostQueueMsg", IBaseErrorInfo::invalidParameter, IException::recoverable);
}
/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to the
 specified message queue handle, using the specified parameters.
------------------------------------------------------------------------------*/
void IMessageQueueHandle::postEvent  ( const IMessageQueueHandle& handle,
         unsigned long eventId, const IEventParameter1 &parm1,
         const IEventParameter2 &parm2)
{
   if (!IPOSTQUEUEMESSAGE(handle.asUnsigned(), eventId, parm1, parm2))
      ITHROWGUIERROR2("WinPostQueueMsg", IBaseErrorInfo::invalidParameter, IException::recoverable);
}

/*------------------------------------------------------------------------------
 postEvent - post (asynchronously)  a window system message/event to
 all threads that have a message queue, using the specified parameters.
------------------------------------------------------------------------------*/
void IMessageQueueHandle::postEvents ( unsigned long eventId,
      const IEventParameter1 &parm1, const IEventParameter2 &parm2)
{
#ifdef IC_PM
   if (!WinBroadcastMsg(0, eventId, parm1, parm2, BMSG_POSTQUEUE))
      ITHROWGUIERROR2("WinBroadcastMsg", IBaseErrorInfo::invalidParameter, IException::recoverable);
#endif
#ifdef IC_WIN
   if (!IPOSTMESSAGE(HWND_BROADCAST, eventId, parm1, parm2))
      ITHROWGUIERROR2("PostMessage", IBaseErrorInfo::invalidParameter, IException::recoverable);
#endif
}

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| iwindowhandleEnumCB                                                          |
| Callback function for EnumWindows/EnumChildWindows                           |
| Used to implement sendEvents/postEvents                                      |
------------------------------------------------------------------------------*/
BOOL CALLBACK iwindowhandleEnumCB( HWND hwnd, LPARAM callbackParms)
{
   CallbackParms* parms( (CallbackParms*)callbackParms );
   IBase::Boolean bSend( parms->fValue != IWindowHandle::frameDescendants );
   if ( !bSend )
      {
      // only need this if using frameDescendants
      IWindow* win = IWindow::windowWithHandle( hwnd );
      if (win && win->isFrameWindow())
         bSend = true;
      }

   // post or send to this window
   if (bSend)
      {
      if (parms->fPost)
         IPOSTMESSAGE(hwnd,
                      (unsigned int)parms->fEventId,
                      parms->fParameter1,
                      parms->fParameter2 );
      else
         ISENDMESSAGE(hwnd,
                      (unsigned int)parms->fEventId,
                      parms->fParameter1,
                      parms->fParameter2 );
      }
   return 1;      // continue enumeration
}
#endif

#pragma info(noext)

