/*******************************************************************************
* FILE NAME: ihandle3.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions for Pointer     *
*   handles declared ihandle.hpp.                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C"
  {
  #define INCL_WINPOINTERS
  #define INCL_WINMESSAGEMGR
  #define INCL_WINWINDOWMGR
  #include <iwindefs.h>
  }

#include <ihandle.hpp>

#ifndef _IRESLIB_
  #include <ireslib.hpp>
#endif

#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif

#ifndef _ICCONST_
  #include <icconst.h>
#endif

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IHANDLE3_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
 Note that pointer handles use references to reduce
  duplicate pointers for better storage utilization.
  See IReference, IRefCounted classes for more info.
------------------------------------------------------------------------------*/

IPointerHandle::IPointerHandle( Value value )
  : IHandle( value )
{
   if(value!=0)
     IResourceLibrary::addRef(*this);
}

IPointerHandle::IPointerHandle(const IPointerHandle& pointerHandle)
  : IHandle( pointerHandle )
{
   if(pointerHandle!=0)
     IResourceLibrary::addRef(pointerHandle);
}

IPointerHandle::~IPointerHandle()
{
   if(handle!=0)
     IResourceLibrary::removeRef(*this);
}


IPointerHandle& IPointerHandle::operator =(const IPointerHandle& pointerHandle)
{
   /* Note: Ensure self assignment works */
     if(this != &pointerHandle)
     {
       if(pointerHandle.handle!=0)
          IResourceLibrary::addRef(pointerHandle);
       if(this->handle!=0)
          IResourceLibrary::removeRef(*this);
       this->handle = pointerHandle.handle;
     }
     return *this;
}


ISystemPointerHandle:: ISystemPointerHandle(Identifier pointerId,
                                            Boolean    fMakeCopy)
   : IPointerHandle(0)
{
long lPMId = 0;
// Flag for Windows macro to determine correct api to use in loading resource
Boolean fIconApi = true;
switch(pointerId)
{
  case  arrow:
        lPMId = SPTR_ARROW;
        break;
  case  text:
        lPMId = SPTR_TEXT;
        break;
  case  wait:
        lPMId = SPTR_WAIT;
        break;
  case  move:
        lPMId = SPTR_MOVE;
        break;
  case  sizeNWSE:
        lPMId = SPTR_SIZENWSE;
        break;
  case  sizeNESW:
        lPMId = SPTR_SIZENESW;
        break;
  case  sizeHorizontal:
        lPMId = SPTR_SIZEWE;
        break;
  case  sizeVertical:
        lPMId = SPTR_SIZENS;
        break;
  case  standardApplication:
        lPMId = SPTR_APPICON;
        fIconApi = false;
        break;
  case  information:
        lPMId = SPTR_ICONINFORMATION;
        fIconApi = false;
        break;
  case  question:
        lPMId = SPTR_ICONQUESTION;
        fIconApi = false;
        break;
  case  error:
        lPMId = SPTR_ICONERROR;
        fIconApi = false;
        break;
  case  warning:
        lPMId = SPTR_ICONWARNING;
        fIconApi = false;
        break;
  case  illegalOperation:
        lPMId = SPTR_ILLEGAL;
        break;
#ifndef IC_WIN
  case  singleFile:
        lPMId = SPTR_FILE;
        break;
  case  multipleFile:
        lPMId = SPTR_MULTFILE;
        break;
  case  folder:
        lPMId = SPTR_FOLDER;
        break;
  case  program:
        lPMId = SPTR_PROGRAM;
        break;
#endif  // !IC_WIN
} /* end switch */

  if (lPMId)
  {
    handle = (IHandle::Value)IResourceLibrary::systemPointer( lPMId, fMakeCopy,
                                                              fIconApi );

    /* If we return a copy we will manage reference count for this resource */
    /* otherwise not (don't want to destroy a system pointer).          */
    if(fMakeCopy)
    {
      IResourceLibrary::addRef(IPointerHandle(*this));
    }
  }
  /* This is for icons that are not part of the operating system.  We
   * need to load them ourselves.  I have set up a switch statement so
   * that when we add more icons, it will be easily expandable.
   */
  else
  {
#ifdef IC_WIN
    Boolean useShell32 = false;
#endif
    switch(pointerId)
    {
      case  dragFade:
        lPMId = IC_FADEICON;
        fMakeCopy = true;
        break;
#ifdef IC_WIN
      // These we attempt to load from shell32.dll.  The resource Ids
      // are NOT documented, but were determined by experimentation.
      case  singleFile:
        lPMId = IC_FILE_SHELL32;
        fMakeCopy = false;
        useShell32 = true;
        break;
      case  multipleFile:
        lPMId = IC_MULTIFILE_SHELL32;
        fMakeCopy = false;
        useShell32 = true;
        break;
      case  folder:
        lPMId = IC_FOLDER_SHELL32;
        fMakeCopy = false;
        useShell32 = true;
        break;
      case  program:
        lPMId = IC_PROGRAM_SHELL32;
        fMakeCopy = false;
        useShell32 = true;
        break;
#endif
    }   // switch
#ifdef IC_WIN
    if (useShell32)
       {
       // Try to load the icon from shell32.dll.  If that fails, then
       // punt and return the standard application icon.
       IPointerHandle ph(0);
       try
          {
          ph =
            IDynamicLinkLibrary(IC_SHELL32_DLL).tryToLoadIcon( lPMId,
                                                               fMakeCopy );
          }
       catch (...)
          {
          // this is to catch nonexistant DLL errors
          }

       if (ph)
          {
          // Got it.  return icon from shell32
          handle = ph;
          }
       else
          {
          // Not found.  This may happen on NT 3.51 without newshell
          handle = (IHandle::Value)
             IResourceLibrary::systemPointer( SPTR_APPICON,
                                              fMakeCopy,
                                              false );
          }
       }
    else
#endif
       {
       IPointerHandle ph =
               IApplication::current().resourceLibrary().loadIcon(lPMId);
       handle = ph;
       if(fMakeCopy)
         {
         IResourceLibrary::addRef(IPointerHandle(*this));
         }
       }
  }   // else nonstandard icons
}

ISystemPointerHandle :: ~ISystemPointerHandle( )
{ }

