/*******************************************************************************
* FILE NAME: ikeyevt.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ikeyevt.hpp.                                                            *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

extern "C"
  {
  #define INCL_WININPUT
  #define INCL_NLS
  #include <iwindefs.h>
  }

#include <ikeyevt.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <istring.hpp>


// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IKEYEVT_CPP_
  #include <ipagetun.h>
#endif


class IKeyboardEventPrivateData {
public:
  IKeyboardEventPrivateData ( )
   : fMappedCharCode( 0 ),
     fSpecialCharacter( false )
  { }
 ~IKeyboardEventPrivateData ( )
  { }
private:
friend class IKeyboardEvent;
char
  fMappedCharCode;
IBase::Boolean
  fSpecialCharacter;
}; // IKeyboardEventPrivateData


/*------------------------------------------------------------------------------
| isSpecialCharacter                                                           |
|                                                                              |
| Local API for determining if the value passed is one of the keys which       |
| require special processing on our part.                                      |
------------------------------------------------------------------------------*/
IBase::Boolean isSpecialCharacter ( unsigned long keyCode )
{
  return ( ( keyCode == VK_TAB ||
             keyCode == VK_RETURN ||
             keyCode == VK_BACK ||
             keyCode == VK_BREAK ||
             keyCode == VK_ESCAPE ) ? true : false );
}


/*******************************************************************/
/* Initialize static flags used by the testing functions.          */
/*******************************************************************/
const unsigned long IKeyboardEvent::ulAltFlag              ( 0x20000000 );
const unsigned long IKeyboardEvent::ulUpTransitionFlag     ( 0x80000000 );
#ifndef IC_WIN_FLAGNOP
const unsigned long IKeyboardEvent::ulCharacterFlag        ( 0x00000000 );
const unsigned long IKeyboardEvent::ulScanCodeFlag         ( 0x00000000 );
const unsigned long IKeyboardEvent::ulVirtualFlag          ( 0x00000000 );
const unsigned long IKeyboardEvent::ulRepeatFlag           ( 0x00000000 );
const unsigned long IKeyboardEvent::ulUncombinedFlag       ( 0x00000000 );
const unsigned long IKeyboardEvent::ulShiftFlag            ( 0x00000000 );
const unsigned long IKeyboardEvent::ulCtrlFlag             ( 0x00000000 );
const unsigned long IKeyboardEvent::ulForCompositeFlag     ( 0x00000000 );
const unsigned long IKeyboardEvent::ulCompositeFlag        ( 0x00000000 );
const unsigned long IKeyboardEvent::ulInvalidCompositeFlag ( 0x00000000 );
#endif

/*------------------------------------------------------------------------------
| IKeyboardEvent::IKeyboardEvent                                               |
|                                                                              |
| Constructor.                                                                 |
------------------------------------------------------------------------------*/
IKeyboardEvent::IKeyboardEvent ( const IEvent& event )
  : IKeyboardEvent::Inherited( event ),
    fValidComposite( false ),
    fInvalidComposite( false ),
    pData( new IKeyboardEventPrivateData )
{
  //---------------------------------------------------------
  // DEFECT 8579
  // NOTE: Windows is generating some bad character values
  //       in the WM_CHAR, for Ctl-I, Ctl-M.  This causes
  //       the keyboard event to treat a Ctl-I as a VK_TAB,
  //       and a Ctl-M as a VK_ENTER.
  //---------------------------------------------------------
  pData->fSpecialCharacter = isSpecialCharacter( this->parameter1() );

  switch ( event.eventId() )
  {
    case WM_KEYDOWN:
    case WM_SYSKEYDOWN:
    case WM_KEYUP:
    case WM_SYSKEYUP:
       if ( pData->fSpecialCharacter == false )
       {
          unsigned int charCode = MapVirtualKey( event.parameter1(), 2 );
          if ( charCode  &&  this->isCtrlDown() )
          {
             // Windows generates virtual keys for Ctrl key combinations.
             // Treat Ctrl + character key as a character key, rather
             // than a virtual key.  However, we will not modify the
             // message parameters, so a client can still access the data
             // that Windows stores in the message.
             Boolean
               shiftDown = this->isShiftDown(),
               capsLockDown = ( GetKeyState( VK_CAPSLOCK ) & 0x0001 ) ? true : false;
             if ( ( shiftDown  &&  capsLockDown )  ||
                  ( ! shiftDown  &&  ! capsLockDown ) )
             {
                // The character code returned for VK_A-VK_Z is always an
                // uppercased character.  Convert it to a lowercase
                // character if appropriate (no conversion occurs if the
                // data is not alphabetic).
                charCode = tolower( charCode );
             }
             pData->fMappedCharCode = (char) charCode;
          }
       }
       break;
    default:
       break;
  }
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::IKeyboardEvent                                               |
|                                                                              |
| Copy Constructor. (DEFECT 25402)                                             |
------------------------------------------------------------------------------*/
IKeyboardEvent::IKeyboardEvent ( const IKeyboardEvent& event )
  : IKeyboardEvent::Inherited( event ),
    fValidComposite( event.fValidComposite ),
    fInvalidComposite( event.fInvalidComposite ),
    pData( new IKeyboardEventPrivateData( *(event.pData) ) )
{
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::IKeyboardEvent                                               |
|                                                                              |
| Default Constructor. (DEFECT 25402)                                          |
------------------------------------------------------------------------------*/
IKeyboardEvent::IKeyboardEvent ( )
  : IKeyboardEvent::Inherited( 0,0,0,0 ),
    fValidComposite( false ),
    fInvalidComposite( false ),
    pData( new IKeyboardEventPrivateData )
{
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::~IKeyboardEvent                                              |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
IKeyboardEvent::~IKeyboardEvent ( )
{
  delete pData;
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::character                                                    |
|                                                                              |
| Return the SBCS character code for the event.  Before calling this function, |
| the user should first verify whether the event contains a character code by  |
| calling the isCharacter() function.                                          |
------------------------------------------------------------------------------*/
char IKeyboardEvent::character ( ) const
{
  if ( this->isCharacter() == false )
  {                               // No character data.
     ITHROWLIBRARYERROR( IC_INVALIDCHARCODE,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }

  if ( pData->fMappedCharCode )
  {
     return pData->fMappedCharCode;
  }
  else
  {
     return this->parameter1();
  }
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::mixedCharacter                                               |
|                                                                              |
| Return the SBCS or DBCS character code for the event.  Before calling this   |
| function, the user should first verify whether the event contains a          |
| character code by calling the isCharacter() function.                        |
------------------------------------------------------------------------------*/
IString IKeyboardEvent::mixedCharacter ( ) const
{
  if ( this->isCharacter() == false )
  {                               // No character data.
     ITHROWLIBRARYERROR( IC_INVALIDCHARCODE,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }

  if ( pData->fMappedCharCode )
  {
     return IString( pData->fMappedCharCode );
  }
  else
  {
     return IString( (char)( this->parameter1() ) );
  }
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::scanCode                                                     |
|                                                                              |
| Get the hardware scan code for the key.  Before calling this function, the   |
| user should first verify whether the event contains a scan code by calling   |
| the isScanCode() function.                                                   |
------------------------------------------------------------------------------*/
unsigned long IKeyboardEvent::scanCode ( ) const
{
  if ( this->isScanCode() == false )
  {         // No scan code data.
     ITHROWLIBRARYERROR( IC_INVALIDSCANCODE,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }

  return this->parameter2().char3();
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::repeatCount                                                  |
|                                                                              |
| Return the number of times this key event should be applied.                 |
------------------------------------------------------------------------------*/
unsigned long IKeyboardEvent::repeatCount ( ) const
{
  return this->parameter2().number1();
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::virtualKey                                                   |
|                                                                              |
| Get the virtual key value of the key event.  Before calling this function,   |
| the user should first verify whether the event contains a virtual key code   |
| by calling the isVirtual() function.                                         |
------------------------------------------------------------------------------*/
IKeyboardEvent::VirtualKey IKeyboardEvent::virtualKey ( ) const
{
  VirtualKey vkey = other;        // Assume no match found.

  if ( this->isVirtual() == false )
  {                               // Not a virtual key.
     ITHROWLIBRARYERROR( IC_INVALIDVIRTUALKEY,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }

  switch ( this->parameter1() )
  {                               // Switch on the virtual key code.
     case VK_BUTTON1:
        vkey = button1;
        break;
     case VK_BUTTON2:
        vkey = button2;
        break;
     case VK_BUTTON3:
        vkey = button3;
        break;
     case VK_BREAK:
        vkey = breakKey;
        break;
     case VK_BACKSPACE:
        vkey = backSpace;
        break;
     case VK_TAB:
        vkey = tab;
        // Check shift state to determine type of tab.
        if (( this->isShiftDown() ) && ( this->isCtrlDown() == false ) &&
            ( this->isAltDown() == false ))
           vkey = backTab;
        break;
     case VK_SHIFT:
        vkey = shift;
        break;
     case VK_CTRL:
        vkey = ctrl;
        break;
     case VK_PAUSE:
        vkey = pause;
        break;
     case VK_CAPSLOCK:
        vkey = capsLock;
        break;
     case VK_ESC:
        vkey = esc;
        break;
     case VK_SPACE:
        vkey = space;
        break;
     case VK_PAGEUP:
        vkey = pageUp;
        break;
     case VK_PAGEDOWN:
        vkey = pageDown;
        break;
     case VK_END:
        vkey = end;
        break;
     case VK_HOME:
        vkey = home;
        break;
     case VK_LEFT:
        vkey = left;
        break;
     case VK_UP:
        vkey = up;
        break;
     case VK_RIGHT:
        vkey = right;
        break;
     case VK_DOWN:
        vkey = down;
        break;
     case VK_INSERT:
        vkey = insert;
        break;
     case VK_DELETE:
        vkey = deleteKey;
        break;
     case VK_SCRLLOCK:
        vkey = scrollLock;
        break;
     case VK_NUMLOCK:
        vkey = numLock;
        break;
     case VK_ENTER:
        vkey = enter;
        // Check for extended key bit on.
        if ( this->parameter2().asUnsignedLong() & 0x01000000 )
           vkey = newLine;
        break;
     case VK_F1:
        vkey = f1;
        break;
     case VK_F2:
        vkey = f2;
        break;
     case VK_F3:
        vkey = f3;
        break;
     case VK_F4:
        vkey = f4;
        break;
     case VK_F5:
        vkey = f5;
        break;
     case VK_F6:
        vkey = f6;
        break;
     case VK_F7:
        vkey = f7;
        break;
     case VK_F8:
        vkey = f8;
        break;
     case VK_F9:
        vkey = f9;
        break;
     case VK_F10:
        vkey = f10;
        break;
     case VK_F11:
        vkey = f11;
        break;
     case VK_F12:
        vkey = f12;
        break;
     case VK_F13:
        vkey = f13;
        break;
     case VK_F14:
        vkey = f14;
        break;
     case VK_F15:
        vkey = f15;
        break;
     case VK_F16:
        vkey = f16;
        break;
     case VK_F17:
        vkey = f17;
        break;
     case VK_F18:
        vkey = f18;
        break;
     case VK_F19:
        vkey = f19;
        break;
     case VK_F20:
        vkey = f20;
        break;
     case VK_F21:
        vkey = f21;
        break;
     case VK_F22:
        vkey = f22;
        break;
     case VK_F23:
        vkey = f23;
        break;
     case VK_F24:
        vkey = f24;
        break;

     // The below virtual keys are not supported because a
     // keyboard handler will not get the key press (F1, F10),
     // release (Alt), or both (PrintScreen, though maybe it can
     // get the key release) of these keys.  Supporting
     // enumerators for them implies that all presses and
     // releases of these keys can be processed.
     // The application can still process the non-missing events
     // for the below keys by overriding the key() function.
     case VK_ALT:                 // Fall into default case.
     case VK_PRINTSCRN:           // Fall into default case.
     default:
        break;
  } /* endswitch */

  return vkey;
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::isCharacter                                                  |
|                                                                              |
| Tell the caller whether this keyboard event is for a character key.          |
| To maintain compatibility with OS/2 design, certain keys which Windows       |
| converts into character messages will be excluded and will show up as        |
| virtual keys                                                                 |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isCharacter ( ) const
{
  Boolean characterData = false;

  if ( pData->fMappedCharCode )
  {
     // Treat a virtual key generated from Ctrl + character key as a
     // character key.
     characterData = true;
  }
  else if ( ( ( this->eventId() == WM_CHAR ) ||
              ( this->eventId() == WM_SYSCHAR ) ) &&
            ( pData->fSpecialCharacter == false ) )
  {
     characterData = true;
  }
  return characterData;
}

/*------------------------------------------------------------------------------
| IKeyboardEvent::isVirtual                                                    |
|                                                                              |
| Tell the caller whether this keyboard event is for a virtual key.            |
| To maintain compatibility with OS/2 design, certain keys which Windows       |
| converts into character messages will be included.                           |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isVirtual ( ) const
{
  Boolean vkeyData = false;

  if ( ! pData->fMappedCharCode )
  {
     unsigned long eventIdentifier = this->eventId();
     if ( pData->fSpecialCharacter )
     {
        if ( ( eventIdentifier == WM_CHAR )  ||
             ( eventIdentifier == WM_SYSCHAR ) )
        {
           vkeyData = true;
        }
        else if ( this->isCtrlDown() )
        {
           vkeyData = true;
        }
     }
     else if ( ( eventIdentifier == WM_KEYDOWN )  ||
               ( eventIdentifier == WM_SYSKEYDOWN )  ||
               ( eventIdentifier == WM_KEYUP )  ||
               ( eventIdentifier == WM_SYSKEYUP ) )
     {
        vkeyData = true;
     }
  }

  return vkeyData;
}

/*------------------------------------------------------------------------------
| IBase::Boolean  IKeyboardEvent :: isShiftDown                                |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isShiftDown ( ) const
{
  return (( GetKeyState(VK_SHIFT) & 0x8000 ) ? true : false );
}

/*------------------------------------------------------------------------------
| IBase::Boolean  IKeyboardEvent :: isCtrlDown                                 |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isCtrlDown ( ) const
{
  return (( GetKeyState(VK_CONTROL) & 0x8000 ) ? true : false );
}

/*------------------------------------------------------------------------------
| IBase::Boolean  IKeyboardEvent :: isForComposite                             |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isForComposite ( ) const
{
  return ((( this->eventId() == WM_DEADCHAR ) ||
           ( this->eventId() == WM_SYSDEADCHAR )) ? true : false);
}

/*------------------------------------------------------------------------------
| IBase::Boolean  IKeyboardEvent :: isUncombined                               |
------------------------------------------------------------------------------*/
IBase::Boolean IKeyboardEvent::isUncombined ( ) const
{
  return (!(fValidComposite || fInvalidComposite));
}
