/*******************************************************************************
* FILE NAME: imousevt.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in imousevt.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_WINFRAMEMGR
#define INCL_WINWINDOWMGR
#define INCL_WININPUT
extern "C"
  {
  #include <iwindefs.h>
  }

#include <imousevt.hpp>
#include <icconst.h>
#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <ihandle.hpp>
#include <ipoint.hpp>
#include <iwindow.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IMOUSEVT_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IMouseEvent::IMouseEvent                                                     |
|                                                                              |
| Constructor here for page tuning.                                            |
------------------------------------------------------------------------------*/
IMouseEvent::IMouseEvent ( const IEvent& event )
  : IMouseEvent::Inherited ( event )
{ }


/*------------------------------------------------------------------------------
| IMouseEvent::~IMouseEvent                                                    |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IMouseEvent::~IMouseEvent ( )
{ }


/*------------------------------------------------------------------------------
| IMouseEvent::mousePosition                                                   |
|                                                                              |
| Return the mouse pointer position.                                           |
------------------------------------------------------------------------------*/
IPoint IMouseEvent::mousePosition ( ) const
{
#ifdef IC_WIN
   IPoint pt((short)parameter2().number1(), (short)parameter2().number2());
   RECT rect;
   GetWindowRect(handle(), &rect);
   ISize size(rect.right - rect.left, rect.bottom - rect.top );
#else
   IPoint pt((short)parameter1().number1(), (short)parameter1().number2());
   SWP swp;
   WinQueryWindowPos(handle(), &swp);
   ISize size(swp.cx,swp.cy);
#endif
   return ICoordinateSystem::convertToApplication( pt, size );
}

/*------------------------------------------------------------------------------
| IMouseEvent::windowUnderPointer                                              |
|                                                                              |
| Get the window handle that the mouse is positioned over.                     |
------------------------------------------------------------------------------*/
IWindowHandle IMouseEvent::windowUnderPointer ( ) const
{
#ifdef IC_WIN
   RECT rect;
   GetWindowRect(handle(), &rect);
   ISize size(rect.right - rect.left, rect.bottom - rect.top );
#else
   SWP swp;
   WinQueryWindowPos(handle(), &swp);
   ISize size(swp.cx,swp.cy);
#endif
  POINTL ptMouse =
     ICoordinateSystem::convertToNative(  mousePosition(),
                                          size ).asPOINTL();
#ifdef IC_WIN
  // Map the point to screen coordinates.  This is necessary
  // because the macro IWINDOWFROMPOINT was changed to use
  // WindowFromPoint instead of ChildWindowFromPoint.  It appears
  // that ChildWindowFromPoint does not return any window beyond
  // its immediate children.
  IMAPWINDOWPOINTS( handle(), 0, &ptMouse, 1 );
#endif

  return IWINDOWFROMPOINT(handle(), &ptMouse, true);
  // no exception thrown as handle is always valid in event object
}

/*------------------------------------------------------------------------------
| IMouseEvent::isAltKeyDown                                                    |
|                                                                              |
| Determine the action taken on the mouse.                                     |
------------------------------------------------------------------------------*/
Boolean IMouseEvent::isAltKeyDown( ) const
{
#ifdef IC_WIN
   return (Boolean)HIBYTE(GetKeyState(VK_ALT));
#else
   return parameter2().number2() & KC_ALT;
#endif
}

/*------------------------------------------------------------------------------
| IMouseEvent::isCtrlKeyDown                                                   |
|                                                                              |
| Determine the action taken on the mouse.                                     |
------------------------------------------------------------------------------*/
Boolean IMouseEvent::isCtrlKeyDown( ) const
{
#ifdef IC_WIN
   return parameter1() & MK_CONTROL;
#else
   return parameter2().number2() & KC_CTRL;
#endif
}

/*------------------------------------------------------------------------------
| IMouseEvent::isShiftKeyDown                                                  |
|                                                                              |
| Determine the action taken on the mouse.                                     |
------------------------------------------------------------------------------*/
Boolean IMouseEvent::isShiftKeyDown( ) const
{
#ifdef IC_WIN
   return parameter1() & MK_SHIFT;
#else
   return parameter2().number2() & KC_SHIFT;
#endif
}

/*------------------------------------------------------------------------------
| IMouseClickEvent::IMouseClickEvent                                           |
|                                                                              |
| Constructor here for page tuning.                                            |
------------------------------------------------------------------------------*/
IMouseClickEvent::IMouseClickEvent ( const IEvent& event )
  : IMouseClickEvent::Inherited ( event )
{ }


/*------------------------------------------------------------------------------
| IMouseClickEvent::~IMouseClickEvent                                          |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IMouseClickEvent::~IMouseClickEvent ( )
{ }

/*------------------------------------------------------------------------------
| IMouseClickEvent::mouseButton                                                |
|                                                                              |
| Determine which mouse button is generating the event.                        |
------------------------------------------------------------------------------*/
IMouseClickEvent::MouseButton  IMouseClickEvent::mouseButton ( ) const
{
   MouseButton button = button1;  // default value
   switch (eventId())
   {
      case WM_BUTTON1DOWN:
      case WM_BUTTON1UP:
      case WM_BUTTON1CLICK:
      case WM_BUTTON1DBLCLK:
         break;                   // use default value

      case WM_BUTTON2DOWN:
      case WM_BUTTON2UP:
      case WM_BUTTON2CLICK:
      case WM_BUTTON2DBLCLK:
         button = button2;
         break;

      case WM_BUTTON3DOWN:
      case WM_BUTTON3UP:
      case WM_BUTTON3CLICK:
      case WM_BUTTON3DBLCLK:
         button = button3;
         break;

      case WM_CHORD:
         button = buttonChord;
         break;

      default:
         ITHROWLIBRARYERROR(IC_INVALIDEVENT,
                            IBaseErrorInfo::invalidRequest,
                            IException::recoverable);
         break;
   } /* endswitch */

   return button;
}


/*------------------------------------------------------------------------------
| IMouseClickEvent::mouseAction                                                |
|                                                                              |
| Determine the action taken on the mouse.                                     |
------------------------------------------------------------------------------*/
IMouseClickEvent::MouseAction  IMouseClickEvent::mouseAction ( ) const
{
   MouseAction action = down;     // default value
   switch (eventId())
   {
      case WM_BUTTON1DOWN:
      case WM_BUTTON2DOWN:
      case WM_BUTTON3DOWN:
      case WM_CHORD:
         break;                   // use default value
      case WM_BUTTON1UP:
      case WM_BUTTON2UP:
      case WM_BUTTON3UP:
         action = up;
         break;

      case WM_BUTTON1CLICK:
      case WM_BUTTON2CLICK:
      case WM_BUTTON3CLICK:
         action = click;
         break;

      case WM_BUTTON1DBLCLK:
      case WM_BUTTON2DBLCLK:
      case WM_BUTTON3DBLCLK:
         action = doubleClick;
         break;
      default:
         ITHROWLIBRARYERROR(IC_INVALIDEVENT,
                            IBaseErrorInfo::invalidRequest,
                            IException::recoverable);
         break;
   } /* endswitch */

   return action;
}

/*------------------------------------------------------------------------------
| IMousePointerEvent::IMousePointerEvent                                       |
|                                                                              |
| Constructor here for page tuning.                                            |
------------------------------------------------------------------------------*/
IMousePointerEvent::IMousePointerEvent ( const IEvent& event )
  : IMousePointerEvent::Inherited ( event )
{ }


/*------------------------------------------------------------------------------
| IMousePointerEvent::~IMousePointerEvent                                      |
|                                                                              |
| Empty destructor here for page tuning.                                       |
------------------------------------------------------------------------------*/
IMousePointerEvent::~IMousePointerEvent ( )
{ }


/*------------------------------------------------------------------------------
| IMousePointerEvent::controlId                                                |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IMousePointerEvent::windowId( ) const
{
#ifdef IC_WIN
  return IIDOF( parameter1() );
#else
  return parameter1().number1();
#endif
}

/*------------------------------------------------------------------------------
| IMousePointerEvent::setMousePointer                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IMousePointerEvent& IMousePointerEvent::setMousePointer(
                                            const IPointerHandle& mousePointer )
{
#ifdef IC_WIN
  ISETPOINTER( mousePointer );
  setResult( true );
#else
  setResult((unsigned long)mousePointer);
#endif
  return *this;
}

/*------------------------------------------------------------------------------
| IMousePointerEvent::defaultMousePointer                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointerHandle IMousePointerEvent::defaultMousePointer( ) const
{
#ifdef IC_WIN
  return IPointerHandle( GetCursor() );
#else
  return IPointerHandle((unsigned long)parameter2());
#endif
}
