/*******************************************************************************
* FILE NAME: ioutlbox.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ioutlbox.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define INCL_WINERRORS
  #define INCL_WINSTATICS
  #define INCL_WINSYS
  #include <iwindefs.h>
}

#ifndef _IOUTLBOX_
  #include <ioutlbox.hpp>
#endif
#ifndef _ICCONST_
  #include <icconst.h>
#endif
#ifndef _IEXCEPT_
  #include <iexcept.hpp>
#endif
#ifndef _IHANDLE_
  #include <ihandle.hpp>
#endif
#ifndef _IPOINT_
  #include <ipoint.hpp>
#endif
#ifndef _IRECT_
  #include <irect.hpp>
#endif
#ifndef _ICOLOR_
  #include <icolor.hpp>
#endif
#ifndef _ICMNFUN_
  #include <icmnfun.hpp>
#endif
#ifndef _INOTIFEV_
  #include <inotifev.hpp>
#endif
#ifdef IC_WIN
  #include <iplatfrm.hpp>
#endif


// Segment definitions
#ifdef IC_PAGETUNE
  #define _IOUTLBOX_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| Public outline box styles.                                                   |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLStaticConst)
const IOutlineBox::Style
  IOutlineBox::backgroundFrame   ( 0, IOBS_BACKGROUNDFRAME ),
  IOutlineBox::foregroundFrame   ( 0, IOBS_FOREGROUNDFRAME ),
  IOutlineBox::halftoneFrame     ( 0, IOBS_HALFTONEFRAME ),
#ifdef IC_WIN
  IOutlineBox::border3D          ( 0, WS_EX_STATICEDGE ),
#endif
#ifndef IC_WIN
  IOutlineBox::border3D          ( 0, 0 ),
#endif
#ifdef IC_WIN
  IOutlineBox::classDefaultStyle ( WS_VISIBLE,
                                   IOBS_FOREGROUNDFRAME | WS_EX_STATICEDGE );
#endif
#ifndef IC_WIN
  IOutlineBox::classDefaultStyle ( WS_VISIBLE, IOBS_FOREGROUNDFRAME );
#endif
#pragma data_seg()

/*------------------------------------------------------------------------------
| Default style for new objects (initial value).                               |
------------------------------------------------------------------------------*/
#pragma data_seg(ICLNonConst)
#ifdef IC_WIN
IOutlineBox::Style
  IOutlineBox::currentDefaultStyle ( WS_VISIBLE,
                                     IOBS_FOREGROUNDFRAME | WS_EX_STATICEDGE );
#endif
#ifndef IC_WIN
IOutlineBox::Style
  IOutlineBox::currentDefaultStyle ( WS_VISIBLE, IOBS_FOREGROUNDFRAME );
#endif
#pragma data_seg()


/*------------------------------------------------------------------------------
| IOutlineBox::IOutlineBox                                                     |
|                                                                              |
| Constructor to create an outline box control on a standard window.           |
------------------------------------------------------------------------------*/
IOutlineBox :: IOutlineBox(unsigned long ulId,
                           IWindow* pwndParent,
                           IWindow* pwndOwner,
                           const IRectangle& rectInit,
                           const Style& obsStyle)
                         : fOutlineBoxData( 0 )
{
  // assertions on input parms
  IASSERTPARM(pwndParent!=0);
  IASSERTPARM((obsStyle & (backgroundFrame | foregroundFrame |
                           halftoneFrame))!=0);

  // Save the extended style to make sure we have a copy of it stored
  setExtendedStyle( extendedStyle() | obsStyle.asExtendedUnsignedLong() );

  IWindowHandle whOutlineBox =
      this -> create( ulId,
                      0,
                      convertToGUIStyle( obsStyle ),
                      WC_STATIC,
                      pwndParent->handle(),
                      (pwndOwner==0)?IWindowHandle(0):pwndOwner->handle(),
                      rectInit,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( obsStyle, true ) );

  startHandlingEventsFor(whOutlineBox);

  //------------------------------------------------------------
  // DEFECT 23723 (ugly fix, but it works)
  //    Work-around for MS bug on ArabicWin95; style resulting
  //    after the create() is incorrect for Static Style value.
  //    (Only need to do it if we're in BiDirectional mode.)
  //------------------------------------------------------------
  if (IBidiSettings::isBidiSupported()) {
    setOutlineType( outlineType() );
  }
  //------------------------------------------------------------
}


/*------------------------------------------------------------------------------
| IOutlineBox::IOutlineBox                                                     |
|                                                                              |
| Constructor to create instance for a template window.                        |
------------------------------------------------------------------------------*/
IOutlineBox :: IOutlineBox(unsigned long ulId,
                           IWindow* pdlgwndParent)
                         : fOutlineBoxData( 0 )
{
  setAutoDestroyWindow( false );
  startHandlingEventsFor(ulId, pdlgwndParent);
}


/*------------------------------------------------------------------------------
| IOutlineBox::IOutlineBox                                                     |
|                                                                              |
| Constructor to instantiate from an existing outline box control.             |
------------------------------------------------------------------------------*/
IOutlineBox :: IOutlineBox(const IWindowHandle& wh)
                         : fOutlineBoxData( 0 )
{
  setAutoDestroyWindow( false );
  startHandlingEventsFor(wh);
}

/*------------------------------------------------------------------------------
| IOutlineBox::~IOutlineBox                                                    |
|                                                                              |
| Destructor for cleanup.                                                      |
------------------------------------------------------------------------------*/
IOutlineBox :: ~IOutlineBox( )
{
  // delete fOutlineBoxData;
}


/*------------------------------------------------------------------------------
| IOutlineBox::defaultStyle                                                    |
|                                                                              |
| Return the default style for new outline box objects.                        |
------------------------------------------------------------------------------*/
IOutlineBox::Style  IOutlineBox :: defaultStyle()
{
  return currentDefaultStyle;
}


/*------------------------------------------------------------------------------
| IOutlineBox::setDefaultStyle                                                 |
|                                                                              |
| Replace the default style for new outline box objects.                       |
------------------------------------------------------------------------------*/
void  IOutlineBox :: setDefaultStyle(const IOutlineBox::Style& obsStyle)
{
  currentDefaultStyle = obsStyle;
}


/*------------------------------------------------------------------------------
| IOutlineBox::convertToGUIStyle                                               |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long IOutlineBox::convertToGUIStyle(const IBitFlag& guiStyle,
                                             Boolean bExtOnly) const
{
  // Obtain the style from the class (IControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( guiStyle, bExtOnly );

  if (bExtOnly)
  {
    // Use mask to only return extended styles in the user defined range
    ulStyle |= guiStyle.asExtendedUnsignedLong() & ( IS_EXTMASK | IOB_EXTGUIMASK);
//    if ((IPlatform::isWin95()) || (IPlatform::isNTNewShell()))
//       ulStyle |= WS_EX_STATICEDGE;
  }
  else
  {
    // Let the clean styles flow thru
    ulStyle |= guiStyle.asUnsignedLong();

    // Get the extended styles and set the GUI style accordingly ...
    unsigned long ulTmp = guiStyle.asExtendedUnsignedLong();
    //------------------------------------------------------------
    // DEFECT 23723
    //   Enforce that only ONE of these 3 Static Styles gets set.
    //------------------------------------------------------------
    ulStyle &= ~(SS_HALFTONEFRAME | SS_BKGNDFRAME | SS_FGNDFRAME);
    //------------------------------------------------------------
    if (ulTmp & backgroundFrame.asExtendedUnsignedLong())
    {
     ulStyle |= SS_BKGNDFRAME;
    }
    else if (ulTmp & halftoneFrame.asExtendedUnsignedLong())
    {
     ulStyle |= SS_HALFTONEFRAME;
    }
    else if (ulTmp & foregroundFrame.asExtendedUnsignedLong())
    {
     ulStyle |= SS_FGNDFRAME;
    }
    //------------------------------------------------------------
#ifdef IC_WIN
    ulStyle |= WS_CHILD;
#endif
  }

  return( ulStyle );
}


/*------------------------------------------------------------------------------
| IOutlineBox::setOutlineType                                                  |
|                                                                              |
| Sets the type of outline.                                                    |
------------------------------------------------------------------------------*/
IOutlineBox&  IOutlineBox :: setOutlineType(const OutlineType& outlineType)
{
  unsigned long ulStyles = style();
  unsigned long ulOldStyles = ulStyles;
  ulStyles &= ~(SS_HALFTONEFRAME | SS_BKGNDFRAME | SS_FGNDFRAME);

  /*************************************************************/
  /* Replace with new outline type.                            */
  /*************************************************************/
  switch (outlineType)
  {
    case foreground:
      ulStyles |= SS_FGNDFRAME;
    break;

    case background:
      ulStyles |= SS_BKGNDFRAME;
    break;

    case halftone:
      ulStyles |= SS_HALFTONEFRAME;
    break;

    default:                          //Uninitialized enum
      ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
    break;
  }

  if (ulStyles != ulOldStyles)
  {
    setStyle(ulStyles);                 //Change style
    refresh();                          //Force repaint
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IOutlineBox::outlineType                                                     |
------------------------------------------------------------------------------*/
IOutlineBox::OutlineType  IOutlineBox :: outlineType() const
{
  unsigned long ulStyles =
     style() & (SS_HALFTONEFRAME | SS_BKGNDFRAME | SS_FGNDFRAME);
  if (ulStyles == SS_BKGNDFRAME)
     return background;
  if (ulStyles == SS_HALFTONEFRAME)
     return halftone;
  return foreground;
}

/*------------------------------------------------------------------------------
| WIN32 Color Support                                                          |
|                                                                              |
| Windows has no API equivalent for Presentation Parameters.  The following    |
| outline box color function is NOPed for Windows.  The query function returns |
| an appropriate default IColor object.                                        |
------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------
| IOutlineBox::foregroundColor                                                 |
|                                                                              |
| Return the foreground color of the outline box based on the outline type.    |
------------------------------------------------------------------------------*/
IColor IOutlineBox::foregroundColor () const
{
  IGUIColor guiColor(IGUIColor::windowBgnd);

  switch (outlineType())
  {
    case background:
#ifdef IC_PM
      return (IWindow::color(PP_FOREGROUNDCOLOR,
                             defaultbackgroundcolor(handle())));
#else
      return (defaultbackgroundcolor(handle()));
#endif // IC_PM

    case foreground:
      guiColor = IGUIColor(IGUIColor::frameBorder);
    break;

    case halftone:
      guiColor = IGUIColor(IGUIColor::desktopBgnd);
    break;
  }
#ifdef IC_PM
  return (IWindow::color(PP_FOREGROUNDCOLOR, guiColor));
#else
  return (guiColor);
#endif // IC_PM
}

/*------------------------------------------------------------------------------
| IOutlineBox::visibleRectangle                                                |
|                                                                              |
| Returns the area painted by the outline box.                                 |
------------------------------------------------------------------------------*/
IRectangle IOutlineBox :: visibleRectangle () const
{
  return IRectangle(position(),ISize(0,0));
}

