/*******************************************************************************
* FILE NAME: iwindow4.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iwindow.hpp that are Windows unique.                                    *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1996       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

extern "C" {
  #define IUSING_OLE
  #ifdef WIN32_LEAN_AND_MEAN
    #undef WIN32_LEAN_AND_MEAN
  #endif
  #ifdef _OLE2_H_
    #undef _OLE2_H_
  #endif
  #include <iwindefs.h>
}

#ifdef IC_WIN  // Ensure clean compile in PM/AIX environment

#include <iwindow.hpp>
#include <iwinpriv.hpp>
#include <iwinlsts.hpp>

#include <icconst.h>
#include <icolor.hpp>
#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <iplatfrm.hpp>
#include <irect.hpp>
#include <istring.hpp>
#include <ithread.hpp>
#include <itrace.hpp>
#include <iwcname.hpp>
#include <icmnfun.hpp>
#include <imcevt.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IWINDOW4_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
|ITheObjectWindow::ITheObjectWindow                                            |
------------------------------------------------------------------------------*/
ITheObjectWindow::ITheObjectWindow()
   : hwnd(0)
{
   IWindowHandle focusWindow = IQUERYFOCUS(0);
   hwnd = ICREATEWINDOW(
              0,                        // parent
              WC_STATIC,                // class
              0,                        // window text
              WS_POPUP | SS_GRAYRECT,   // Note not WS_VISIBLE
              0,                        // x
              0,                        // y
              0,                        // width
              0,                        // height
              0,                        // owner (unused in WIN)
              0,                        // behind (unused in WIN)
              0,                        // id/menu handle
              0,                        // window creation data
              0,                        // unused in win
              0     );                  // hInst
   if (hwnd == 0)
      {
      ITHROWGUIERROR2("CreateWindow",
                      IBaseErrorInfo::accessError,
                      IException::recoverable);
      }
   // restore focus back to original window
   ISETFOCUS(0, focusWindow);
}


/*------------------------------------------------------------------------------
| IWindow::size                                                                |
|                                                                              |
| Get the window size.                                                         |
------------------------------------------------------------------------------*/
ISize IWindow::size() const
{
   return rect().size();
}

/*------------------------------------------------------------------------------
| IWindow::parentSize                                                          |
------------------------------------------------------------------------------*/
ISize IWindow::parentSize ( ) const
{
   return parentSize( this->handle() );
}

/*------------------------------------------------------------------------------
| IWindow::parentSize                                                          |
------------------------------------------------------------------------------*/
ISize IWindow::parentSize ( const IWindowHandle& windowhandle )
{
   IMODTRACE_ALL("parentSize");
   IWindowHandle  parentHwnd = IPARENTOF( windowhandle );
   if ( parentHwnd == IWindowHandle(0) )
      {
      parentHwnd = IWindow::desktopWindow()->handle();
      }

   // the size needed is the size of the coordinate space which is used
   // to position child windows.
#ifdef   IC_WIN
   // If the window is a popup, use the desktop as the parent since
   //    1) By default this window has no parent
   //    2) A query for the parent always returns the owner window
   if ( ISTYLEOF( windowhandle ) & WS_POPUP )
     parentHwnd = IWindow::desktopWindow()->handle();

   RECT rect;
   GetClientRect(parentHwnd, &rect);

   ITRACE_ALL( IString("parentSize=") +
               ISize(rect.right - rect.left, rect.bottom - rect.top ).asString() );
   return  ISize(rect.right - rect.left, rect.bottom - rect.top );
#endif   //IC_WIN
#ifdef   IC_PM
   SWP swp;
   WinQueryWindowPos(parentHwnd, &swp);
   ITRACE_ALL( IString("parentSize=") +
               ISize(swp.cx, swp.cy ).asString() );
   return (ISize(swp.cx,swp.cy));
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| IWindow::position                                                            |
|                                                                              |
| Gets the object position.                                                    |
------------------------------------------------------------------------------*/
IPoint IWindow::position() const
{
   return rect().minXMinY();
}

/*------------------------------------------------------------------------------
| IWindow::rect                                                                |
|                                                                              |
| Gets the object position and size.                                           |
------------------------------------------------------------------------------*/
IRectangle IWindow::rect() const
{
   if ( ICoordinateSystem::isConversionNeeded() )
      {
      return ICoordinateSystem::convertToApplication( nativeRect( ),
                                                      this->parentSize( ) );
      }
   else
      {
      return nativeRect( );
      }
}

/*------------------------------------------------------------------------------
| IWindow::nativeRect                                                          |
------------------------------------------------------------------------------*/
IRectangle IWindow::nativeRect ( ) const
{
   RECT   rect;
   IQUERYWINDOWRECT(handle(), &rect);
   IWindowHandle  parentHwnd = IPARENTOF( handle() );

#ifdef   IC_WIN
   // If the window is a popup, use the desktop as the parent since
   //    1) By default this window has no parent
   //    2) A query for the parent always returns the owner window
   if ( ISTYLEOF( handle() ) & WS_POPUP )
     parentHwnd = IWindow::desktopWindow()->handle();
#endif   //IC_WIN

   MapWindowPoints( HWND_DESKTOP, parentHwnd, (POINT*)&rect, 2);
   return IRectangle( rect );
}

/*------------------------------------------------------------------------------
| IWindow::sizeTo                                                              |
|                                                                              |
| Size the window.                                                             |
------------------------------------------------------------------------------*/
IWindow& IWindow::sizeTo(const ISize& siz)
{
   IMODTRACE_ALL("Win::sizeTo");
   ITRACE_ALL(siz.asString());
   moveSizeTo( IRectangle( position(), siz ) );
   return *this;
}


/*------------------------------------------------------------------------------
| IWindow::moveTo                                                              |
|                                                                              |
| Move the object.                                                             |
------------------------------------------------------------------------------*/
IWindow& IWindow::moveTo(const IPoint& pt)
{
   IMODTRACE_ALL("Win::moveTo");
   ITRACE_ALL(pt.asString());
   moveSizeTo(IRectangle( pt, size()) );
   return *this;
}

/*------------------------------------------------------------------------------
| IWindow::moveSizeTo                                                          |
|                                                                              |
| Move and size the object.                                                    |
------------------------------------------------------------------------------*/
IWindow& IWindow::moveSizeTo(const IRectangle& rect)
{
   IMODTRACE_ALL("Win::moveSizeTo");
   ITRACE_ALL(rect.asString());

   IWindowHandle  hwnd  = this->handle();

   // Sorry, but you need to mess about with display drivers to change
   // desktop window size.
   if ( hwnd == IWindow::desktopWindow()->handle() )
      return *this;

   // recoordination ... translate to native coordinate system
   IRectangle nativeRectangle = rect;
   if ( ICoordinateSystem::isConversionNeeded() )
      {
      nativeRectangle = ICoordinateSystem::convertToNative( rect,
                                                            parentSize() );
      }

   //------------------------------------------------------------------------------
   // DEFECT 9451
   // preserve current position prior to the SetWindowPos call
   //------------------------------------------------------------------------------
   IPoint wPos = position();
   //------------------------------------------------------------------------------

   // set the window position
   ITRACE_ALL( IString("SetWindowPos hwnd=") +
               IString(hwnd.asUnsigned()).d2x() +
               IString(" position=") +
               nativeRectangle.minXMinY().asString() +
               IString(" size=") +
               nativeRectangle.size().asString()  );
   if (! SetWindowPos( hwnd, 0,
                       (int)nativeRectangle.minX(),
                       (int)nativeRectangle.minY(),  // really X origin
                       (int)nativeRectangle.width(),
                       (int)nativeRectangle.height(),
                       SWP_NOZORDER | SWP_NOACTIVATE ) )
      {
      ITHROWGUIERROR2("SetWindowPos",
                      IBaseErrorInfo::invalidParameter,
                      IException::recoverable);
      }

#ifdef IC_WIN
   //------------------------------------------------------------------------------
   // DEFECT 9451
   // if they resize smaller than Windows can accomodate...
   // in UpperLeft orientation, window does not move or change height
   // in LowerLeft orientation, y-position is adjusted, then height restored to min
   // (so, in this case, force it back to original location, so both behave same).
   //------------------------------------------------------------------------------
   {
      ISize newSize = size();
      ISize reqSize = ISize(rect.width(), rect.height());
      IPoint newPos = position();
      IPoint reqPos = rect.topLeft();

      if ( (newSize != reqSize) && (newPos != reqPos) ) {
         // ReSize failed, and caused a Size
         return( moveTo( wPos ) );
      }
   }
   //------------------------------------------------------------------------------
#endif

#ifdef IC_TRACE_DEVELOP
   RECT tRect;
   GetWindowRect( hwnd, &tRect);
   IRectangle  newRect(tRect);
   // GetWindowRect always returns screen coordinates
   // and not relative to parent.
   ITRACE_DEVELOP( IString(" real rect=") + newRect.asString() );
#endif

   return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setColor                                                            |
|                                                                              |
| Change the window attribute color using Red, Green and                       |
| Blue mix.                                                                    |
------------------------------------------------------------------------------*/
IWindow& IWindow::setColor(unsigned long ulColorArea, const IColor& color)
{
#ifdef IC_PM
   unsigned long ulClr=color.asRGBLong();
   WinSetPresParam (handle(), ulColorArea, sizeof(ulClr), &ulClr);
#endif // IC_PM
#ifdef IC_WIN
   if ((ulColorArea == PP_FOREGROUNDCOLOR) ||
       (ulColorArea == PP_BACKGROUNDCOLOR) )
   {
     if (ulColorArea == PP_FOREGROUNDCOLOR)
     {
       pWindowData->state |= frgndColorWasSet;
       pWindowData->clrforeground = color;
     }
     else if (ulColorArea == PP_BACKGROUNDCOLOR)
     {
       pWindowData->state |= bkgndColorWasSet;
       pWindowData->clrbackground = color;
     }

     /**************************************************************/
     /* Visually update the window                                 */
     /**************************************************************/
     setLayoutDistorted(IWindow::colorChanged, 0);
     refresh();

     /**************************************************************/
     /* Send private message to support window notify handler      */
     /* because Windows does not have presparam messages           */
     /**************************************************************/
     this->sendEvent( IC_UM_DELAYNOTIFY,
                      IEventParameter1( ulColorArea ),
                      IEventParameter2( IC_NOTDEL_WINDOW_COLOR ));
   }
#endif // IC_WIN
   return *this;
   // no exception b'cos handle is validated.
}

/*------------------------------------------------------------------------------
| IWindow::color                                                               |
|                                                                              |
| Return the color of window attribute.                                        |
------------------------------------------------------------------------------*/
IColor IWindow::color(unsigned long ulColorArea) const
{
#ifdef IC_PM
   unsigned long ulClr=0;
   unsigned long ulRetLen = WinQueryPresParam (handle(), ulColorArea, 0, 0,
                             sizeof(ulClr), &ulClr, 0);

   if (ulRetLen == 0) {
      ITHROWLIBRARYERROR(IC_PRESPARAMNOTFOUND,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable);
   } /* endif */

   unsigned char blue  = (unsigned char)(ulClr);
   unsigned char green = (unsigned char)(ulClr >> 8);
   unsigned char red   = (unsigned char)(ulClr >> 16);
   return IColor(red, green, blue);
#endif // IC_PM
#ifdef IC_WIN
   IColor clr( 0, 0, 0 );

   if ( (ulColorArea == PP_FOREGROUNDCOLOR) &&
        (pWindowData->state & frgndColorWasSet) )
   {
     clr = pWindowData->clrforeground;
   }
   else if ( (ulColorArea == PP_BACKGROUNDCOLOR) &&
             (pWindowData->state & bkgndColorWasSet) )
   {
     clr = pWindowData->clrbackground;
   }
   else if ( pWindowData->fhwndOwner )
   {
     IWindow* pwin = IWindow::windowWithHandle(pWindowData->fhwndOwner, false);
     if (pwin)
       clr = pwin->color(ulColorArea);
   }
   else if ( (ulColorArea == PP_FOREGROUNDCOLOR) ||
             (ulColorArea == PP_BACKGROUNDCOLOR) )
   {
      ITHROWLIBRARYERROR(IC_PRESPARAMNOTFOUND,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable);
   }
   return clr;
#endif // IC_WIN
}

/*------------------------------------------------------------------------------
| IWindow::color                                                               |
|                                                                              |
| Return the color of window attribute.                                        |
------------------------------------------------------------------------------*/
IColor IWindow::color(unsigned long ulColorArea, const IColor& defaultColor) const
{
#ifdef IC_PM
   unsigned long ulClr=0;
   unsigned long ulRetLen = WinQueryPresParam (handle(), ulColorArea, 0, 0,
                             sizeof(ulClr), &ulClr, 0);

   // if no presentation parameter has been set for this
   // area return the default color.
   if (ulRetLen == 0) {
      return defaultColor;
   } /* endif */

   unsigned char blue  = (unsigned char)(ulClr);
   unsigned char green = (unsigned char)(ulClr >> 8);
   unsigned char red   = (unsigned char)(ulClr >> 16);
   return IColor(red, green, blue);
#endif // IC_PM
#ifdef IC_WIN
   if ( (ulColorArea == PP_FOREGROUNDCOLOR) &&
        (pWindowData->state & frgndColorWasSet) )
   {
     return (pWindowData->clrforeground);
   }
   else if ( (ulColorArea == PP_BACKGROUNDCOLOR) &&
           (pWindowData->state & bkgndColorWasSet) )
   {
     return (pWindowData->clrbackground);
   }
   else if ( (ulColorArea == PP_BACKGROUNDCOLOR) &&
             (IWindowClassName(handle()) == WC_SCROLLBAR))
   {
       return (IGUIColor(IGUIColor::scrollBar));
   }
   else if ( ( pWindowData->fhwndOwner ) &&
             ( (IWindowClassName(handle()) != WC_ENTRYFIELD) &&
               (IWindowClassName(handle()) != WC_NATIVESPINBUTTON) &&
               (IWindowClassName(handle()) != WC_LISTBOX) &&
               (IWindowClassName(handle()) != WC_COMBOBOX) ) )
   {
     IWindow* pwin = IWindow::windowWithHandle(pWindowData->fhwndOwner, false);
     if (pwin)
       return (pwin->color(ulColorArea, defaultColor));
     else
       return defaultColor;
   }
   else
   {
     return defaultColor;
   }
#endif // IC_WIN
}

/*------------------------------------------------------------------------------
| IWindow::positionBehindSiblings                                              |
|                                                                              |
| Put this window in the Z order behind all of its siblings.                   |
------------------------------------------------------------------------------*/
IWindow& IWindow::positionBehindSiblings ( )
{

   IMODTRACE_DEVELOP("Win::positionBehindSiblings");
   SetWindowPos(handle(), HWND_BOTTOM, 0,0,0,0,
                SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

   return *this;
}

/*------------------------------------------------------------------------------
| IWindow::positionOnSiblings                                                  |
|                                                                              |
| Put this window in the Z order on top of all of its siblings.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::positionOnSiblings ( )
{

   IMODTRACE_DEVELOP("Win::positionOnSiblings");
   SetWindowPos(handle(), HWND_TOP, 0,0,0,0,
                SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE );    
   // DEFECT 20907 : use NOACTIVATE to match os/2 behavior

   return *this;
}

/*------------------------------------------------------------------------------
| IWindow::positionBehindSibling                                               |
|                                                                              |
| Put this window in the Z order behind the passed window.                     |
------------------------------------------------------------------------------*/
IWindow& IWindow::positionBehindSibling ( const IWindowHandle& siblingWindow)
{

   IMODTRACE_DEVELOP("Win::positionBehindSibling");
   SetWindowPos(handle(), siblingWindow, 0,0,0,0,
                SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

   return *this;
}


/*------------------------------------------------------------------------------
| IWindow::isShowing                                                           |
|                                                                              |
| Is the window showing?  True if the window not being moved                   |
| and some part of the window is currently on the screen.                      |
------------------------------------------------------------------------------*/
Boolean IWindow::isShowing() const
{
//   return WinIsWindowShowing(handle());

// Not supported in Windows.  For now if the window is visible, we will
// assume it is also showing
//
   return isVisible();
}

/*------------------------------------------------------------------------------
| IWindow::setOwner                                                            |
|                                                                              |
| Change the window's owner                                                    |
------------------------------------------------------------------------------*/
IWindow& IWindow::setOwner(const IWindow* newOwner)
{
   HWND hwndOwner = 0;
   if(newOwner != 0)
      {
      hwndOwner = newOwner->handle();
      }

   if ( ( ISTYLEOF( handle() ) & WS_CHILD )  )
      {
      // Simulated ownership used for child controls
      pWindowData->fhwndOwner = hwndOwner;
      }
   else
      {
      // You cannot change the ownership of a window which really has
      // an owner known to Windows ... that is, an overlapped or popup
      // window.  We ignore the change.
      hwndOwner = IOWNEROF( handle() );
      }

   HWND hwndParent = IPARENTOF(handle());

   if ( (hwndOwner == 0 || hwndOwner == desktopWindow()->handle())
                        &&
          hwndParent == desktopWindow()->handle() ) {
      pWindowData->state |= primaryWindow;
   } /* endif */
   else
      pWindowData->state &= (unsigned long)~primaryWindow;

   return *this;

   // no exception required b'cos handle() check for valid
   // window handle
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::ChildCursor                                            |
|                                                                              |
| The constructor simply saves the parent handle and initializes the cursor    |
| state to "invalid" (enumHandle = windowHandle == 0).                         |
------------------------------------------------------------------------------*/
IWindow::ChildCursor::ChildCursor ( IWindow &parent )
  : enumHandle( 0 ),
    hwnd( 0 ),
    hwndParent( parent.handle() )
{
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::~ChildCursor                                           |
|                                                                              |
| If the cursor enum handle is valid, issue WinEndEnumWindows.                 |
------------------------------------------------------------------------------*/
IWindow::ChildCursor::~ChildCursor ( )
{
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::setToFirst                                             |
|                                                                              |
------------------------------------------------------------------------------*/
Boolean IWindow::ChildCursor::setToFirst ( )
{

  if ( ( this->hwnd = GetWindow( this->hwndParent, GW_CHILD ) ) )
    return true;

  // No exception, since there either is a handle (return true) or
  // not (return false).
  return false;
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::setToNext                                              |
|                                                                              |
| Set the saved window handle to the value returned by WinGetNextWindow.       |
------------------------------------------------------------------------------*/
Boolean IWindow::ChildCursor::setToNext ( )
{
  // If First, get next, else return first
  if ( !this->hwnd )
     return setToFirst();
  else if ( ( this->hwnd = GetWindow( this->hwnd, GW_HWNDNEXT ) ) )
      // Got it, return true...
      return true;

  // Failure...
  return false;
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::isValid                                                |
|                                                                              |
| Return true if the saved window handle isn't null.                           |
------------------------------------------------------------------------------*/
Boolean IWindow::ChildCursor::isValid ( ) const
{
  return this->hwnd != 0;
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::invalidate                                             |
|                                                                              |
| If enum handle is valid, end the enumeration.                                |
------------------------------------------------------------------------------*/
void IWindow::ChildCursor::invalidate ( )
{
  this->hwnd = 0;
}

/*------------------------------------------------------------------------------
| IWindow::foregroundColor                                                     |
|                                                                              |
| Returns the foreground color of the window.                                  |
------------------------------------------------------------------------------*/
IColor IWindow::foregroundColor () const
{
  return (color(PP_FOREGROUNDCOLOR, IGUIColor(IGUIColor::windowText)));
}

/*------------------------------------------------------------------------------
| IWindow::backgroundColor                                                     |
|                                                                              |
| Returns the background color of the window.                                  |
------------------------------------------------------------------------------*/
IColor IWindow::backgroundColor () const
{
#ifdef IC_PM
  return (color(PP_BACKGROUNDCOLOR, IGUIColor(IGUIColor::windowBgnd)));
#else
  return (color(PP_BACKGROUNDCOLOR, defaultbackgroundcolor(handle())));
#endif // IC_PM
}

/*------------------------------------------------------------------------------
| IWindow::borderColor (NOPed for Windows; returns default)                    |
|                                                                              |
| Returns the border color of the window.                                      |
------------------------------------------------------------------------------*/
IColor IWindow::borderColor () const
{
//  return (color(PP_BORDERCOLOR, IGUIColor(IGUIColor::frameBorder)));
    return (IGUIColor(IGUIColor::frameBorder));
}

/*------------------------------------------------------------------------------
| IWindow::shadowColor  (NOPed for Windows; returns default)                   |
|                                                                              |
| Returns the shadow color of the window.                                      |
------------------------------------------------------------------------------*/
IColor IWindow::shadowColor () const
{
//  return (color(PP_SHADOW, IGUIColor(IGUIColor::dialogShadow)));
    return (IGUIColor(IGUIColor::dialogShadow));
}

/*------------------------------------------------------------------------------
| IWindow::disabledForegroundColor (NOPed for Windows; returns default)        |
|                                                                              |
| Returns the disabled foreground color of the window.                         |
------------------------------------------------------------------------------*/
IColor IWindow::disabledForegroundColor () const
{
//  return (color(PP_DISABLEDFOREGROUNDCOLOR,
//                IGUIColor(IGUIColor::defaultControl)));
    return (IGUIColor(IGUIColor::defaultControl));
}

/*------------------------------------------------------------------------------
| IWindow::disabledBackgroundColor (NOPed for Windows; returns default)        |
|                                                                              |
| Returns the disabled background color of the window.                         |
------------------------------------------------------------------------------*/
IColor IWindow::disabledBackgroundColor () const
{
//  return (color(PP_DISABLEDBACKGROUNDCOLOR,
//                IGUIColor(IGUIColor::windowBgnd)));
    return (IGUIColor(IGUIColor::windowBgnd));
}

/*------------------------------------------------------------------------------
| IWindow::hilitedForegroundColor (NOPed for Windows; returns default)         |
|                                                                              |
| Returns the hilite foreground color of the window.                           |
------------------------------------------------------------------------------*/
IColor IWindow::hiliteForegroundColor () const
{
//  return (color(PP_HILITEFOREGROUNDCOLOR,
//                IGUIColor(IGUIColor::hiliteFgnd)));
    return (IGUIColor(IGUIColor::hiliteFgnd));
}

/*------------------------------------------------------------------------------
| IWindow::hilitedBackgroundColor (NOPed for Windows; returns default)         |
|                                                                              |
| Returns the hilite background color of the window.                           |
------------------------------------------------------------------------------*/
IColor IWindow::hiliteBackgroundColor () const
{
//  return (color(PP_HILITEBACKGROUNDCOLOR,
//                IGUIColor(IGUIColor::hiliteBgnd)));
    return (IGUIColor(IGUIColor::hiliteBgnd));
}

/*------------------------------------------------------------------------------
| IWindow::activeColor (NOPed for Windows; returns default)                    |
|                                                                              |
| Returns the active color of the window.                                      |
------------------------------------------------------------------------------*/
IColor IWindow::activeColor () const
{
//  return (color(PP_ACTIVECOLOR,
//                IGUIColor(IGUIColor::activeFrameBorder)));
    return (IGUIColor(IGUIColor::activeFrameBorder));
}

/*------------------------------------------------------------------------------
| IWindow::inactiveColor (NOPed for Windows; returns default)                  |
|                                                                              |
| Returns the inactive color of the window.                                    |
------------------------------------------------------------------------------*/
IColor IWindow::inactiveColor () const
{
//  return (color(PP_INACTIVECOLOR,
//                IGUIColor(IGUIColor::inactiveFrameBorder)));
    return (IGUIColor(IGUIColor::inactiveFrameBorder));
}

/*------------------------------------------------------------------------------
| IWindow::setForegroundColor                                                  |
|                                                                              |
| Sets the foreground color of the window.                                     |
------------------------------------------------------------------------------*/
IWindow& IWindow::setForegroundColor (const IColor &color)
{
  return (setColor(PP_FOREGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setBackgroundColor                                                  |
|                                                                              |
| Sets the background color of the window.                                     |
------------------------------------------------------------------------------*/
IWindow& IWindow::setBackgroundColor (const IColor &color)
{
  return (setColor(PP_BACKGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setBorderColor (NOPed for Windows)                                  |
|                                                                              |
| Sets the border color of the window.                                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::setBorderColor (const IColor &color)
{
//  return (setColor(PP_BORDERCOLOR, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setShadowColor (NOPed for Windows)                                  |
|                                                                              |
| Sets the shadow color of the window.                                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::setShadowColor (const IColor &color)
{
//  return (setColor(PP_SHADOW, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setDisabledForegroundColor (NOPed for Windows)                      |
|                                                                              |
| Sets the disabled foreground color of the window.                            |
------------------------------------------------------------------------------*/
IWindow& IWindow::setDisabledForegroundColor (const IColor &color)
{
//  return (setColor(PP_DISABLEDFOREGROUNDCOLOR, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setDisabledBackgroundColor (NOPed for Windows)                      |
|                                                                              |
| Sets the disabled background color of the window.                            |
------------------------------------------------------------------------------*/
IWindow& IWindow::setDisabledBackgroundColor (const IColor &color)
{
//  return (setColor(PP_DISABLEDBACKGROUNDCOLOR, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setHiliteForegroundColor (NOPed for Windows)                        |
|                                                                              |
| Sets the hilite foreground color of the window.                              |
------------------------------------------------------------------------------*/
IWindow& IWindow::setHiliteForegroundColor (const IColor &color)
{
//  return (setColor(PP_HILITEFOREGROUNDCOLOR, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setHiliteBackgroundColor (NOPed for Windows)                        |
|                                                                              |
| Sets the hilite background color of the window.                              |
------------------------------------------------------------------------------*/
IWindow& IWindow::setHiliteBackgroundColor (const IColor &color)
{
//  return (setColor(PP_HILITEBACKGROUNDCOLOR, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setActiveColor (NOPed for Windows)                                  |
|                                                                              |
| Sets the active color of the window.                                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::setActiveColor (const IColor &color)
{
//  return (setColor(PP_ACTIVECOLOR, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::setInactiveColor (NOPed for Windows)                                |
|                                                                              |
| Sets the inactive color of the window.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::setInactiveColor (const IColor &color)
{
//  return (setColor(PP_INACTIVECOLOR, color));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetForegroundColor                                                |
|                                                                              |
| Resets the foreground color by undoing a previous set.                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetForegroundColor ()
{
  return (resetColor(PP_FOREGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetBackgroundColor                                                |
|                                                                              |
| Resets the background color by undoing a previous set.                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetBackgroundColor ()
{
  return (resetColor(PP_BACKGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetBorderColor (NOPed for Windows)                                |
|                                                                              |
| Resets the border color by undoing a previous set.                           |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetBorderColor ()
{
//  return (resetColor(PP_BORDERCOLOR));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetShadowColor (NOPed for Windows)                                |
|                                                                              |
| Resets the shadow color by undoing a previous set.                           |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetShadowColor ()
{
//  return (resetColor(PP_SHADOW));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetDisabledForegroundColor (NOPed for Windows)                    |
|                                                                              |
| Resets the disabled foreground color by undoing a previous set.              |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetDisabledForegroundColor ()
{
//  return (resetColor(PP_DISABLEDFOREGROUNDCOLOR));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetDisabledBackgroundColor (NOPed for Windows)                    |
|                                                                              |
| Resets the disabled background color by undoing a previous set.              |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetDisabledBackgroundColor ()
{
//  return (resetColor(PP_DISABLEDBACKGROUNDCOLOR));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetHiliteForegroundColor (NOPed for Windows)                      |
|                                                                              |
| Resets the hilite foreground color by undoing a previous set.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetHiliteForegroundColor ()
{
//  return (resetColor(PP_HILITEFOREGROUNDCOLOR));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetHiliteBackgroundColor (NOPed for Windows)                      |
|                                                                              |
| Resets the hilite background color by undoing a previous set.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetHiliteBackgroundColor ()
{
//  return (resetColor(PP_HILITEBACKGROUNDCOLOR));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetActiveColor (NOPed for Windows)                                |
|                                                                              |
| Resets the active color by undoing a previous set.                           |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetActiveColor ()
{
//  return (resetColor(PP_ACTIVECOLOR));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetInactiveColor (NOPed for Windows)                              |
|                                                                              |
| Resets the inactive color by undoing a previous set.                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetInactiveColor ()
{
//  return (resetColor(PP_INACTIVECOLOR));
    return *this;
}

/*------------------------------------------------------------------------------
| IWindow::resetColor                                                          |
|                                                                              |
| Worker function to remove presentation parameters from windows.              |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetColor (unsigned long colorArea)
{
#ifdef IC_PM
  if (!WinRemovePresParam (handle(), colorArea))
  {
    ITHROWLIBRARYERROR(IC_PRESPARAMNOTFOUND,
                       IBaseErrorInfo::invalidRequest,
                       IException::recoverable);
  }
#endif // IC_PM
#ifdef IC_WIN
  if ( (colorArea == PP_FOREGROUNDCOLOR ) &&
       (pWindowData->state & frgndColorWasSet) )
  {
    pWindowData->state &= (unsigned long)~frgndColorWasSet;
    refresh();

    /**************************************************************/
    /* Send private message to support window notify handler      */
    /* because Windows does not have presparam messages           */
    /**************************************************************/
    this->sendEvent( IC_UM_DELAYNOTIFY,
                     IEventParameter1( colorArea ),
                     IEventParameter2( IC_NOTDEL_WINDOW_COLOR ));
  }
  else if ( (colorArea == PP_BACKGROUNDCOLOR ) &&
       (pWindowData->state & bkgndColorWasSet) )
  {
    pWindowData->state &= (unsigned long)~bkgndColorWasSet;
    refresh();

    /**************************************************************/
    /* Send private message to support window notify handler      */
    /* because Windows does not have presparam messages           */
    /**************************************************************/
    this->sendEvent( IC_UM_DELAYNOTIFY,
                     IEventParameter1( colorArea ),
                     IEventParameter2( IC_NOTDEL_WINDOW_COLOR ));
  }
#endif // IC_WIN
  return (*this);
}


/*------------------------------------------------------------------------------
| IDialogControl::IDialogControl                                               |
------------------------------------------------------------------------------*/
IDialogControl::IDialogControl( )
  : fHwnd        (0),
    fOriginalProc(0),
    fDialog      (0)
{}

/*------------------------------------------------------------------------------
| IDialogControl::IDialogControl                                               |
------------------------------------------------------------------------------*/
IDialogControl::IDialogControl( const IDialogControl& control)
  : fHwnd        ( control.fHwnd ),
    fOriginalProc( control.fOriginalProc ),
    fDialog      ( control.fDialog )
{}

/*------------------------------------------------------------------------------
| IDialogControl::IDialogControl                                               |
------------------------------------------------------------------------------*/
IDialogControl& IDialogControl:: operator= ( const IDialogControl& control )
{
  fHwnd         = control.fHwnd;
  fOriginalProc = control.fOriginalProc;
  fDialog       = control.fDialog;
  return *this;
}

/*------------------------------------------------------------------------------
| IDialogControl::IDialogControl                                               |
------------------------------------------------------------------------------*/
IDialogControl::~IDialogControl   ()
{
}

/*------------------------------------------------------------------------------
| IDialogControls::IDialogControls                                             |
------------------------------------------------------------------------------*/
IDialogControls::IDialogControls () :
     IHashKeySet< IDialogControl, unsigned long>(19)
{
}

/*------------------------------------------------------------------------------
| IDialogControls::IDialogControls                                             |
------------------------------------------------------------------------------*/
IDialogControls::~IDialogControls ( )
{
  // Remove the subclassing for any valid windows
  IDialogControls::Cursor cursor( *this );
  for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext() )
     {
     if ( IsWindow( (HWND)cursor.element().fHwnd ) )
        {
        // Verify that no additional subclassing has occurred.  If not,
        // put back the original proc.  We can't do anything if the
        // control has been subclassed again.
        unsigned long  currentP = GetWindowLong( (HWND)cursor.element().fHwnd,
                                                 GWL_WNDPROC) ;
        if (currentP == (unsigned long)pfnwpICControlProc)
           {
           ISUBCLASSWINDOW( (HWND)cursor.element().fHwnd,
                            (unsigned long)cursor.element().fOriginalProc );
           }
        }
     }
}


/*------------------------------------------------------------------------------
| _pfnwpICControlProc                                                          |
------------------------------------------------------------------------------*/
unsigned long __stdcall pfnwpICControlProc(void*         hwnd,
                                           unsigned long ulMsg,
                                           void*         mp1,
                                           void*         mp2)
{
   IMODTRACE_ALL("pfnwpICControlProc");
   IDialogControls*        dialogControls =
      IThread::current().dialogControls();
   IDialogControls::Cursor cursor( *dialogControls );
   FARPROC                 originalProc = 0;
   if (dialogControls->locateElementWithKey( (unsigned long)hwnd, cursor) )
      {
      originalProc = (FARPROC) cursor.element().fOriginalProc;
      }

   if (originalProc)
      {
      switch (ulMsg)
         {
         case  WM_KEYDOWN:
         case  WM_KEYUP:
         case  WM_CHAR:
            // We need to route these messages to the parent IWindow to allow for
            // keyboard navigation.  We can't use the Windows IsDialogMessage
            // function because it requires knowledge of which dialog (read
            // IFrameWindow) we want to route events to.
            {
            ITRACE_DEVELOP( IString("pfnwpICControlProc hwnd=") +
                            IString((unsigned long)hwnd).d2x()+
                            IString(", Msg=") + IString(ulMsg).d2x()+
                            IString(", wParam=") + IString((unsigned long)mp1).d2x()+
                            IString(", lParam=") + IString((unsigned long)mp2).d2x() );
            unsigned long result = 0;
            IWindowHandle dialog = cursor.element().fDialog;

            if (!IWindow::windowWithHandle( dialog, false ))
               {
               // For some reason the dialog is no longer an IWindow.
               // Just call the original procedure and return.
               result =
                  CallWindowProc( originalProc,
                                  hwnd, (UINT)ulMsg, (WPARAM)mp1, (LPARAM)mp2);
               }
            else
               {
               // get the dialog code to see what this control is.
               unsigned long dlgCode =
                  SendMessage( hwnd, WM_GETDLGCODE, 0, 0);

               ITRACE_DEVELOP( IString("GETDLGCODE=") + IString(dlgCode).d2x() );

               // Special case for edit fields
               if ( ( ulMsg == WM_CHAR ) && ( dlgCode & DLGC_HASSETSEL ) )
                  {
                  IBase::Boolean toDialog = true;

                  // An edit field.  Need to check multiline edit controls
                  // for ES_WANTRETURN to determine return behavior.
                  unsigned long ulStyle = ISTYLEOF(hwnd);

                  // Note that WC_MLE and WC_ENTRYFIELD are same class in Win
                  // Have to check the style.
                  if ( ( IWindowClassName(hwnd) == WC_ENTRYFIELD ) &&
                       ( ulStyle & ES_MULTILINE) )
                     {
                     if ( ((unsigned long)mp1 != VK_RETURN ) ||
                          (ulStyle & ES_WANTRETURN) )
                        toDialog = false;
                     }

                  if ( toDialog &&
                       ( ((unsigned long)mp1 == VK_RETURN ) ||
                         ((unsigned long)mp1 == VK_ESCAPE ) ||
                         ((unsigned long)mp1 == VK_TAB    ) ) )
                     // Send to dialog and return result now
                     return dialog.sendEvent( ulMsg, mp1, mp2 );
                  }

               // Call the default procedure for the control.
               result =
                  CallWindowProc( originalProc,
                                  hwnd, (UINT)ulMsg, (WPARAM)mp1, (LPARAM)mp2);

               // If not processed, send event to the dialog.
               if (!result)
                  {
                  ITRACE_ALL("pfnwpICControlProc - to dialog");
                  result = dialog.sendEvent(ulMsg, mp1, mp2);
                  }
               } // else dialog is IWindow
            return result;
            }

         case  WM_MATCHMNEMONIC:
            // We implement WM_MATCHMNEMONIC  to allow canvas and frame
            // navigation.
            // IWindow
            return IWindowData::matchMnemonicHwnd(
                                   LOWORD((unsigned long)mp1),
                                   hwnd );


         case  WM_DESTROY:
            // remove this control from the list.  Don't bother removing
            // the subclass since the window is being destroyed
            dialogControls->removeAt( cursor );
            return 0;

         default:
            return CallWindowProc( originalProc,
                                   hwnd, (UINT)ulMsg, (WPARAM)mp1, (LPARAM)mp2);
         }  // switch
      }  // if originalProc
   else
      {
      // Can't find IWindow ... call default proc
      return DefWindowProc( hwnd, (UINT)ulMsg, (WPARAM)mp1, (LPARAM)mp2);
      }
}


/*------------------------------------------------------------------------------
| IWindowData::matchMnemonicHwnd                                               |
| This function used to implement WM_MATCHMNEMONIC.  Not to be confused        |
| with the non static matchForMnemonic which is overridden by canvases         |
| to check children for mnemonics.                                             |
------------------------------------------------------------------------------*/
unsigned long IWindowData::matchMnemonicHwnd ( unsigned short character,
                                               IWindowHandle hwnd )
{
   Boolean result = false;
   // only visible, enabled, child controls need to do anything
   if (  (ISTYLEOF( hwnd ) & (WS_CHILD | WS_DISABLED | WS_VISIBLE)) ==
         (WS_CHILD | WS_VISIBLE)  )
      {
      char  textBuffer[256];   // arbitrary max len
      int copied = GetWindowText(hwnd, textBuffer, 255);
      if (copied)
         {
         IString text(textBuffer);
         // find first non-escaped & character
         unsigned beginPos = text.indexOf('&');
         while ((beginPos != 0) &&
                (text.indexOf('&', beginPos+1) == beginPos+1) )
            beginPos = text.indexOf('&', beginPos + 2);

         if (beginPos)
            {
            IString target =
               IString("&") +
               IString( (char)character ) ;

            // See if there is a match
            if ((text.indexOf(target.upperCase(), beginPos) ==
                 beginPos) ||
                (text.indexOf(target.lowerCase(), beginPos) ==
                 beginPos))
               result =  true ;
            }  // found &
         }  // if has text
      }  // if child
   return result;
}


/*------------------------------------------------------------------------------
| IWindow::isDragStarting                                                       |
|
| Windows does not generate drag messages, so therefore it is up to a
| a control to decide if the conditions exist to start a drag. This is
| is the basic implementation that will just look for a down followed by
| a move more than xxx. If this occurs before we have an up event then
| we have a drag.
|
| Note: if you don't want the event to continue up the owner chain,
|       then set the event result to true.
------------------------------------------------------------------------------*/
Boolean IWindow::isDragStarting( IEvent &event )
{
  IMODTRACE_DEVELOP("IWindow::isDragStarting");
  Boolean brc = false;

  switch ( event.eventId() )
  {
    case WM_LBUTTONDOWN:
    case WM_RBUTTONDOWN:
    {
      ITRACE_DEVELOP("Got a button down, start to look for drag");
      IMouseClickEvent mevt( event );

      // We need to use ChildWindowFromPoint instead of WindowFromPoint
      // which IMouseEvent::windowUnderPointer uses to find disabled window
      POINT pt = { (short)mevt.parameter2().number1(),
                   (short)mevt.parameter2().number2() };
      IWindowHandle childHwnd ( ChildWindowFromPoint( mevt.handle(), pt) );

      if ( (childHwnd) && !(ISTYLEOF( childHwnd ) & WS_DISABLED) )
      {
        // In Windows 95, the system apparently captures the pointer itself.
        // during the down event and releases it in the up event.  If we
        // mess with this incorrectly it causes no WM_COMMANDs for the
        // button clicks to be sent.
        IWindowHandle captureHandle = GetCapture( );
        if (( captureHandle != NULL ) && (captureHandle != event.handle()) )
        {
          // This means that someone already has a capture...do nothing
//          SetCapture( captureHandle );
          ITRACE_DEVELOP(IString("Pointer already captured hwnd=") +
                         IString(captureHandle.asUnsigned()).d2x() +
                         IString(" event.handle()=")+
                         IString(event.handle().asUnsigned()).d2x());
        }
        else
        {
          if (captureHandle == NULL)
             {
             ITRACE_DEVELOP(IString("capturing pointer hwnd=") +
                            IString(event.handle().asUnsigned()).d2x());
             SetCapture( event.handle() );
             }
          pWindowData->bCheckingDragStart = True;
          pWindowData->pointLButtonDown =
              mevt.mousePosition();
        }
      }
    break;
    }

    case WM_LBUTTONUP:
    case WM_RBUTTONUP:
    {
      ITRACE_DEVELOP("Got a button up event");
      if ( pWindowData->bCheckingDragStart) //&&
      {                                     
         if (!( IPlatform::isWin95() || IPlatform::isNTNewShell() ))
           {
             // Win95 appears to release the capture when processing buttonup
             // If we release it we get no WM_COMMAND for button clicks.
             ITRACE_DEVELOP(IString("Releasing capture GetCapture=") +
                            IString((unsigned long)GetCapture()).d2x());
             SetCapture( NULL );
           }
           else
           {
             // On 95, if captured by us, and not already released by WM_MOUSEMOVE
             // processing, (for example user clicks and releases button, without moving
             // mouse) release it here.

              IWindowClassName className( event.handle());
              if ( className != WC_BUTTON )
              {
               IWindowHandle captureHandle = GetCapture( );    //asm
               if (( captureHandle != NULL ) && (captureHandle == event.handle()) )
               {
                  ITRACE_DEVELOP(IString("asm Releasing capture GetCapture=") +
                              IString((unsigned long)GetCapture()).d2x()  );
                  SetCapture( NULL );
               }
              }
           }
      }
      pWindowData->bCheckingDragStart = False;
      pWindowData->pointLButtonDown = IPoint(0,0);
      break;
    }
    case WM_MOUSEMOVE:
    {
      if (pWindowData->bCheckingDragStart)
      {
        ITRACE_DEVELOP("Got a move and DragStart is true");
        IMouseClickEvent mevt( event );
        IPoint mPoint( pWindowData->pointLButtonDown );
        if ( abs(mevt.mousePosition().x() - mPoint.x()) >
                 DD_DEFDRAGMINDIST ||
             abs(mevt.mousePosition().y() - mPoint.y()) >
                 DD_DEFDRAGMINDIST )
        {
          ITRACE_DEVELOP("We have a drag" );
          SetCapture( NULL );
          pWindowData->bCheckingDragStart = FALSE;
          //We have detected a drag occurring
          brc = true;
        }
      }
    }
    break;
#ifdef IC_TRACE_DEVELOP
    // Windows 95 only
    case WM_CAPTURECHANGED:
    {
       ITRACE_DEVELOP(IString("WM_CAPTURECHANGED new capture=") +
                      IString(event.parameter2().asUnsignedLong()).d2x() +
                      IString(" event.handle()=")+
                      IString(event.handle().asUnsigned()).d2x());
    }
    break;
#endif
  }
  return brc;
}
#endif // IC_WIN
